// --- C++ ---
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2023 CEA/DEN
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   TLauncherEnv.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2023-02-22
 
 \brief  working dir for launcher code
*/


// ROOT
//#include "

// Uranie
#include "TLauncherEnv.h"

#include "TSystem.h"

using std::cout;
using std::endl;

ClassImp(URANIE::Launcher::TLauncherEnv)

namespace URANIE {
namespace Launcher {

// Constructor Destructor
TLauncherEnv::TLauncherEnv() :
    _code(NULL), _sdirectory(gSystem->pwd()), _sworkingDirectory(""), _blog(kFALSE)
{
}

TLauncherEnv::TLauncherEnv(TCode *code) :
    _code(code), _sdirectory(gSystem->pwd()), _sworkingDirectory(""), _blog(kFALSE)
{
    try {
        if (code == NULL) {
             throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                  "TLauncherEnv::constructor ERROR : the pointer of\"code\" is empty");
        }
    }
    catch(URANIE::Exceptions::UErrorExceptions &ue) {
        ue.printMessage();
        throw ue;
    }
}

TLauncherEnv::~TLauncherEnv() {}

// work dir
void TLauncherEnv::verifyWorkingDirectory()
{

    try {
        // The working directory is not specified in the TLauncher object
        if (_sworkingDirectory.Length() == 0) {
            // Get the working directory in the TCode object
            _sworkingDirectory = _code->getWorkingDirectory();
            // Else the current directory is current path/URANIE
            if (_sworkingDirectory.Length() == 0) {
                _sworkingDirectory = gSystem->pwd();
		        char *foo=gSystem->ConcatFileName( _sworkingDirectory.Data(), "URANIE"); //Needed prevent memleak
                _sworkingDirectory = foo;
                delete[] foo;
	        }
        }
        if (_sworkingDirectory == gSystem->pwd()) {
	        char *foo=gSystem->ConcatFileName(_sworkingDirectory.Data(), "URANIE");
            _sworkingDirectory = foo;
	        delete[] foo;
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                Form(
                "Beware! You have set the current directory as your working directory. Performing a setClean() method will delete the current directory. The directory %s/URANIE is created as working directory.",
                _sworkingDirectory.Data()));
        }
        if (_sworkingDirectory[0] == '.') {
            TString tmpDir = _sworkingDirectory;
            tmpDir.Remove(0, 1);
            tmpDir.Prepend(_sdirectory);
            _sworkingDirectory = tmpDir;
        }
        if (_code->getNOutputFiles() == 0) {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
		       "Beware! The list of output files is empty. You have to set the output files by using the addOutputFile method on the TCode object.");
        }
    }
    catch (URANIE::Exceptions::UWarningExceptions& ue) {
        ue.printMessage();
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue) {
        ue.printMessage();
        throw(ue);
    }

    _code->setWorkingDirectory(_sworkingDirectory);
}

///////////////////////////////////////////////////////////
//  createResultsDirectory Method
///////////////////////////////////////////////////////////  
void TLauncherEnv::createResultsDirectory()
{
    // Make the Results directory for distribution
    try {
        char *foo=gSystem->ConcatFileName(_sworkingDirectory.Data(),"UranieResults");
        TString sresultsDirectory = foo;
        delete[] foo;
        void * openDirResult = gSystem->OpenDirectory(sresultsDirectory);
        if (openDirResult == 0)
	    if (gSystem->mkdir(sresultsDirectory, kTRUE) == -1)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
            Form(
            "TLauncher::createResultsDirectory ERROR\n the subdirectory \"UranieResults\" can not be created in the workindirectory \"%s\" with name\"%s\"",
            _sworkingDirectory.Data(), sresultsDirectory.Data())
        );
        gSystem->FreeDirectory(openDirResult);
    } 
    catch (URANIE::Exceptions::UExceptions& ue) {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
	    throw ue;
    }
}


///////////////////////////////////////////////////////////
//  Clean Method
///////////////////////////////////////////////////////////
bool TLauncherEnv::cleanWorkingDirectory()
{
    // Go to the working directoryPlace in the good directory
    if (_sworkingDirectory.Length() != 0) {
#ifdef WIN32
        gSystem->Exec(Form("if exist %s rmdir /s /q %s", _sworkingDirectory.Data(), _sworkingDirectory.Data()));
#else
        gSystem->Exec(Form("rm -fr %s", _sworkingDirectory.Data()));
#endif
        return true;
    }
    else return false;

}


// ----------------------------------------------------------------------
// setWorkingDirectory() method
// ----------------------------------------------------------------------
//-----------------------------------------------------------------------
void TLauncherEnv::setWorkingDirectory(TString str)
{
    TString strDirectory = str;
    try {
        if ((strDirectory.EndsWith("\\")) || (strDirectory.EndsWith("/"))) {
            strDirectory.Remove(strDirectory.Length() - 1);
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                Form(
                "TLauncher::setWorkingDirectory method\n the working directory \"%s\" is redefined without last backslash/slash \"%s\"",
                str.Data(), strDirectory.Data())
            );
        }
    }
    catch (URANIE::Exceptions::UExceptions& ue) {
        ue.printMessage();
    }
    _sworkingDirectory = strDirectory;
    //if (_blog) cout << " *** _sworkingDirectory [" << _sworkingDirectory << "]" << endl;
}


}} // namespace
