//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2025 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLauncher.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9 11:37:19 CEST 2006
 
 \brief Implementation de la classe URANIE::Launcher::TLauncher
 
 */

#include "TLauncherMulti.h"
#include "TSlurmCluster.h"
#include "TLSFCluster.h"
#include "TSGECluster.h"
#include "TLocalHostCluster.h"
#include "TLoadLevelerCluster.h"
#include "forkcode.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <unistd.h>
#include <fstream>

#include "TSystem.h"

#include "TFile.h"
#include "TChain.h"

using std::cout;
using std::cerr;
using std::endl;

ClassImp (URANIE::Launcher::TLauncherMulti);

namespace URANIE
{
namespace Launcher
{

TLauncherMulti::TLauncherMulti() :
        TNamed("_uranie_TLauncherMulti_", "multi lancher uranie"),
        _blog(kFALSE), _bemptyMasterNode(kFALSE), njobs_(0), nb_avail_procs_(1)
{
    _edistrib = NULL;
    createHostsList("local=2");
    if (_blog)
        cout << "end of TLauncherMulti () constructor" << endl;
}
TLauncherMulti::TLauncherMulti(const char* name, const char* title,
                               Option_t *opt, Bool_t emptyMasterNode, bool dumpmulti) :
        TNamed(name, title),
        _blog(kFALSE), _bemptyMasterNode(emptyMasterNode), _bmultidump(dumpmulti), 
        njobs_(0), nb_avail_procs_(1)
{
    _edistrib = NULL;
    createHostsList(opt);
    if (_blog)
        cout << "end of TLauncherMulti (name,title,opt) constructor" << endl;
}

TLauncherMulti::~TLauncherMulti()
{
    if (_edistrib)
        delete _edistrib;
    _edistrib = NULL;
}

void TLauncherMulti::createHostsList(Option_t *opt)
{
    TString sListOfHosts = "";
    const char *label = "Unknown";

    if (_blog)
        cerr << " ** sListOfHosts[" << sListOfHosts << "]" << endl;

    if (!_edistrib) {
        if (TSlurmCluster::eligible()) 
        {
            _edistrib = new TSlurmCluster();
            if (_blog)
                _edistrib->setLog();
            label = "SLUM_NODELIST";
        }
        else if (TLoadLevelerCluster::eligible())
        {
            _edistrib = new TLoadLevelerCluster();
            label = "TLoadLevelerCluster";
        }
        else if (TLSFCluster::eligible())
        {
            _edistrib = new TLSFCluster();
            label = "LSF_HOSTS";
        }
        else if (TSGECluster::eligible())
        {
            _edistrib = new TSGECluster();
            label = "SGE_HOSTFILE";
        }
        else if (TLocalHostCluster::eligible(opt))
        {
            _edistrib = new TLocalHostCluster();
            ((TLocalHostCluster*)_edistrib)->setMultiDump(_bmultidump);
            label = "LocalHost";
        }
    }

    // In case of any TLocalHost::eligible is OK, we are in sequential mode
    if (_edistrib == NULL)
    {
        if (_blog)
            cout << " ** Mode : Sequential" << endl;
        return;
    }
    else {
        sListOfHosts = _edistrib->getListOfNodes();
        cerr << " ***************** " << label << " = [" << _edistrib->GetTitle()
                << "]" << endl;
        cerr << " ***************** " << label << " = [" << sListOfHosts << "]"
                << endl;
        cout << " ***************** " << label << " = [" << sListOfHosts << "]"
                << endl;
    }

    if (_blog)
        cout << " ** Mode : Distribution " << endl;

    istringstream hostsTmp((char *) sListOfHosts.Data());

    while (!hostsTmp.eof())
    {
        string tmp;
        hostsTmp >> tmp;
        if (tmp.size())
            hostslist_.push_back(tmp);
    }
    // eliminating the procs on the master node for large jobs (bottleneck problem)
    if (_bemptyMasterNode)
    {
        string master = *(hostslist_.begin());
        list<string>::iterator iter = hostslist_.begin();
        while (iter != hostslist_.end())
        {
            if (strcmp(iter->c_str(), master.c_str()) == 0)
                iter = hostslist_.erase(iter);
            else
                iter++;
        }
    }
    nb_avail_procs_ = hostslist_.size();
    nb_total_procs_ = hostslist_.size();
    //cout << "nombre de pros dispos : " << nb_avail_procs_ << endl;

    if (_blog)
    {
        cout << " ** MASTER JOB LAUNCHED, Proc[" << nb_avail_procs_ << "]" << endl;
        cout << " ** LIST OF HOSTS **" << _edistrib->getListOfNodes() << endl;
        list<string>::iterator iterHost, iterEnd;
        for (iterHost = hostslist_.begin(), iterEnd = hostslist_.end();
                iterHost != iterEnd; iterHost++)
            cout << "\t" << *iterHost;
    }
}

void TLauncherMulti::defineJob(const char* exe, int njobs, int nprocs)
{
    exe_ = " ";
    exe_ += exe;

    njobs_ = njobs;
    procsperjob_ = nprocs;
    if (0)
    {
        cout << "\texe : " << exe_ << endl;
        cout << "\tnjobs : " << njobs_ << endl;
        cout << "\tnombre de pros dispos : " << nb_avail_procs_ << endl;
    }

}

bool TLauncherMulti::availableRessource()
{
    return (nb_avail_procs_ >= procsperjob_);
}

void TLauncherMulti::createJob(forkcode *fjob, TCode *code)
{
    list<string>::iterator iterHost, iterEnd;

    // ---------------------------------------------
    // ---------- PROCS AVAILABLE, CREATE NEW ------
    // ---------------------------------------------
    char mpiexe[2500];
    char procs[2000];
    char singlehost[200];
    list < string > tmpHosts;
    list<string>::iterator iter = hostslist_.begin();
    
    strcpy(procs, "");
    for (int nproc = 1; nproc <= procsperjob_; nproc++)
    {
        tmpHosts.push_back(*iter);
        strcpy(singlehost, iter->c_str());
        strcat(procs, singlehost);
        if (nproc < procsperjob_)
            strcat(procs, ",");
        iter++;
        hostslist_.pop_front();
    }

    if( _bmultidump )
    {
        cout << " ** TLauncherMulti::createJob method :: Job Created, pid = "
             << getpid() << " Iteration[" << fjob->pattern + 1 << "] host["
             << procs << "]" << endl;
        cerr << " ** TLauncherMulti::createJob method :: Job Created, pid = "
             << getpid() << " Iteration[" << fjob->pattern + 1 << "] host["
             << procs << "]" << endl;
    }
    // prefix command
    if (_edistrib) _edistrib->getCommandPrefix(mpiexe, procs);
    fjob->command = mpiexe;
    fjob->command += " ";
    fjob->command += code->getCommand();
    if (_blog)
    {
        cout << " ** createJob method :: Debug de la Ligne de commande["
                << mpiexe << "]" << endl;
        cerr << " ** createJob method :: Debug de la Ligne de commande["
                << mpiexe << "]" << endl;
    }
    njobs_--;
    nb_avail_procs_ -= procsperjob_;
    if (_blog)
        cout << " ** TLauncherMulti::createJob method :: fork with nb_avail ["
            << nb_avail_procs_ << "] slots" << endl;
}


void TLauncherMulti::mergeAll(const int niteration)
{
    TString expname = GetName();
    cerr << " **  TLauncherMulti::mergeAll Name[" << expname << "]" << endl;

    cout << "TLauncherMulti::mergeAll(" << niteration << ")" << endl;
    cerr << "TLauncherMulti::mergeAll(" << niteration << ")" << endl;
    cout << "expname[" << expname << "]" << endl;
    cerr << "expname[" << expname << "]" << endl;
    TChain ch(expname);
    TString sfilsfile;

    for (int i = 0; i < niteration; i++)
    {
        sfilsfile = _sworkingDirectory + "/UranieResults/uranielauncher_";
        sfilsfile += i;
        sfilsfile += ".root";
        std::ifstream file(sfilsfile);
        if (file.fail())
        {
            cout << "Skip " << sfilsfile
                    << " for merging because it doesn't exist!" << endl;
            cerr << "Skip " << sfilsfile
                    << " for merging because it doesn't exist!" << endl;
        }
        else
        {
            ch.Add(sfilsfile);
        }
    }

    sfilsfile = _sworkingDirectory + "/UranieResults/uranielauncher.root";
    ch.Merge(sfilsfile);
}

void TLauncherMulti::display()
{

    std::ofstream os(_sworkingDirectory + "/UranieResults/display.C");

    os << "void display() {" << endl << endl;
    ;
    // On ouvre le fichier ROOT "uranielauncher.root" en mode update
    os << "\tTFile *rt = new TFile(\"uranielauncher.root\",\"read\");" << endl;
    // Informations de ce fichier
    os << "\trt->Print();" << endl;
    // Listing du contenu du fichier
    os << "\trt->ls();" << endl;
    os << endl << "}" << endl;

    os.close();

}

// ----------------------------------------------------------------------
// Printing Log
// ----------------------------------------------------------------------
//---------------------------------------------
void TLauncherMulti::printLog(Option_t *option)
{
    cout << "TLauncherMulti::printLog[" << option << "]" << endl;
    cout << "Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << " _edistrib[" << _edistrib << "]" << endl;
    cout << " njobs_[" << njobs_ << "]" << endl;
    cout << " nb_avail_procs_[" << nb_avail_procs_ << "]" << endl;
    cout << " _sworkingDirectory[" << _sworkingDirectory << "]" << endl;
}

} // Fin du namespace LAUNCHER
} // Fin du namespace URANIE

