//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2025 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*! \file TMetropHasting.cxx
 *
 */


#include <fstream>

//ROOT
#include "TFile.h"
#include "TH1.h"
#include "TCanvas.h"
#include "TMath.h"
#include "TROOT.h"
#include "TRandom3.h"

//URANIE
#include "TMetropHasting.h"
#include "TStochAttribut.h"

ClassImp (URANIE::Calibration::TMetropHasting)

namespace URANIE
{
namespace Calibration
{

// -----------------------------------------------------------------------
//  Constructors & Initialisation
// -----------------------------------------------------------------------

TMetropHasting::TMetropHasting(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run, int ns, Option_t *option) :
    TCalibration(tds, run, ns, option), _burnin(-1), _lag(-1), _nbDump(1000), _lowAccRange(-1.), _higAccRange(-1.), _mbGDVersion(false), _bcleaningAtt(false)
{
    if (_blog)
        cout << "constructor TMetropHasting by TDataServer[" << tds->GetName() << "]"
	     << endl;

    parseOption(option);
    
    checktdsParContent();
 
}

TMetropHasting::TMetropHasting(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    TCalibration(tds, fcn, varexpinput, varexpoutput, ns, option), _burnin(-1), _lag(-1), _nbDump(1000), _lowAccRange(-1.), _higAccRange(-1.), _mbGDVersion(false), _bcleaningAtt(false)
{

    parseOption(option);

    checktdsParContent();
}

TMetropHasting::TMetropHasting(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    TCalibration(tds, fcn, varexpinput, varexpoutput, ns, option), _burnin(-1), _lag(-1),_nbDump(1000), _lowAccRange(-1.), _higAccRange(-1.), _mbGDVersion(false), _bcleaningAtt(false)
{

    parseOption(option);

    checktdsParContent();
}

TMetropHasting::TMetropHasting(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns, Option_t *option):
    TCalibration(tds, fcode, ns, option), _burnin(-1), _lag(-1),_nbDump(1000), _lowAccRange(-1.), _higAccRange(-1.), _mbGDVersion(false), _bcleaningAtt(false)
{
    TString opt = option;
    opt.ToLower();
    
    checktdsParContent();
}

void TMetropHasting::checktdsParContent()
{
    // Dans le cas d'un plan SRS
    setMethodName("MetropolisHasting");

    if(_blog)
        XYZ<<"In checktdsParContent"<<endl;

    // Get the number of attributes
    _nPar=_tdsPar->getNAttributes();
    TIter next(_listOfParameters);
    URANIE::DataServer::TAttribute *att=NULL;
    URANIE::DataServer::TStochasticAttribute *stoatt=NULL;
    try
    {
        while ((att = (URANIE::DataServer::TAttribute *) next()))
        {
            stoatt = dynamic_cast<URANIE::DataServer::TStochasticAttribute *>(att);
            if(stoatt==0)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TMetropHasting::constructor\n\
 The provided dataserver must contains stochastic attributes. Attribute [%s] is not one of them", att->GetName()));

            _vatt.push_back( stoatt );
            if (_blog)
                cout << " ** Storing " << att->GetName() << "] in our attribute vector"<<endl;
        }
        
        if( _vatt.size()==0 )
	        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TMetropHasting: construction with tds not containing any TStochasticAttribute");

        // number of attributes different from number of stochastic attributes
        if( int(_vatt.size()) != _nPar )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                Form("TMetropHasting: construction with _nPar [%d] different from _vatt content [%lu]", _nPar, _vatt.size()));
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    
}

void TMetropHasting::setAcceptationRatioRange(double lower, double higher)
{
    try
    {
        if( lower >= higher )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                Form("TMetropHasting::setAcceptationRatioRange. Lower bound [%g] should be lower than the higher one [%g]", lower, higher));
            if( lower > 1.0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                Form("TMetropHasting::setAcceptationRatioRange. Lower bound [%g] is greater than 1.0. It should not be.", lower));
            if( lower < 0.0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                Form("TMetropHasting::setAcceptationRatioRange. Lower bound [%g] is lower than 0.0. It should not be.", lower));
            if( higher > 1.0 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                Form("TMetropHasting::setAcceptationRatioRange. Higher bound [%g] is greater than 1.0. It should not be.", higher));
            if( higher < 0.0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                Form("TMetropHasting::setAcceptationRatioRange. Higher bound [%g] is higher than 0.0. It should not be.", higher));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _lowAccRange=lower;
    _higAccRange=higher;
}

void TMetropHasting::setInitialisation(int n, double *values, double *standDev)
{
    try
    {
	    if( n != _nPar )
	        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMetropHasting::setInitialisation. Arrays should contain %d elements. The number provided is %d",
							    _nPar,n));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    try
    {
        // initialise the parameters values
        for (int iatt=0;iatt<_nPar; iatt++)
        {
            _values.push_back(values[iatt]);
            if( standDev[iatt] < 0 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TMetropHasting::setInitialisation. Standard deviation provided is negative [%g]. This cannot be !",
                                                                standDev[iatt]));
            _vstd.push_back(standDev[iatt]);
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    
}

void TMetropHasting::setInitialisation(vector<double> values, vector<double> standDev)
{
    try
    {
	    if( (int(values.size()) != _nPar) || (int(standDev.size()) != _nPar) )
	        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMetropHasting::setInitialisation. Vectors should contain [%d] elements. Value and standard deviation does only have [%lu] and [%lu]",
							    _nPar,values.size(),standDev.size()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    setInitialisation(_nPar, &values[0], &standDev[0]);
}


// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TMetropHasting::~TMetropHasting()
{
    if (_blog)
        cout << "************ Destructor URANIE::Calibration::TMetropHasting" << endl;

    if (_blog)
        cout << "************ End of Destructor URANIE::Calibration::TMetropHasting"<< endl;
}


// -----------------------------------------------------------------------
//  Setters
// -----------------------------------------------------------------------

void TMetropHasting::setDistanceAndReference(URANIE::Calibration::TDistanceFunction *distFunc, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *output, const char *weight)
{
    UNUSED(distFunc);
    TCalibration::setDistanceAndReference("weightedls", tdsRef, input, output, weight);
}

void TMetropHasting::setDistanceAndReference(const char *funcName, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *output, const char *weight)
{
    UNUSED(funcName);
    TCalibration::setDistanceAndReference("weightedls", tdsRef, input, output, weight);
}
   

// -----------------------------------------------------------------------
//  Compute the parameters
// -----------------------------------------------------------------------

void TMetropHasting::computeParameters(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    
    parseOption(option); 

    int iatt=0;

    if(_blog)
        XYZ<<"In computeParameters"<<endl;

    // If computeParameters has already been run, we need to delete the log_prior, log_likelihood, log_post_unormalized, agreement and acceptation_ratio attribute to restart the computation
        
    _tdsPar->delAttribute("agreement");        

    try
    {
        if(_tdsPar->isAttribute("log_prior")==1||_tdsPar->isAttribute("log_likelihood")==1||_tdsPar->isAttribute("log_post_unormalized")==1)
            throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__, "TMetropHasting::computeParameters\n\
                    A new calculation will begin from scratch. To continue the previous calculation, please use the continueCalculation function.");    
    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        ue.printMessage();

        _tdsPar->delAttribute("log_prior");
        _tdsPar->delAttribute("log_likelihood");
        _tdsPar->delAttribute("log_post_unormalized");

        for (int iatt=0;iatt<_nPar; iatt++)
        {
            if (_tdsPar->isAttribute(Form("acceptation_ratio_%s",_vatt.at(iatt)->GetName())))
                _tdsPar->delAttribute(Form("acceptation_ratio_%s",_vatt.at(iatt)->GetName()));
        }

        delete _listOfParameters;
        _listOfParameters = NULL;

        _listOfParameters=_tdsPar->getListOfAttributes();
        _nPar=_listOfParameters->GetEntries();
    }

    if(!_bcleaningAtt)
    {    
        //Add attributes corresponding to the values of acceptation ratio per components
        for (int iatt=0;iatt<_nPar; iatt++)
        {
            _tdsPar->addAttribute(new URANIE::DataServer::TAttribute(Form("acceptation_ratio_%s",_vatt.at(iatt)->GetName())));
            _tdsPar->getAttribute(Form("acceptation_ratio_%s",_vatt.at(iatt)->GetName()))->delShare();
        }
        
        //Add the log prior as an attibute
        _tdsPar->addAttribute(new URANIE::DataServer::TAttribute("log_prior"));
        _tdsPar->getAttribute("log_prior")->delShare();
        
        //Add the log likelihood as an attribute
        _tdsPar->addAttribute(new URANIE::DataServer::TAttribute("log_likelihood"));
        _tdsPar->getAttribute("log_likelihood")->delShare();  

        //Add the unormalized log posterior as an attribute
        _tdsPar->addAttribute(new URANIE::DataServer::TAttribute("log_post_unormalized"));
        _tdsPar->getAttribute("log_post_unormalized")->delShare();   
    }

    // delete the TDSNtupleD and the saved vectors (_dBaseSaved, _acceptSaved, _rejectSaved) if it exists to restart the computation
    try
    {
        if (_tdsPar->hasTuple())
            throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                      Form("TMetropHasting::computeParameters\n\
    The TDS [%s] has been cleared to restart the process.",_tdsPar->GetName()));
    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        ue.printMessage();
        _tdsPar->deleteTuple();

        // delete the saved data

        if (!_dBaseSaved.empty())
            _dBaseSaved.clear();

        if (!_acceptSaved.empty())
            _acceptSaved.clear();

        if (!_rejectSaved.empty())
            _rejectSaved.clear();
    }

    _tdsPar->addAttribute("agreement");  

    // Initialise the random generator
    TRandom3* rand = new TRandom3();
    if(_nSeed != 0)
        rand->SetSeed(_nSeed);

    // initialise the parameters values
    if( _values.size()==0 && _vstd.size()==0)
    {
        for (int iatt=0;iatt<_nPar; iatt++)
        {
            _values.push_back(_vatt.at(iatt)->getInvCDF(rand->Rndm()));
            _vstd.push_back( _vatt.at(iatt)->getTheoreticalStdDev());
        }
    }
    
    // Initialise the first ratio
    double log_prior_value=0, log_lik_value=0, loglik_prop=0.;
    logPriorPdf(log_prior_value);
    _dFunc->eval(&_values[0],&log_lik_value);
    log_lik_value*=-0.5;
    
    if(_mbGDVersion)
    {
        _tdsPar->addAttribute("yes_or_no");

        int baseSize=2*_nPar+6;

        // Create the database
        vector<double> dBase(baseSize,0.);
        for(int iatt=0; iatt<_nPar; iatt++)
        {
            dBase[iatt+1]=_values.at(iatt);
            dBase[_nPar+1+iatt]=1.;
        }

        dBase[2*_nPar+1]=log_prior_value; 
        dBase[2*_nPar+2]=log_lik_value; 
        dBase[2*_nPar+3]=log_lik_value+log_prior_value;
        dBase[2*_nPar+4]=1.;

        double retval=0;
        _dFunc->eval(&_values[0], &retval);
        dBase[2*_nPar+5]=retval;

        _tdsPar->createTuple();
        _tdsPar->getTuple()->Fill(&dBase[0]);

        vector<int> accept(_nPar, 0);
        vector<int> reject(_nPar, 0);
        vector<int> LenBurn(_nPar, 0);

        // Initialise
        int Niter=1;    
        vector<bool> vOK(_nPar, false);
        vector<bool> vBatchDone(_nPar, false);
        bool PrintDone=false;
        int LenBatch=100;
        
        for (int ievt=0; ievt<_nSam; ievt++)
        {                                	
            for(int iatt=0; iatt<_nPar; iatt++)
            {
                if ( ( Niter%_nbDump==0 ) && ! PrintDone  )
                {
                    cout<<Niter<<" events done"<<endl;
                    PrintDone=true;
                }

                _values[iatt]=rand->Gaus( dBase[iatt+1], _vstd[iatt] );

                if (_blog)                    
                    printf("From (%6.6f,%6.6f) => %6.6f\n",dBase[iatt+1],_vstd[iatt],_values.at(iatt));
                
                if ( ( _values[iatt] < _vatt.at(iatt)->getLowerBound() )  || ( _values[iatt] > _vatt.at(iatt)->getUpperBound() ) )
                {
                    dBase[0]=double(Niter);
                    dBase[2*_nPar+4]=0.;
                    reject[iatt]+=1;
                    dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                    _values[iatt]=dBase[iatt+1];
                    _tdsPar->getTuple()->Fill(&dBase[0]);
                    Niter+=1;
                    continue;
                }                   

                log_prior_value=0.;
                // Recompute the proba with new values.
                logPriorPdf(log_prior_value);                        
                _dFunc->eval(&_values[0], &loglik_prop);
                dBase[2*_nPar+5]=loglik_prop;
                log_lik_value*=-0.5;

                // Recompute the ratio	    
                double value_prop=loglik_prop+log_prior_value;
                double log_ratio=value_prop-dBase[2*_nPar+3];
                double log_rand=log(rand->Rndm());

                //cout<<" Niter "<<Niter<<" Value  "<<_values[0]<<" ==> logratio "<<log_ratio<<"  value_prop "<<value_prop<<" - dbase  "<<dBase[2*_nPar+3]<<"  logrand  "<<log_rand<<endl;
                if ( log_ratio > log_rand )
                {
                    //cout<<" ACCEPTED "<<Niter<<endl;
                    dBase[0]=double(Niter);
                    accept[iatt]+=1;
                    dBase[2*_nPar+1]=log_prior_value;
                    dBase[2*_nPar+2]=loglik_prop;
                    dBase[2*_nPar+3]=value_prop;
                    dBase[2*_nPar+4]=1;
                    dBase[iatt+1]=_values[iatt];
                    dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                    PrintDone=false;
                    // cout<<"Info are ";
                    // for(int info=0; info<2*_nPar+4; info++)
                    //     cout<<info<<":"<<dBase[info]<<", ";
                    // cout<<endl;                    
                    _tdsPar->getTuple()->Fill(&dBase[0]);
                }
                else
                {
                    //cout<<" REJECTED"<<endl;;
                    dBase[0]=double(Niter);
                    dBase[2*_nPar+4]=0;                   
                    reject[iatt]+=1;
                    _values[iatt]=dBase[iatt+1];
                    dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                }

                Niter++;
                vOK[iatt]=true;
                vBatchDone[iatt]=false;
                if ( ievt%LenBatch==0  && _lowAccRange != -1. && _higAccRange !=-1)
                {
                    if( dBase[_nPar+1+iatt] < _lowAccRange )
                    {
                     	LenBurn[iatt]=double(Niter);
                        accept[iatt]=0;
                        reject[iatt]=0;
                        _vstd[iatt]=0.9*_vstd[iatt];
                        dBase[_nPar+1+iatt]=0.;
                    }
                    else if( dBase[_nPar+1+iatt] > _higAccRange )
                    {
                        LenBurn[iatt]=double(Niter);
                        accept[iatt]=0;
                        reject[iatt]=0;
                        _vstd[iatt]=1.1*_vstd[iatt];
                        dBase[_nPar+1+iatt]=0.;
                    }
                    vBatchDone[iatt]=true;
                }

                // if(reject[iatt]>=10)
                // 	return;
            }// _nPar
        }// _nS
        
        // Save the last dBase, the number of acceptation and rejection in case the computation is continued
        
        _dBaseSaved = dBase; 
        _acceptSaved = accept;
        _rejectSaved = reject;

        int burn=TMath::MaxElement(_nPar,&LenBurn[0]);
        cout<<"la longueur de chauffe pour un taux d'acceptation acceptable est de %d"<<burn<<endl;
    }
    else // JBB
    {
        int baseSize=2*_nPar+5;

        // Create the database
        vector<double> dBase(baseSize,0.);
        for(int iatt=0; iatt<_nPar; iatt++)
        {
            dBase[iatt+1]=_values.at(iatt);
            dBase[_nPar+1+iatt]=1.;
        }

        dBase[2*_nPar+1]=log_prior_value; 
        dBase[2*_nPar+2]=log_lik_value; 
        dBase[2*_nPar+3]=log_lik_value+log_prior_value;

        double retval=0;
        _dFunc->eval(&_values[0], &retval);
        dBase[2*_nPar+4]=retval;

        _tdsPar->createTuple();
        // XYZ<<"DBase content: ";
        // for(unsigned int ip=0; ip<2*_nPar+4; ip++)
        //     cout<<dBase.at(ip)<<", ";
        // cout<<endl;
        
        _tdsPar->getTuple()->Fill(&dBase[0]);

        vector<int> accept(_nPar, 0);
        vector<int> reject(_nPar, 0);

        // The proposal distribution is specified through a TRamdom3 object
    
        // Initialise
        int Niter=1;    
        vector<bool> vOK(_nPar, false);
        vector<bool> vBatchDone(_nPar, false);
        bool PrintDone=false;
        int LenBatch=100;

        while( Niter < _nSam )
        {                                	
            for(int iatt=0; iatt<_nPar; iatt++)
            {

                // Print the iteration every _nbDump step
                if ( ( Niter%_nbDump==0 ) && ! PrintDone  )
                {
                    cout<<Niter<<" events done"<<endl;
                    PrintDone=true;
                }
            
                // Draw a candidate from the proposal distribution
                _values[iatt]=rand->Gaus( dBase[iatt+1], _vstd[iatt] );

                // If the candidate value is out of bound, we reject it directly, actualise the rejection vector and restart a new step in the while loop
                if ( ( _values[iatt] < _vatt.at(iatt)->getLowerBound() )  || ( _values[iatt] > _vatt.at(iatt)->getUpperBound() ) )
                {
                    reject[iatt]+=1;
                    _values[iatt]=dBase[iatt+1];
                    dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                    continue;
                }                   

                // If not, we recompute the proba with new value
                logPriorPdf(log_prior_value);                        
                _dFunc->eval(&_values[0], &loglik_prop);
                dBase[2*_nPar+4]=loglik_prop;
                loglik_prop*=-0.5;

                // Recompute the ratio	    
                double value_prop=loglik_prop+log_prior_value;
                double log_ratio=value_prop-dBase[2*_nPar+3];
                
                // Draw uniformly a double between 0 and 1
                double log_rand=log(rand->Rndm());
                
                dBase[0]=float(Niter);
                
                // If the ratio is superior to the random double, we keep the candidate, update the dBase vector and fill the TDS 
                if ( log_ratio > log_rand )
                {
                    accept[iatt]+=1;
                    dBase[2*_nPar+1]=log_prior_value;
                    dBase[2*_nPar+2]=loglik_prop;
                    dBase[2*_nPar+3]=value_prop;
                    dBase[iatt+1]=_values[iatt];
                    dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                    Niter+=1;
                    PrintDone=false;

                    _tdsPar->getTuple()->Fill(&dBase[0]);
                }
                // If not, the candidate is rejected and we update the rejection vector 
                else
                {
                    reject[iatt]+=1;
                    _values[iatt]=dBase[iatt+1];
                    dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                }
                
                vOK[iatt]=true;
                vBatchDone[iatt]=false;

                // If bounds on the acceptation ratio have been defined, 
                if ( (Niter-2-iatt)%LenBatch==0 && ! vBatchDone[iatt] && _lowAccRange != -1 && _higAccRange != -1)
                {
                    // And if the current acceptation ratio is too small,
                    if( dBase[_nPar+1+iatt] < _lowAccRange )
                    {
                        // We reduce the standard deviation of the proposal in order that the candidate are more likely to be accepted
                        _vstd[iatt]=0.9*_vstd[iatt];
                    }
                    // And if the current acceptation ratio is too big,
                    else if( dBase[_nPar+1+iatt] > _higAccRange )
                    {
                        // We increase the standard deviation of the proposal in order that we explore more efficiently the parameter space
                        _vstd[iatt]=1.1*_vstd[iatt];
                    }
                    vBatchDone[iatt]=true;
                }
            }

        }

        // Save the most recent dBase, along with the count of acceptances and rejections, and the seed number, in case the computation needs to be continued.
        _dBaseSaved = dBase; 
        _acceptSaved = accept;
        _rejectSaved = reject;
    }

    _randSaved = *rand;

    delete rand;

    computeAPosterioriForDistribution();
}

void TMetropHasting::continueCalculation(int new_Ns)
{
    try
    {
        // Verify if a calculation has been run before (if not return an error message)
        if (_tdsPar->hasTuple()==false)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                      Form(
                                                          "TMetropHasting::continueCalculation\n\
                The TDS [%s] is empty, nothing to continue. You should start the calculation with 'computeParameters' before using 'continueCalculation'.",
                                                          _tdsPar->GetName()));
        
        // Verify if the number of additional step is superior to 0 (if not return an error message)
        if (new_Ns<1)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                      Form("TMetropHasting::continueCalculation\n\
The number of additional iterations has to be > 1 (here %i). The calculation is not continued.", new_Ns));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    TRandom3* rand = new TRandom3();

    // Read the random generator
    *rand = _randSaved;
    
    // Recover the database, the number of acceptation and rejection, and the seed of the MH computation
    vector<double> dBase = _dBaseSaved;
    vector<int> accept = _acceptSaved;
    vector<int> reject = _rejectSaved;
    
    if(_blog)
        XYZ<<"In continueCalculation"<<endl;

    //////////////////////////////////////////
    // Continuing calculation from JBB version
    //////////////////////////////////////////

    double log_prior_value, loglik_prop;

    // Initialise
    int Niter = _nSam;    
    vector<bool> vOK(_nPar, false);
    vector<bool> vBatchDone(_nPar, false);
    bool PrintDone=false;
    int LenBatch=100;

    // new_Ns + _nSam because _nSam steps has already been computed
    while(Niter < new_Ns + _nSam) 
    {                                	
        for(int iatt=0; iatt<_nPar; iatt++)
        {
            // Print the iteration every _nbDump step
            if ( ( Niter%_nbDump==0 ) && ! PrintDone  )
            {
                cout<<Niter<<" events done"<<endl;
                PrintDone=true;
            }

            // Draw a candidate from the proposal distribution
            _values[iatt]=rand->Gaus( dBase[iatt+1], _vstd[iatt] );

            // If the candidate value is out of bound, we reject it directly, actualise the rejection vector and restart a new step in the while loop
            if ( ( _values[iatt] < _vatt.at(iatt)->getLowerBound() ) || ( _values[iatt] > _vatt.at(iatt)->getUpperBound() ) )
            {
                reject[iatt]+=1;
                _values[iatt]=dBase[iatt+1];
                dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                continue;
            }                   

            // If not, we recompute the proba with new value
            logPriorPdf(log_prior_value);
            _dFunc->eval(&_values[0], &loglik_prop);    
            dBase[2*_nPar+4]=loglik_prop;
            loglik_prop*=-0.5;
            
            // Recompute the ratio	    
            double value_prop=loglik_prop+log_prior_value;
            double log_ratio=value_prop-dBase[2*_nPar+3];
            
            // Draw uniformly a double between 0 and 1
            double log_rand=log(rand->Rndm());
            
            dBase[0]=float(Niter);
            
            // If the ratio is superior to the random double, we keep the candidate, update the dBase vector and fill the TDS 
            if ( log_ratio > log_rand )
            {
                accept[iatt]+=1;
                dBase[2*_nPar+1]=log_prior_value;
                dBase[2*_nPar+2]=loglik_prop;
                dBase[2*_nPar+3]=value_prop;
                dBase[iatt+1]=_values[iatt];
                dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                Niter+=1;
                PrintDone=false;

                _tdsPar->getTuple()->Fill(&dBase[0]);
            }
            // If not, the candidate is rejected and we update the rejection vector 
            else
            {
                reject[iatt]+=1;
                _values[iatt]=dBase[iatt+1];
                dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
            }

            vOK[iatt]=true;
            vBatchDone[iatt]=false;
            
            // If bounds on the acceptation ratio have been defined, 
            if ( (Niter-2-iatt)%LenBatch==0 && ! vBatchDone[iatt] && _lowAccRange != -1 && _higAccRange != -1)
            {
                // And if the current acceptation ratio is too small,
                if( dBase[_nPar+1+iatt] < _lowAccRange )
                {
                    // We reduce the standard deviation of the proposal in order that the candidate are more likely to be accepted
                    _vstd[iatt]=0.9*_vstd[iatt];
                }
                // And if the current acceptation ratio is too big,
                else if( dBase[_nPar+1+iatt] > _higAccRange )
                {
                    // We increase the standard deviation of the proposal in order that we explore more efficiently the parameter space
                    _vstd[iatt]=1.1*_vstd[iatt];
                }
                vBatchDone[iatt]=true;
            }
        }
     
    }

    // Save the last dBase, the number of acceptation and rejection in case the computation is continued
    _nSam = Niter;
    _dBaseSaved = dBase; 
    _acceptSaved = accept;
    _rejectSaved = reject;
    
    _randSaved = *rand;

    delete rand;

    computeAPosterioriForDistribution();
}

// -----------------------------------------------------------------------
//  Save and read
// -----------------------------------------------------------------------

void TMetropHasting::exportMHData(const char *fileName)
{
    // Verify if computeParameters has been run (if not nothing to save and print an error message)
    try
    {
        if(_tdsPar->getNPatterns()==0)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
"TMetropHasting::exportMHData\n\
ComputeParameters has been run. Nothing to export.");
        }
    }
    catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    ///////////////////////////////////////////////////////////////
    // Save the parameters value of TMetropHasting inside the title
    ///////////////////////////////////////////////////////////////

    TString title("");

    // The number of simulations _nSam
    title += Form("Number of simulations: %i;", _nSam);

    // The number of parameters _nPar
    title += Form("Number of parameters: %i;", _nPar);
    
    // The final state of the chain _dBaseSaved
    title += "Final state of the chain: [";
    for (Int_t ii=0;ii<2*_nPar+5;ii++)
        if (ii<2*_nPar+4)
            title += Form("%f,",_dBaseSaved.at(ii));
        else
            title += Form("%f];",_dBaseSaved.at(ii));

    // The standard deviation of the proposal distribution _vstd
    title += "Standard deviation of the proposal distribution: [";
    for (Int_t ii=0;ii<_nPar;ii++)
        if (ii<_nPar-1)
            title += Form("%f,",_vstd[ii]);
        else
            title += Form("%f];",_vstd[ii]);

    // The number of accepted candidates _acceptSaved
    title += "Number of accepted candidates: [";
    for(Int_t ii=0; ii<_nPar; ii++)
        if (ii<_nPar-1)
            title += Form("%i,",_acceptSaved[ii]);
        else
            title += Form("%i];",_acceptSaved[ii]);

    // The number of rejected candidates _rejectSaved
    title += "Number of rejected candidates: [";
    for(Int_t ii=0; ii<_nPar; ii++)
        if (ii<_nPar-1)
            title += Form("%i,",_rejectSaved[ii]);
        else
            title += Form("%i];",_rejectSaved[ii]);
    
    // The value of the parameters _values
    title += "Value of the parameters: [";
    for(Int_t ii=0; ii<_nPar; ii++)
        if (ii<_nPar-1)
            title += Form("%f,",_values[ii]);
        else
            title += Form("%f];",_values[ii]);

    // The burn-in size _burnin
    if(_burnin)
        title += Form("The burn-in size: %i;", _burnin);

    // The lower bound of acceptation ratio _lowAccRange
    if(_lowAccRange)
        title += Form("Lower bound of acceptation ratio: %f;", _lowAccRange);

    // The upper bound of acceptation ratio _higAccRange
    if(_higAccRange)
        title += Form("Upper bound of acceptation ratio: %f;", _higAccRange);

    // The lag value _lag
    if(_lag)
        title += Form("The lag value: %i;", _lag);
    
    // The frequency to which the algo dump a line _nbDump
    if(_nbDump)
        title += Form("Frequency to which the algo dump a line: %i;", _nbDump);

    //////////////////////
    // Save the attributes
    //////////////////////

    TString varNames  = "";

    // Extract the name of the attributes separeting them with ":"
    for(Int_t i = 0; i < 2*_nPar+4; i++)
    {
        if (i<2*_nPar+3)
            varNames += Form("%s:",_tdsPar->getAttribute(i)->GetName());
        else
            varNames += Form("%s",_tdsPar->getAttribute(i)->GetName());
    }
    
    TObjArray *varNamesList = varNames.Tokenize(":"); 
    
    // Create the saved TDS
    URANIE::DataServer::TDataServer *tdssaveMH = new URANIE::DataServer::TDataServer("tdssaveMH", title.Data());

    // Add the attributes in the saved TDS
    for(Int_t i = 0; i < varNamesList->GetEntries(); i++)
    {
        URANIE::DataServer::TAttribute *att = new URANIE::DataServer::TAttribute(varNamesList->At(i)->GetName());
        att->delShare();
        tdssaveMH->addAttribute(att);
    }

    ///////////////////////////
    // Create tuple and fill it
    ///////////////////////////

    // Create tuple
    tdssaveMH->createTuple();
    URANIE::DataServer::TDSNtupleD *t = tdssaveMH->getTuple() ;

    // Fill the tds
    
    Int_t nbBr = t->GetNbranches();
    Double_t *line = new Double_t[nbBr];

    for(Int_t i = 0; i < _nSam; i++) 
    {
        // Iterator
        line[0] = Double_t(i+1);
        for(Int_t j = 0; j < nbBr-1; j++)
        {
            // Extract the values of the saved TDS
            line[j+1] = _tdsPar->getValue(varNamesList->At(j)->GetName(),i);
        }
        // Fill line by line the TDS
        t->Fill(line);
    }

    // Export the data
    tdssaveMH->exportData(fileName);

    // Free the allocated memory
    delete [] line;
    line = NULL;

    delete varNamesList;  
    varNamesList = NULL;
        
    delete tdssaveMH;   
    tdssaveMH = NULL;

    // Save the random generator in a ROOT file
    std::string randomFileNameStr = std::string(fileName) + "_random.root";
    const char *randomFileName = randomFileNameStr.c_str();

    TFile file(randomFileName, "RECREATE");
    _randSaved.Write("myRandom");
    file.Close();
}

void TMetropHasting::readMHData(const char *fileName)
{
    
    // Verify if the file exists
    try
    {
        // If yes, read it
        std::ifstream file(fileName);
        if(!file.good())
           throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                        Form("TMetropHasting::readMHData\n\
The file named [%s] does not exist.", fileName));
    }
    catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    vector<double> dBase_Saved;
    vector<double> vstd_Saved;
    vector<int> accept_Saved;
    vector<int> reject_Saved;
    vector<double> values_Saved;

    // Read the TDS saved
    URANIE::DataServer::TDataServer *tdssaveMH = new URANIE::DataServer::TDataServer();;
    tdssaveMH->fileDataRead(fileName);

    /////////////////////////////////////////////////////////////
    // Read the title to extract the information contained inside
    /////////////////////////////////////////////////////////////

    TString title = tdssaveMH->GetTitle();

    // The information are separated by ";"
    TObjArray *titleList = title.Tokenize(";"); 

    for(Int_t i = 0; i < titleList->GetEntries(); i++)
    {
        TString element_title = titleList->At(i)->GetName();

        // The name of the information and its value are separated by ":"
        TObjArray *element_title_List = element_title.Tokenize(":"); 
        
        std::string var_title = element_title_List->At(0)->GetName();

        // Now we can recognize the name of the information and update the hyperparameter value
        try
        {
            // If the information is the number of simulations (_nSam)
            if (var_title.compare("Number of simulations")==0)
            {
                std::string into_string = element_title_List->At(1)->GetName();
                
                // Convert into int
                _nSam  = std::stoi(into_string);
            }

            // If the information is the number of parameters (_nPar)
            else if (var_title.compare("Number of parameters")==0)
            {
                std::string into_string = element_title_List->At(1)->GetName();
                
                // Convert into int
                _nPar = std::stoi(into_string);
            }

            // If the information is the final state of the chain (_dBaseSaved)
            else if (var_title.compare("Final state of the chain")==0)
            {
                std::string Final_state_chain = element_title_List->At(1)->GetName();
                
                // Remove the space and the brackets at the beggining and the end
                TString Final_state_chain_extracted = Final_state_chain.substr(2,Final_state_chain.length() - 3);
                
                // The values are separated by ","
                TObjArray *Final_state_chain_list = Final_state_chain_extracted.Tokenize(","); 
                    
                for(Int_t ii = 0; ii < Final_state_chain_list->GetEntries(); ii++)
                {           
                    std::string into_string = Final_state_chain_list->At(ii)->GetName();
                    // Add the value to the vector dBase_Saved after converting it into double
                    dBase_Saved.push_back(std::stod(into_string));
                }
                
                delete Final_state_chain_list;
                Final_state_chain_list = NULL;
            }

            // If the information is the standard deviation of the proposal distribution (_vstd)
            else if (var_title.compare("Standard deviation of the proposal distribution")==0)
            {
                std::string Standard_deviation = element_title_List->At(1)->GetName();

                // Remove the space and the brackets at the beggining and the end
                TString Standard_deviation_extracted = Standard_deviation.substr(2,Standard_deviation.length() - 3);
                
                // The values are separated by ","
                TObjArray *Standard_deviation_list = Standard_deviation_extracted.Tokenize(","); 
                    
                for(Int_t ii = 0; ii < Standard_deviation_list->GetEntries(); ii++)
                {
                    std::string into_string = Standard_deviation_list->At(ii)->GetName();
                    // Add the value to the vector vstd_Saved after converting it into double
                    vstd_Saved.push_back(std::stod(into_string));
                }

                delete Standard_deviation_list;
                Standard_deviation_list = NULL;
            }

            // If the information is the number of accepted candidates (_acceptSaved)
            else if (var_title.compare("Number of accepted candidates")==0)
            {
                std::string accepted_candidates = element_title_List->At(1)->GetName();

                // Remove the space and the brackets at the beggining and the end
                TString accepted_candidates_extracted = accepted_candidates.substr(2,accepted_candidates.length() - 3);
                
                // The values are separated by ","
                TObjArray *accepted_candidates_list = accepted_candidates_extracted.Tokenize(","); 
                    
                for(Int_t ii = 0; ii < accepted_candidates_list->GetEntries(); ii++)
                {
                    std::string into_string = accepted_candidates_list->At(ii)->GetName();
                    // Add the value to the vector accept_Saved after converting it into int
                    accept_Saved.push_back(std::stoi(into_string));
                }

                delete accepted_candidates_list;
                accepted_candidates_list = NULL;
            }

            // If the information is the number of rejected candidates (_rejectSaved)
            else if (var_title.compare("Number of rejected candidates")==0)
            {
                std::string rejected_candidates = element_title_List->At(1)->GetName();

                // Remove the space and the brackets at the beggining and the end
                TString rejected_candidates_extracted = rejected_candidates.substr(2,rejected_candidates.length() - 3);
                
                // The values are separated by ","
                TObjArray *rejected_candidates_list = rejected_candidates_extracted.Tokenize(","); 
                    
                for(Int_t ii = 0; ii < rejected_candidates_list->GetEntries(); ii++)
                {
                    std::string into_string = rejected_candidates_list->At(ii)->GetName();
                    // Add the value to the vector reject_Saved after converting it into int
                    reject_Saved.push_back(std::stoi(into_string));
                }

                delete rejected_candidates_list;
                rejected_candidates_list = NULL;
            }

            // If the information is the value of the parameters (_values)
            else if (var_title.compare("Value of the parameters")==0)
            {
                std::string value_param = element_title_List->At(1)->GetName();

                // Remove the space and the brackets at the beggining and the end
                TString value_param_extracted = value_param.substr(2,value_param.length() - 3);
                
                // The values are separated by ","
                TObjArray *value_param_list = value_param_extracted.Tokenize(","); 

                for(Int_t ii = 0; ii < value_param_list->GetEntries(); ii++)
                {
                    std::string into_string = value_param_list->At(ii)->GetName();
                    // Add the value to the vector values_Saved after converting it into double
                    values_Saved.push_back(std::stod(into_string));
                }

                delete value_param_list;
                value_param_list = NULL;
            }

            // If the information is the burn-in size (_burnin)
            else if (var_title.compare("The burn-in size")==0)
            {
                std::string into_string = element_title_List->At(1)->GetName();

                // Convert into int
                _burnin = std::stoi(into_string);
            }

            // If the information is the lower bound of acceptation ratio (_lowAccRange)
            else if (var_title.compare("Lower bound of acceptation ratio")==0)
            {
                std::string into_string = element_title_List->At(1)->GetName();

                // Convert into double
                _lowAccRange = std::stod(into_string);
            }

            // If the information is the upper bound of acceptation ratio (_higAccRange)
            else if (var_title.compare("Upper bound of acceptation ratio")==0)
            {
                std::string into_string = element_title_List->At(1)->GetName();

                // Convert into double
                _higAccRange = std::stod(into_string);
            }

            // If the information is the lag value (_lag)
            else if (var_title.compare("The lag value")==0)
            {
                std::string into_string = element_title_List->At(1)->GetName();

                // Convert into int
                _lag = std::stoi(into_string);
            }

            // If the information is the frequency to which the algo dump a line (_nbDump)
            else if (var_title.compare("Frequency to which the algo dump a line")==0)
            {
                std::string into_string = element_title_List->At(1)->GetName();

                // Convert into int
                _nbDump = std::stoi(into_string);
            }
            
            // If the information is not recognized
            else
            {
                throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                        Form("TMetropHasting::readMHData\n\
The keyword [%s] is not recognized.", var_title.c_str()));
            }
            
            delete element_title_List;
            element_title_List = NULL;
        }
        catch(URANIE::Exceptions::UInfoExceptions& ue)
        {
            ue.printMessage();
        }
    }

    /////////////////////////////////////////////////////////////
    // Construct the attributes
    /////////////////////////////////////////////////////////////
    
    // Extract the number of parameter of the current TDS
    Int_t nPar = _tdsPar->getNAttributes();

    // Extract the number of parameter of the saved TDS
    Int_t nPar_saved = tdssaveMH->getNAttributes();

    TString varNames  = "";
    TString varNames_saved  = "";

    for(Int_t i = 0; i < nPar_saved; i++)
    {
        if (i<nPar_saved-1)
        {
            // We extract the name of the saved attributes separating them with ":"
            varNames_saved += tdssaveMH->getAttribute(i)->GetName();
            varNames_saved += ":";
        }
        else
        {
            // We do not add ":" after the last one
            varNames_saved += tdssaveMH->getAttribute(i)->GetName();
        }
    }
    for(Int_t i = 0; i < nPar; i++)
    {
        if (i<nPar-1)
        {
            // We extract the name of the current attributes separating them with ":"
            varNames += _tdsPar->getAttribute(i)->GetName();
            varNames += ":";
        }
        else
        {
            // We do not add ":" after the last one
            varNames += _tdsPar->getAttribute(i)->GetName();
        }
    }

    TObjArray *varNamesList = varNames.Tokenize(":"); 
    TObjArray *varNamesList_saved = varNames_saved.Tokenize(":"); 

    for(Int_t i = 0; i < nPar; i++)
    {
        
        // If one of the current attributes is not inside the saved TDS, we delete this attribute (and print an informative message)
        try
        {
            if(tdssaveMH->isAttribute(varNamesList->At(i)->GetName())==0)
            {    
                throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                Form("TMetropHasting::readMHData\n\
Attribute [%s] has been deleted as it does not exist in the saved dataserver",varNamesList->At(i)->GetName()));
            }
        }
        catch(URANIE::Exceptions::UInfoExceptions& ue)
        {
            _tdsPar->delAttribute(varNamesList->At(i)->GetName());
            ue.printMessage();
        }
    }

    for(Int_t i = 0; i < nPar_saved; i++)
    {
    
        // If one of the saved attributes is not inside the current TDS, we add a new attribute to the current TDS (and print an informative message)
        try
        {
            if(_tdsPar->isAttribute(varNamesList_saved->At(i)->GetName())==0)
            {    
                throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                Form("TMetropHasting::readMHData\n\
Attribute [%s] has been added as it exists in the saved dataserver",varNamesList_saved->At(i)->GetName()));
            }
        }
        catch(URANIE::Exceptions::UInfoExceptions& ue)
        {
            _tdsPar->addAttribute(varNamesList_saved->At(i)->GetName());
            ue.printMessage();
        }
    }

    /////////////////////////////////////////////////////////////
    // Prepare the Tuple and fill it
    /////////////////////////////////////////////////////////////

    // Remove the actual Tuple and create a new one empty
    _tdsPar->deleteTuple();
    _tdsPar->createTuple();

    URANIE::DataServer::TDSNtupleD *t = _tdsPar->getTuple() ;
    
    Int_t nbBr = t->GetNbranches();
    Double_t *line = new Double_t[nbBr];

    for(Int_t i = 0; i < _nSam; i++) 
    {
        // Iterator
        line[0] = Double_t(i+1);
        for(Int_t j = 0; j < nbBr-1; j++)
        {            
            // Extract the values of the saved TDS
            line[j+1] = tdssaveMH->getValue(varNamesList_saved->At(j)->GetName(),i);
        }
        // Fill line by line the TDS
        t->Fill(line);
    }

    // Free the allocated memory
    delete [] line;
    line = NULL;
    
    delete varNamesList;  
    varNamesList = NULL;

    delete varNamesList_saved;  
    varNamesList_saved = NULL;
    
    delete titleList;   
    titleList = NULL;

    delete tdssaveMH;   
    tdssaveMH = NULL;

    // Save the variables into the global variables used by the continueCalculation function
    _dBaseSaved = dBase_Saved;
    _vstd = vstd_Saved;
    _acceptSaved = accept_Saved;
    _rejectSaved = reject_Saved;
    _values = values_Saved;

    // Read the ROOT file containing the random generator
    std::string randomFileNameStr = std::string(fileName) + "_random.root";
    const char *randomFileName = randomFileNameStr.c_str();

    TFile file(randomFileName, "READ");
    TRandom3* rand = nullptr;

    file.GetObject("myRandom", rand);
    file.Close();

    _randSaved = *rand;

    delete rand;
}

// -----------------------------------------------------------------------
//  Visualisation and diagnostic
// -----------------------------------------------------------------------

void TMetropHasting::drawTrace(TString sTitre, const char *variable, const char *select, Option_t *option)
{
    TString soption = option;
    TString cleanoption = option;
    soption.ToLower();

    if ( _blog )
        cout << " ** TMetropHasting::drawTrace for method[" << _sMethodName << "] option[" << option << "]" << endl;
    
    if (_tdsPar->getTuple()->GetEntries() == 0) {
        cout << "TMatropHAsting::drawTrace Error, the tdsPar is empty" << endl;
        return;
    }

    bool newcan=false;
    if (!soption.Contains("nonewcanvas"))
	newcan=true;
    else
        cleanoption.ReplaceAll("nonewcanvas","");

    checkCanvasCreation(newcan);

    string svariable=variable;
    vector<string> variables;
    if( svariable=="*" )
        for(unsigned int ivar=0; ivar<_vatt.size(); ivar++)
            variables.push_back( string(_vatt.at(ivar)->GetName() ) );
    else
        URANIE::DataServer::UEntry::StringParser(svariable, ":", variables);

    if (soption.Contains("vertical"))
    {
        _canvas->Divide( variables.size(), 1);
        cleanoption.ReplaceAll("vertical","");
    }
    else
        _canvas->Divide(1, variables.size());    

    URANIE::DataServer::TAttribute *att=NULL;
    TH1 *h=NULL;
    string sselect=select;
    string cut = getDefaultCut() + ((sselect!="" && sselect!="1>0") ? ((getDefaultCut()!="")?" && ":"")+sselect : "");

    for(unsigned int ivar=0; ivar<variables.size(); ivar++)
    {
        _canvas->cd(ivar+1);

        _tdsPar->getTuple()->Draw(Form("%s:%s",variables.at(ivar).c_str(), _tdsPar->getIteratorName()), cut.c_str(), (TString("l")+cleanoption).Data());

        att=(URANIE::DataServer::TAttribute*)_listOfParameters->FindObject(variables.at(ivar).c_str());

        h = (TH1*)gPad->GetPrimitive("htemp");
        h->SetName( (variables.at(ivar)+"_trace").c_str() );
        h->SetTitle( ((sTitre=="") ? (variables.at(ivar)+"_trace").c_str() : sTitre.Data() ) );
        h->GetXaxis()->SetTitle("Nb_{Calc}");
        h->GetYaxis()->SetTitle( ((att!=NULL)?att->getTitle():"") );
    }
}

void TMetropHasting::drawAcceptationRatio(TString sTitre, const char *variable, const char *select, Option_t *option)
{
    TString soption = option;
    TString cleanoption = option;
    soption.ToLower();

    if ( _blog )
        cout << " ** TMetropHasting::drawAcceptationRatio for method[" << _sMethodName << "] option[" << option << "]" << endl;
    
    if (_tdsPar->getTuple()->GetEntries() == 0) 
    {
        cout << "TMatropHAsting::drawAcceptationRatio Error, the tdsPar is empty" << endl;
        return;
    }

    bool newcan=false;
    if (!soption.Contains("nonewcanvas"))
	    newcan=true;
    else
        cleanoption.ReplaceAll("nonewcanvas","");

    checkCanvasCreation(newcan);

    string svariable=variable;
    vector<string> variables;
    if( svariable=="*" )
        for(unsigned int ivar=0; ivar<_vatt.size(); ivar++)
            variables.push_back( string(_vatt.at(ivar)->GetName() ) );
    else
        URANIE::DataServer::UEntry::StringParser(svariable, ":", variables);

    if (soption.Contains("vertical"))
    {
        _canvas->Divide( variables.size(), 1);
        cleanoption.ReplaceAll("vertical","");
    }
    else
        _canvas->Divide(1, variables.size());    

    URANIE::DataServer::TAttribute *att=NULL;
    TH1 *h=NULL;
    string sselect=select;
    string cut = getDefaultCut() + ((sselect!="" && sselect!="1>0") ? ((getDefaultCut()!="")?" && ":"")+sselect : "");
    for(unsigned int ivar=0; ivar<variables.size(); ivar++)
    {
        _canvas->cd(ivar+1);

        _tdsPar->getTuple()->Draw(Form("acceptation_ratio_%s:%s",variables.at(ivar).c_str(), _tdsPar->getIteratorName()), cut.c_str(), (TString("l")+cleanoption).Data());

        att=(URANIE::DataServer::TAttribute*)_listOfParameters->FindObject(variables.at(ivar).c_str());

        h = (TH1*)gPad->GetPrimitive("htemp");
        h->SetName( (variables.at(ivar)+"_trace").c_str() );
        h->SetTitle( ((sTitre=="") ? (variables.at(ivar)+"_trace").c_str() : sTitre.Data() ) );
        h->GetXaxis()->SetTitle("Nb_{Calc}");
        h->GetYaxis()->SetTitle( ((att!=NULL)?att->getTitle():"") );
    }
}

void TMetropHasting::drawParameters(TString sTitre, const char *variable, const char *select, Option_t *option)
{
    TString soption = option;
    TString cleanoption = option;
    soption.ToLower();

    if ( _blog )
        cout << " ** TMetropHasting::drawParameters for method[" << _sMethodName << "] option[" << option << "]" << endl;
    
    if (_tdsPar->getTuple()->GetEntries() == 0) 
    {
        cout << "TMetropHasting::drawParameters Error, the tdsPar is empty" << endl;
        return;
    }

    bool newcan=false;
    if (!soption.Contains("nonewcanvas"))
	    newcan=true;
    else
        cleanoption.ReplaceAll("nonewcanvas","");
    
    checkCanvasCreation(newcan);

    string svariable=variable;
    vector<string> variables;
    if( svariable=="*" )
        for(unsigned int ivar=0; ivar<_vatt.size(); ivar++)
            variables.push_back( string(_vatt.at(ivar)->GetName() ) );
    else
        URANIE::DataServer::UEntry::StringParser(svariable, ":", variables);

    if( !soption.Contains("same") )
    {
        if (soption.Contains("vertical"))
        {
            _canvas->Divide(1, variables.size());
            cleanoption.ReplaceAll("vertical","");
        }
        else
            _canvas->Divide(variables.size(), 1);
    }
    
    URANIE::DataServer::TAttribute *att=NULL;
    TH1 *h=NULL;
    string sselect=select;
    string cut = getDefaultCut() + ((sselect!="" && sselect!="1>0") ? ((getDefaultCut()!="")?" && ":"")+sselect : "");
    for(unsigned int ivar=0; ivar<variables.size(); ivar++)
    {
        _canvas->cd(ivar+1);

        att=(URANIE::DataServer::TAttribute*)_listOfParameters->FindObject(variables.at(ivar).c_str());

        _tdsPar->getTuple()->Draw( variables.at(ivar).c_str(), cut.c_str(), cleanoption.Data() );
        h = (TH1*)gPad->GetPrimitive("htemp");
        h->SetName( ("h"+variables.at(ivar)+"_aposteriori").c_str() );
        h->SetTitle( ((sTitre=="") ? ("h"+variables.at(ivar)+"_aposteriori").c_str() : sTitre.Data() ) );
        h->GetXaxis()->SetTitle(att->getTitle());
    }
}
	
void TMetropHasting::getAutoCorrelation(vector<int> l, vector<double> *out, int cut)
{
    out->clear();
    out->resize(l.size()*_vatt.size());
    double *data;
    int np=0, arrsize=0;
    for(unsigned int ip=0; ip<_vatt.size(); ip++)
    {
        np=_tdsPar->getTuple()->Draw(_vatt.at(ip)->GetName(), Form("(%s > %d)",_tdsPar->getIteratorName(),cut),"goff");
        data=_tdsPar->getTuple()->GetV1();
        double mY=TMath::Mean(np, data), sdY=TMath::RMS(np, data);

        for(unsigned int il=0; il<l.size(); il++)
        {
                arrsize=np-l.at(il);
            int Nech=np-l.size();
            double cross=0.;
            for(int i=0; i<arrsize; i++)
            cross+= (data[i]-mY)/sdY * (data[i+l.at(il)]-mY)/sdY;

            cross*=1./(Nech-1);
            out->at(l.size()*ip+il)=cross;
        }
    }
}

// -----------------------------------------------------------------------
//  internal methodprintLog
// -----------------------------------------------------------------------

void TMetropHasting::logPriorPdf(double &ret)
{
    ret=0;
    for(unsigned int iatt=0; iatt<_vatt.size(); iatt++)
	ret+=log(_vatt.at(iatt)->getPDF(_values.at(iatt)));
}


// -----------------------------------------------------------------------
//  printLog
// -----------------------------------------------------------------------

void TMetropHasting::parseOption(Option_t *option)
{

    TCalibration::parseOption(option);
    
    TString soption=option;
    soption.ToLower();
    
    if( soption.Contains("clean") )
	    _bcleaningAtt=true;

    if( soption.Contains("gd") )
	    _mbGDVersion=true;


}

void TMetropHasting::printLog(Option_t *option)
{
    cout << "TMetropHasting::printLog[" << option << "] Name[" << GetName() << "] Title[" << GetTitle() << "] " << endl;
    cout << "** _nSam   [" << _nSam << "]" << endl;
    cout << "** _nPar   [" << _nPar << "]" << endl;
    cout << "** _sMethodName   [" << _sMethodName << "]" << endl;

    cout << "End Of TMetropHasting::printLog" << endl;
}

}
}

