//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2025 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////

/*!
 \file   TFCNCode.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date     
 \brief Implementation of the class URANIE::Optimizer::TFCNCode

 \par Modifications
 \li M. Salvoldi (Areva/OCC) 2011/09/28
 - add a IFDEF WIN32 to replace rm -fr by del /Q on Windows platform
 
 */

// URANIE
#include "TFCNCode.h"
#include "UExceptions.h"
#include "TSystem.h"
#include "forkcode.h"

ClassImp (URANIE::Optimizer::TFCNCode)

namespace URANIE
{
namespace Optimizer
{
void TFCNCode::init()
{
    try
    {
        if (_tds != NULL)
        {
            if (_tds->getTuple() == NULL)
            {
                _tds->createTuple();
            }
        }
        _ninput = _tds->getNInputAttributes();
        _noutput = _tds->getNOutputAttributes();

        if (_noutput == 0)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    Form(
                            "TFCNCode::init [%s] ERROR\n the Number of output is not equal to 1",
                            _tds->GetName(), _noutput));

        Int_t ncpt = 1;
        Int_t nvar = _tds->getNAttributes();
        for (Int_t ivar = 0; ivar < nvar; ivar++)
        {
            if (_tds->getAttribute(ivar)->isOutput())
            {
                _soutput = _tds->getAttribute(ivar)->GetName();
                if (_soutput == _sCost)
                    _nCost = ncpt;
                ncpt++;
            }
        }
        cout << " ** TFCNCode::init _sCost[" << _sCost << "] _nCost[ " << _nCost
                << "]" << endl;

        // The size is increased by one for the iterator
        _dinputValue = new double[_ninput + _noutput + 1];
        _doutputValue = new double[_noutput + 1];

        clean();

        _code->init();

        // Go to the working directory
        gSystem->cd(_code->getWorkingDirectory());
        // Make the Results directory for distribution
        TString sresultsdirectory = TString("UranieResults");
        if (!gSystem->OpenDirectory(sresultsdirectory))
            if (gSystem->mkdir(sresultsdirectory, kTRUE) == -1)
                throw URANIE::Exceptions::UExceptions(__FILE__,
                __LINE__,
                        Form(
                                "TLauncher::init ERROR : the subdirectory \"%s\" can not be created in the workindirectory \"%s\"",
                                sresultsdirectory.Data(),
                                _code->getWorkingDirectory().Data()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        exit(-1);
    }

    catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

double TFCNCode::operator()(const vector<double> & par)
{
    Int_t niter = _tds->getTuple()->GetEntries();
    double fval = 0.1;
    if (_code->getLog())
        cout << "iter[" << niter << "]" << endl;
    if (_code != NULL)
    {
        _dinputValue[0] = 1.0 + niter;
        // Remark: If we want to save all computations, we need a constant Bool_t _bSave,
        // and call the the method _code->init() at each new computation.
        URANIE::Launcher::forkcode fjob;
        fjob.pattern = niter;
        fjob.command = "";
        _code->initRun(&fjob);
        // Pretraitment of the data
        for (int i = 0; i < _ninput; i++)
        {
            _dinputValue[1 + i] = par[i];
            if (_code->getLog())
                cout << "**** input i[" << i << "] val["
                        << _dinputValue[1 + i] << "]" << endl;
        }
        int ind = 1;
        _code->preTraitment(&fjob, _dinputValue);
        // Lancement du code
        _code->run(&fjob);
        // Post-traitment of the data
        ind = 0;
        _code->postTraitment(&fjob, ind, _doutputValue);
        if (_code->getLog())
        {
            cout << endl << "*************************************" << endl
                    << " *** TFCNCode operator () _nCost[" << _nCost << "]"
                    << endl;
            for (Int_t ii = 0; ii <= _nCost; ii++)
                cout << " ** ii(" << ii << "/" << _nCost << ") val("
                        << _doutputValue[ii] << ")" << endl;
            cout << " *** End Of TFCNCode operator ()" << endl
                    << "*************************************" << endl;
        }
        fval = _doutputValue[_nCost];
        if (_code->getLog())
            cout << " *** output _nCost[" << _nCost << "] val[" << fval
                    << "]" << endl;

        //Fill the output values in the input vector
        // +1 because the first value is the index
        for (int i = 1; i <= _noutput; i++)
            _dinputValue[_ninput + i] = _doutputValue[i];
        _tds->getTuple()->Fill(_dinputValue);

        // Visualisation
        if (niter % 10)
        {
            //	    _ntd->Draw("val:x", "", "lp");
            _tds->draw(
                    Form("%s:%s", _sCost.Data(), _tds->getIteratorName()),
                    "", "lp");
            gPad->Modified();
            gPad->Update();
        }
    }

    return fval;
}

void TFCNCode::clean()
{
    TString sworkindirectory = _code->getWorkingDirectory();

    // Place in the good directory
    if (sworkindirectory.Length() != 0)
        gSystem->cd(sworkindirectory);

    // Clean the subdirectory "UranieLauncher_*"
#ifdef WIN32
    gSystem->Exec("rd /s /q UranieLauncher_*");
#else
    gSystem->Exec("rm -fr UranieLauncher_*");
#endif
}
} // Fin du namespace Optimizer
} // Fin du namespace URANIE

