////////////////////////////////////////////////////////////// 
// Copyright (C) 2013-2025 CEA/DES 
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
\file   TCramerVonMises.cxx
\author Rudy Chocat <rudy.chocat@cea.fr>
\date   2024-09-09 11:19:38

\brief Implement the URANIE::Sensitivity::TCramerVonMises class

*/

#include "TAttribute.h"

#include "mat3D_alloc.h"

#ifndef ROOT_TVectorD
#include "TVectorD.h"
#endif
#include "TMLPAnalyzer.h"

#include "TDirectory.h"
#include "TMath.h"
#include "TLeaf.h"
#include "TSystem.h"
#include "TROOT.h"

// Uranie
#include "TSobol.h"
#include "TCramerVonMises.h"

#include <memory>

using std::cout;
using std::endl;

ClassImp (URANIE::Sensitivity::TCramerVonMises)

namespace URANIE
{
namespace Sensitivity
{


TCramerVonMises::TCramerVonMises() : TSobol()
{

}

TCramerVonMises::TCramerVonMises(URANIE::DataServer::TDataServer *tds, 
        void (*fcn)(Double_t*,Double_t*), const char *varexpinput, 
        const char *varexpoutput, Int_t ns, Option_t * option) :
    TSobol(tds, fcn, varexpinput, varexpoutput, ns,"cvm")
{
}

// -----------------------------------------------------------------------
//  Constructors with a function
// -----------------------------------------------------------------------
TCramerVonMises::TCramerVonMises(URANIE::DataServer::TDataServer *tds, const char *fcn, Int_t ns, const char *varexpinput,
	       const char *varexpoutput, Option_t* option) :
    TSobol(tds, fcn, ns, varexpinput, varexpoutput, "cvm")
{
}

// -----------------------------------------------------------------------
//  Constructors with a TDataServer filled
// -----------------------------------------------------------------------
TCramerVonMises::TCramerVonMises(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
        const char *varexpoutput, Option_t* option) :
        TSobol(tds, varexpinput, varexpoutput, "cvm")
{
}
// -----------------------------------------------------------------------
//  Constructors with a TCode
// -----------------------------------------------------------------------
TCramerVonMises::TCramerVonMises(URANIE::DataServer::TDataServer *tds,
        URANIE::Launcher::TCode *code, Int_t ns, Option_t * option) :
        TSobol(tds, code, ns, "cvm")
{
}

// -----------------------------------------------------------------------
//  Constructors with a TRun
// -----------------------------------------------------------------------
TCramerVonMises::TCramerVonMises(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TRun *run, Int_t ns, Option_t * option) :
    TSobol(tds, run, ns, "cvm")
{
}

TCramerVonMises::~TCramerVonMises()
{
    // if (_blog)
    //     cout << " ** Destroy TSobol[" << GetName() << "]" << endl;
    // if( _bWithData && _withdatantp != NULL)
    // {
    //     delete _withdatantp;
    //     _withdatantp=NULL;
    // }
    // if( _preloadedntp != NULL )
    // {
    //     delete _preloadedntp;
    //     _preloadedntp=NULL;
    // }
}


///////////////////////////////////////////////////////////
//  Generate the Sample 
///////////////////////////////////////////////////////////
void TCramerVonMises::generateSample(Option_t * option)
{
    prepareSizePerIndex(option);
    TSobol::generateSample(option);
    auto tdsTemp = std::make_unique<URANIE::DataServer::TDataServer>();
    for(int i = 0; i < _nX ;i++)
        tdsTemp->addAttribute( (URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(i)));
    
    switch (_nSamplerMethod)
    {
        case kSRS:
        {
            URANIE::Sampler::TSampling *_fsampler = new URANIE::Sampler::TSampling(
                    tdsTemp.get(), "srs", _nSimPerIndex);
            _fsampler->setSeed(_nSeed);
            _fsampler->generateSample();
            delete _fsampler;
            _fsampler = NULL;
            break;
        }
        case kLHS:
        {
            URANIE::Sampler::TSampling *_fsampler = new URANIE::Sampler::TSampling(
                    tdsTemp.get(), "lhs", _nSimPerIndex);
            _fsampler->setSeed(_nSeed);
            _fsampler->generateSample();
            delete _fsampler;
            _fsampler = NULL;
            break;
        }
        case kqMCSobol:
        {
            URANIE::Sampler::TQMC *_fsampler = new URANIE::Sampler::TQMC(tdsTemp.get(),
                    "sobol", _nSimPerIndex);
            _fsampler->generateSample();
            delete _fsampler;
            _fsampler = NULL;
            break;
        }
        case kqMCHalton:
        {
            URANIE::Sampler::TQMC *_fsampler = new URANIE::Sampler::TQMC(tdsTemp.get(),
                    "halton", _nSimPerIndex);
            _fsampler->generateSample();
            delete _fsampler;
            _fsampler = NULL;
            break;
    }
    default:
        cout << " ** Default case of TSobol::initSample" << endl;
    }
    Double_t *valInpO = new Double_t[_nX + 2];

    
    valInpO[1] = -10.0;
    
    if( !_bWithData )
    {
        for (Int_t ipat = 0; ipat < _nSimPerIndex; ipat++)
        {
            valInpO[0] = (Double_t)(_tds->getTuple()->GetEntries()) + (Double_t)ipat;
            for (Int_t ix = 0; ix < _nX; ix++)
            {
                valInpO[2 + ix] = tdsTemp->getValue(_listOfInputAttributes->At(ix)->GetName(), ipat);
            }

            _tds->getTuple()->Fill(valInpO);
        }
    }

    delete[] valInpO;
    valInpO = NULL;


    

}

///////////////////////////////////////////////////////////
///  evaluateIndexes
///////////////////////////////////////////////////////////
void TCramerVonMises::evaluateIndexes(Option_t * option)
{
  _valO.ResizeTo(_nSimPerIndex, _nbOut);
  if (_blog)
  {
    cout << "***********************************" << endl;
    cout << " ** TCramerVonMises::evaluateIndexes[" << option << "]" << endl;
  }
  bool _bq = _bquiet;
  if(!_bquiet)
  {
      _bquiet=true;
  }
  TSobol::evaluateIndexes(option);
  _bquiet=_bq;
  setMethodName("CramerVonMises");
  URANIE::DataServer::TAttribute *attx;
  URANIE::DataServer::TAttribute *atty;

  sprintf(_sMethod, "%s", _sMethodName.Data());
  TString soutputy, sSelectComputeStat;
  atty = (URANIE::DataServer::TAttribute *) (_listOfOutputAttributes->At(_iy));
  soutputy = atty->GetName();
  double *Values=NULL, *IterV=NULL;
  vector<vector<double> > M_N_Mi; M_N_Mi.resize(3);
  vector<double> iterVals;
  double mi, ni;
  /// End of FGA::remarks 2010-06-25 10:40:57
  //////////////////////
  std::stringstream sstr;   
  //Make the loop for components here to avoid recompute statistic so many times
  for( _iely=0; _iely<_nElY; _iely++)
  {

    if (_nElY!=1)
    {
      sstr.str("");
      sstr << atty->GetName() <<  "[" << _iely << "]";
      sprintf(_sOutputAttribute, "%s", sstr.str().c_str() );

      if (_mAttributeElements.find( string(atty->GetName()) )!=_mAttributeElements.end())
      {
        vector<int> sublist =  _mAttributeElements.find( string(atty->GetName()) )->second;
        if( find( sublist.begin(), sublist.end(), _iely) == sublist.end() )
        {
          _iOut++;
          continue;
        }
      }
    }

      
  if (_blog)
      cout << " ** Case of Output atty [" << _sOutputAttribute << "]  nSimPerIndex " << _nSimPerIndex << endl;  

    TString iter_name = _tds->getIteratorName();
    
    _tds->getTuple()->Draw( iter_name+":"+_sOutputAttribute, Form("%s==-1", _sIteratorName.Data()), "goff" );
    Values = _tds->getTuple()->GetV2(); M_N_Mi[0].assign(Values, Values + _nSimPerIndex); 
    IterV = _tds->getTuple()->GetV1(); iterVals.assign(IterV, IterV + _nSimPerIndex);
    //   TSobol::our_parallel_sort(&M_N_Mi[0], &iterVals);

    _tds->getTuple()->Draw( iter_name+":"+_sOutputAttribute, Form("%s==0", _sIteratorName.Data()), "goff" );
    Values = _tds->getTuple()->GetV2(); M_N_Mi[1].assign(Values, Values + _nSimPerIndex); 
    IterV = _tds->getTuple()->GetV1(); iterVals.assign(IterV, IterV + _nSimPerIndex);
    //   TSobol::our_parallel_sort(&M_N_Mi[1], &iterVals);

    _tds->getTuple()->Draw( iter_name+":"+_sOutputAttribute, Form("%s==-10", _sIteratorName.Data()), "goff" );
    Values = _tds->getTuple()->GetV2(); M_N_Mi[2].assign(Values, Values + _nSimPerIndex); 
    IterV = _tds->getTuple()->GetV1(); iterVals.assign(IterV, IterV + _nSimPerIndex);
    //   TSobol::our_parallel_sort(&M_N_Mi[2], &iterVals);
    /// Store the two matrix M and N
    for (Int_t ipat = 0; ipat < _nSimPerIndex; ipat++)
    {
    /// Matrix M
    _valO(ipat, _iOut-1) = M_N_Mi[2][ipat]; //_tds->getValue(soutputy, _telm->GetEntry(ipat), _iely);
    }
    TMatrixD Ni;
    Ni.ResizeTo(_nSimPerIndex, 1);

    TMatrixD mIndYAYC;
    mIndYAYC.ResizeTo(_nSimPerIndex, _nSimPerIndex);
    TMatrixD mIndYABYC;
    mIndYABYC.ResizeTo(_nSimPerIndex, _nSimPerIndex);
    TMatrixD mIndYBYC;
    mIndYBYC.ResizeTo(_nSimPerIndex, _nSimPerIndex);

    double numXi1 = 0.0;
    double denXi1 = 0.0;
    double numXiT = 0.0;
    double denXiT = 0.0;
    double S1CVMi = 0.0;
    double STCVMi = 0.0;
    // Compute the sum
    Double_t dsumS = 0.;
    Double_t dsumSt = 0.0;
    
    /// Loop on the X attributes
    for (Int_t ix = 0; ix < _nX; ix++)
    {
      attx = (URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(ix));
      if (_blog)
      {
        cout << endl;
        _valSi.Print (Form("*** iy[%d/%d]", _iOut, _nY));
        cout << endl;
        cout << " ** Input att [" << attx->GetName()
        << "] _sIteratorName[" << _sIteratorName << "]_nX["
        << _nX << "]" << endl;
      }

      _tds->getTuple()->Draw( iter_name+":"+_sOutputAttribute, Form("%s==%d", _sIteratorName.Data(), 100 * _nX + ix + 1), "goff" );
      Values = _tds->getTuple()->GetV2(); M_N_Mi[1].assign(Values, Values + _nSimPerIndex);
      IterV = _tds->getTuple()->GetV1(); iterVals.assign(IterV, IterV + _nSimPerIndex);
        // our_parallel_sort(&M_N_Mi[1], &iterVals);

      for (Int_t ipat = 0; ipat < _nSimPerIndex; ipat++)
      {
        /// Matrix M
        Ni(ipat, 0) = M_N_Mi[1][ipat]; //_tds->getValue(soutputy, _telm->GetEntry(ipat), _iely);
      }

      for (Int_t i = 0; i < _nSimPerIndex; i++)
      {
        for (Int_t j = 0; j < _nSimPerIndex; j++)
        {
          if(_valM(j,_iOut-1)<=_valO(i, _iOut-1))
          {
            mIndYAYC(i,j) = 1.0;
          }
          else
          {
            mIndYAYC(i,j) = 0.0;
          }

          if(Ni(j,0)<=_valO(i, _iOut-1))
          {
            mIndYABYC(i,j) = 1.0;
          }
          else
          {
            mIndYABYC(i,j) = 0.0;
          }

          if(_valN(j,_iOut-1)<=_valO(i, _iOut-1))
          {
            mIndYBYC(i,j) = 1.0;
          }
          else
          {
            mIndYBYC(i,j) = 0.0;
          }
        }
      }

      numXi1 = 0.0;
      denXi1 = 0.0;
      numXiT = 0.0;
      denXiT = 0.0;
      double termA1 = 0.0, termA2 = 0.0, termB1 = 0.0, termB2 = 0.0;
      for (Int_t k = 0; k < _nSimPerIndex; k++)
      {
        termA1 = 0.0; 
        termA2 = 0.0;
        termB1 = 0.0; 
        termB2 = 0.0;
        for (Int_t j = 0; j < _nSimPerIndex; j++)
        {
          termA1 += mIndYAYC(k,j) * mIndYABYC(k,j);
          termA2 += ( mIndYAYC(k,j) + mIndYABYC(k,j) );
          termB1 += mIndYBYC(k,j) * mIndYABYC(k,j);
          termB2 += mIndYBYC(k,j) + mIndYABYC(k,j);
        }

        numXi1 += ( termA1 / (double)_nSimPerIndex ) - ( 0.5 * termA2 / (double)_nSimPerIndex ) * ( 0.5 * termA2 / (double)_nSimPerIndex );
        denXi1 += ( 0.5 * termA2 / (double)_nSimPerIndex ) - ( 0.5 * termA2 / (double)_nSimPerIndex ) * ( 0.5 * termA2 / (double)_nSimPerIndex );

        numXiT += ( termB1 / (double)_nSimPerIndex ) - ( 0.5 * termB2 / (double)_nSimPerIndex ) * ( 0.5 * termB2 / (double)_nSimPerIndex );
        denXiT += ( 0.5 * termB2 / (double)_nSimPerIndex ) - ( 0.5 * termB2 / (double)_nSimPerIndex ) * ( 0.5 * termB2 / (double)_nSimPerIndex );
      }

      S1CVMi = numXi1 / denXi1;
      STCVMi = 1.0 - (numXiT / denXiT);

      fillIndex(attx->GetName(), "FirstCVM", S1CVMi, "--first--");
      fillIndex(attx->GetName(), "TotalCVM", STCVMi, "--total--");
      
      dsumS += S1CVMi;
      dsumSt += STCVMi;

      cout << " ** Input att [" << attx->GetName()
        << "] First CVM [" << S1CVMi
        << "] Total Order CVM[" << STCVMi
        << "]" << endl;
    }
    fillIndex("__sum__", "FirstCVM", dsumS, "--first--");
    fillIndex("__sum__", "TotalCVM", dsumSt, "--total--");
  }

}

    void TCramerVonMises::drawIndexes(TString sTitre, const char *select, Option_t * option )
    {
       TString sSelect = TString(select);
      if(sSelect=="")
      {
        TSensitivity::drawIndexes(sTitre, "Method==\"CramerVonMises\"", option);
      }
      else
        TSensitivity::drawIndexes(sTitre, Form("Method==\"CramerVonMises\"&&(%s)",select), option);
    };



} // Fin du namespace Sensitivity
} // Fin du namespace URANIE
