//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2025 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   THSIC.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>, Rudy Chocat <rudy.chocat@cea.fr>, Gabriel Sarazin <gabriel.sarazin@cea.fr>
 \date   2025/01/15

 \brief Implement the URANIE::Sensitivity::THSIC class

 */
/// ROOT
#include "TPad.h"
#include "TProfile.h"
#include "TLeaf.h"
#include "TGaxis.h"
#include "TLatex.h"
#include "TLegend.h"
#include "TMap.h"
#include "TRegexp.h"
#include "TObjString.h"

#include "TDirectory.h"
#include "Math/Math.h"
#include "Math/ProbFuncMathCore.h"

/// Uranie
#include "THSIC.h"
#include "TSensitivity.h"
#include "TKernel.h"
#include "TPatternsEventList.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Sensitivity::THSIC)

namespace URANIE{

    namespace Sensitivity{
    
        ///////////////////
        /// convertRule ///
        ///////////////////

        TString THSIC::convertRule(EEstimatedVariance evRule){
            
            // to initialize the TString
            TString sRule = "";

            // to map each enumerator to its corresponding string representation
            switch(evRule){

                case kUnknown:
                    sRule = "kUnknown";
                    break;

                case kStdEmpirical:
                    sRule = "kStdEmpirical";
                    break;

                case kMedianeDelta:
                    sRule = "kMedianeDelta";
                    break;

                default:
                    sRule = "Invalid enumerator!";
                    
            }

            return sRule;

        } // End of method convertRule

        ////////////////////////////////////
        /// Default constructor (by UTF) ///
        ////////////////////////////////////

        THSIC::THSIC():
            TSensitivityTest(),
            _nEstimatedVariance(kUnknown),
            _nPermutationSample(NDefaultPermutation), _nThresholdPermutation(NDefaultThresholdPermutation), _nThresholdGamma(NDefaultThresholdGamma)
        {
            SetName("HSIC");
            setMethodName("HSic");
            SetTitle("my title of the hSIC indexes");
            cout << " *** Default constrcutor THSIC (empty) ***" << endl;

        } // End of default constructor

        //////////////////////////////////////////
        /// Parameterized constructor (by UTF) ///
        //////////////////////////////////////////

        THSIC::THSIC(URANIE::DataServer::TDataServer *tds, const char *varexpinput, const char *varexpoutput, Option_t * option) :
            TSensitivityTest(tds, varexpinput, varexpoutput, option),
            _nEstimatedVariance(kStdEmpirical),
            _nPermutationSample(NDefaultPermutation), _nThresholdPermutation(NDefaultThresholdPermutation), _nThresholdGamma(NDefaultThresholdGamma)
        {
            SetName("HSIC");
            setMethodName("HSic");
            SetTitle("my title of the hSIC indexes");

            if ( _blog ) {
                cout << " *** _nY["<< _nY << "] with [" << _sSelectedOutput.Data() << "]" << endl;
                cout << " *** _sOutputAttribute[" << _sOutputAttribute << "]" << endl;
            }

            try {
                if ( _nS < NMinimumPatterns )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                            Form("THSIC constructor \nThe number of patterns [%d] is lesser than the number required [%d]", _nS, NMinimumPatterns));

                if ( _sSelectedOutput.Length() == 0 )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                            Form("THSIC constructor \nThe output attribute \"%s\" is empty", _sSelectedOutput.Data()));

            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                throw ue;
            }

            parseOption(option);

        } // End of parametrized constructor

        ///////////////////////////
        /// Destructor (by UTF) ///
        ///////////////////////////

        THSIC::~THSIC()
        {
            if ( _blog )
                cout << " *** destroy THSIC(" << GetName() << ") Title[" << GetTitle() << "] *** " << endl;

        } // End of destructor

        //////////////////////////////////////
        ///  setEstimatedVariance (by UTF) ///
        //////////////////////////////////////

        void THSIC::setEstimatedVariance(EEstimatedVariance nMethod)
        {
            try {
                if ( nMethod == kUnknown )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                            Form("THSIC::setEstimatedVariance\n The method \"kUnknown\" [%d] not authorized.\n We must select one of the \"kStdEmpirical | kMedianeDelta\" methods.", nMethod));
                _nEstimatedVariance = nMethod;
            } catch (URANIE::Exceptions::UErrorExceptions& ue) {
                ue.printMessage();
                throw ue;
            }

        } // End of method setEstimatedVariance

        ///////////////////////////////////////
        ///  setNPermutationSample (by UTF) ///
        ///////////////////////////////////////

        void THSIC::setNPermutationSample(Int_t ns)
        {
            try {
                if ( ns < 1 )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                            Form("THSIC::setNPermutationSample\n The size of permutation (%d) must be greater than 1", ns));
                _nPermutationSample = ns;
            } catch (URANIE::Exceptions::UErrorExceptions& ue) {
                ue.printMessage();
                throw ue;
            }

        } // End of method setNPermutationSample

        /////////////////////////////////////////
        ///  setThresholdPermutation (by UTF) ///
        /////////////////////////////////////////

        void THSIC::setThresholdPermutation(Int_t n)
        {
            try {
                if ( n < 1 )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                            Form("THSIC::setThresholdPermutation\n The Threshold (%d) must be greater than 1", n));
                _nThresholdPermutation = n;
                if ( _nThresholdPermutation > _nThresholdGamma )
                    throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                            Form("THSIC::setThresholdPermutation function\n Change the value of the _nThresholdGamma parameter from [%d] to [%d]", _nThresholdGamma, _nThresholdPermutation));
            } catch (URANIE::Exceptions::UErrorExceptions& ue) {
                ue.printMessage();
                throw ue;
            } catch (URANIE::Exceptions::UInfoExceptions& ue) {
                ue.printMessage();
                _nThresholdGamma = _nThresholdPermutation;
            }

        } // End of method setThresholdPermutation

        ///////////////////////////////////
        ///  setThresholdGamma (by UTF) ///
        ///////////////////////////////////

        void THSIC::setThresholdGamma(Int_t n)
        {
            try {
                if ( n < 1 )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                            Form("THSIC::setThresholdGamma\n The Threshold (%d) must be greater than 1", n));
                _nThresholdGamma = n;
                if ( _nThresholdGamma < _nThresholdPermutation  )
                    throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                            Form("THSIC::setThresholdGamma function\n Change the value of the _nThresholdPermutation parameter from [%d] to [%d]", _nThresholdPermutation, _nThresholdGamma));

            } catch (URANIE::Exceptions::UErrorExceptions& ue) {
                ue.printMessage();
                throw ue;
            } catch (URANIE::Exceptions::UInfoExceptions& ue) {
                ue.printMessage();
                _nThresholdPermutation = _nThresholdGamma ;
            }

        } // End of method setThresholdGamma

        /////////////////////////////
        ///  parseOption (by UTF) ///
        /////////////////////////////

        void THSIC::parseOption(Option_t *option)
        {
            TString soption(option);
            soption.ToLower();

            if( soption.Contains("empiri") )
                _nEstimatedVariance = kStdEmpirical;
            if( soption.Contains("median") )
                _nEstimatedVariance = kMedianeDelta;

        } // End of method parseOption

        /////////////////////////////////
        ///  generateHMatrix (by UTF) ///
        /////////////////////////////////

        void THSIC::generateHMatrix(Option_t* option){

            TString soption = option;
            soption.ToLower();
            if ( _blog )
                cout << "  ** THSIC::generateHMatrix soption[" << soption << "]" << endl;

            Double_t d1OverN = 1.0 / (1.0*_nS);

            _matH.ResizeTo(_nS, _nS);
            _matH.UnitMatrix();
            _matH -= d1OverN;

            if ( _blog )
                _matH.Print();

        } // End of method generateHMatrix

        ///////////////////////
        /// doubleCentering ///
        ///////////////////////

        TMatrixDSym THSIC::doubleCentering(const TMatrixDSym& matK){

            // The input argument is passed by reference to avoid unnecessary copying, which improves performance when working with large matrices.

            // to get the number of rows
            Int_t nRows = matK.GetNrows();

            // to create a local copy of the argument passed by reference
            // --> this TMatrixDSym object will be returned in the end
            // --> remember that matK is passed as a constant reference, and therefore cannot be modified
            TMatrixDSym matHKH(matK);
            
            // to initialize the vector that will contain row means
            TVectorD vRows(nRows);

            // to initialize the variables used to store results
            // --> for the sum of each row
            Double_t dSumRow = 0.0;
            // --> for the sum of the matrix
            Double_t dSumMat = 0.0;

            /////////////////////
            /// Row Summation ///
            /////////////////////

            for(Int_t i=0; i<nRows; i++){
                // at the beginning of the row, the sum is reset to zero
                dSumRow = 0.0;
                // loop over the columns
                for(Int_t j=0; j<nRows; j++){
                    // incrementation of the row sum
                    dSumRow += matK(i,j); 
                }
                // computing and saving the mean of this row
                vRows(i) = dSumRow / (1.0 * nRows);
                // incrementing the matrix sum
                dSumMat += dSumRow;
            }
            
            // computing the mean of the matrix coefficients
            Double_t dMeanMat = dSumMat / (1.0 * nRows * nRows);

            ////////////////////////
            /// Double Centering ///    
            ////////////////////////

            for(Int_t i=0; i<nRows; i++){
                for(Int_t j=0; j<nRows; j++){
                    matHKH(i, j) = matK(i, j) - vRows(i) - vRows(j) + dMeanMat;
                }
            }

            return matHKH;

        } // End of method doubleCentering

        ///////////////////////////
        /// doubleCenteringRect ///
        ///////////////////////////

        TMatrixD THSIC::doubleCenteringRect(const TMatrixD& matK){

            // Note that the function is written as if matK was a matrix with no specific property.
            // In particular, this function can be used for a rectangular matrix.

            // The input argument is passed by reference to avoid unnecessary copying, which improves performance when working with large matrices.

            // to get the number of rows
            Int_t nRows = matK.GetNrows();
            // to get the number of columns
            Int_t nCols = matK.GetNcols();

            // to create a local copy of the argument passed by reference
            // --> this TMatrixD object will be returned in the end
            // --> remember that matK is passed as a constant reference, and therefore cannot be modified
            TMatrixD matHKH(matK);
            
            // to initialize the vector that will contain row means
            TVectorD vRows(nRows);
            
            // to initialize the vector that will contain column means
            TVectorD vCols(nCols);

            // to initialize the variables used to store results
            // --> for the sum of each row
            Double_t dSumRow = 0.0;
            // --> for the sum of each column
            Double_t dSumCol = 0.0;
            // --> for the sum of the matrix
            Double_t dSumMat = 0.0;

            /////////////////////
            /// Row Summation ///
            /////////////////////

            for(Int_t i=0; i<nRows; i++){
                // at the beginning of the row, the sum is reset to zero
                dSumRow = 0.0;
                // loop over the columns
                for(Int_t j=0; j<nCols; j++){
                    // incrementation of the row sum
                    dSumRow += matK(i,j); 
                }
                // computing and saving the mean of this row
                vRows(i) = dSumRow / (1.0 * nCols);
                // incrementing the matrix sum
                dSumMat += dSumRow;
            }
            
            ////////////////////////
            /// Column Summation ///
            ////////////////////////

            for(Int_t j=0; j<nCols; j++){
                // at the beginning of the column, the sum is reset to zero
                dSumCol = 0.0;
                // loop over the rows
                for(Int_t i=0; i<nRows; i++){
                    // incrementation of the column sum
                    dSumCol += matK(i,j); 
                }
                // computing and saving the mean of this column
                vCols(j) = dSumCol / (1.0 * nRows);
            }

            // computing the mean of the matrix coefficients
            Double_t dMeanMat = dSumMat / (1.0 * nRows * nCols);

            ////////////////////////
            /// Double Centering ///    
            ////////////////////////

            for(Int_t i=0; i<nRows; i++){
                for(Int_t j=0; j<nCols; j++){
                    matHKH(i, j) = matK(i, j) - vRows(i) - vCols(j) + dMeanMat;
                }
            }

            return matHKH;

        } // End of method doubleCenteringRect

        /////////////////////
        /// estimateSigma ///
        /////////////////////

        Double_t THSIC::estimateSigma(const char* sAtt, Option_t* option){
            
            // to construct a TString object from the provided Option_t * (equivalent to const char *)
            // --> this copy is necessary since Option_t means const char
            TString sOption = option;

            // to convert the TString object to lower case
            sOption.ToLower();

            // to change the rule of thumb (according to the information given in option)
            parseOption(option);

            // (in debug mode only)
            // to display all relevant information: attribute name + chosen rule of thumb + option
            if(_blog){
                cout << " *** THSIC::estimateSigma *** " << endl << endl;
                cout << Form("Attribute name [%s]", sAtt) << endl;
                cout << Form("Option passed by the user [%s]", option) << endl;
                TString sRule = convertRule(_nEstimatedVariance);
                cout << Form("Rule applied [%s]", sRule.Data()) << endl;
            }

            // to initialize the double that will contain the final value of the kernel parameter	
            Double_t dSigma = -1.0;
            // --> the initial value is negative to indicate that nothing was done

            try{
                switch(_nEstimatedVariance){
                    
                    /////////////////////////////////////////////////////
                    /// Rule of thumb 1: empirical standard deviation ///
                    /////////////////////////////////////////////////////
                    
                    case kStdEmpirical:{

                        // to compute empirical statistics for the attribute with name sAtt
                        // _tds is a pointer to the TDataServer passed as argument of the THSIC constructor
                        _tds->computeStatistic(sAtt);

                        // to get the empirical standard deviation
                        dSigma = _tds->getAttribute(sAtt)->getStd();

                        // (in debug mode only)
                        // to display the new value of dSigma 
                        if (_blog){
                            cout << Form("Taking the empirical standard deviation leads to: Sigma [%e]", dSigma) << endl;
                        }

                        // to exit the switch statement
                        break;
                    }
                    
                    /////////////////////////////////////////////////
                    /// Rule of thumb 2: median of pair distances ///
                    ///////////////////////////////////////////////// 
                    
                    case kMedianeDelta:{

                        // to create a pointer to a new TDataServer (on the heap)
                        // --> this data server is intended to contain all distances $d_{ij} = |x_{i} - x_{j}| for 1 \leq i < j \leq _nS$
                        // --> the attribute name is used to modify the TDataServer name at each iteration, thus ensuring its uniqueness
                        URANIE::DataServer::TDataServer* tdsMedian = new URANIE::DataServer::TDataServer(
                            Form("__tds_median_%s", sAtt),
                            "Data server to compute the median of pair distances"
                        );

                        // name of the new attribute
                        TString sAttName = "_pair_distances_";

                        // to add a new attribute into the TDataServer object
                        tdsMedian->addAttribute(sAttName);

                        // to add a tuple to the TDataServer object
                        tdsMedian->createTuple();

                        // to initialize the doubles that will contain the values of $x_{i}$ and $x_{j}$
                        Double_t dVali = 0.0, dValj = 0.0;

                        // to initialize the integer used as a counter for the distances $d_{ij}$
                        // it ranges from 0 to _nS * (_nS - 1) / 2
                        Int_t iCrt = 0;

                        // to fill the tuple of the TDataServer with the distances $d_{ij}$

                        for(Int_t i=0; i<_nS; i++){
                            
                            // to extract $x_{i}$ from the original TDataServer
                            dVali = _tds->getValue(sAtt, i);
                            
                            for(Int_t j=0; j<i; j++){
                                
                                // to extract $x_{j}$ from the original TDataServer
                                dValj = _tds->getValue(sAtt, j);
                                
                                // to compute $d_{ij}$ and insert the value in the new TDataServer
                                tdsMedian->getTuple()->Fill(iCrt++, TMath::Abs(dVali - dValj));	
                            }
                        }

                        // to compute the median of the sample of all distances 
                        Double_t dProbaMedian=0.5;
                        tdsMedian->computeQuantile(sAttName, dProbaMedian, dSigma);
                        // the value of the quantile is stored in dSigma (passed by reference)

                        // (in debug mode only)
                        // to display the new value of dSigma
                        if (_blog){
                            cout << Form("Taking the median of pair distances leads to: Sigma [%e]", dSigma) << endl;
                        }

                        // to delete the TDataServer object pointed to by tdsMedian, and to assign the pointer to NULL
                        delete tdsMedian; tdsMedian = NULL;

                        // to exit the switch statement
                        break;
                    }
                    
                    ////////////////////////////////////////////////
                    /// Throwing an exception in all other cases ///
                    ////////////////////////////////////////////////		
                    
                    default:{
                        
                        // to define the message to throw
                        TString sMsgOtherCases = 
                            "THSIC::estimateSigma\n"
                            "Define a method [\"empiri\"|\"median\"] to tune the bandwidth parameters of Gaussian kernels."
                            "This must be specified in the option argument of the THSIC constructor.";

                        // to throw the exception
                        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__, sMsgOtherCases);
                        
                        // to exit the switch statement
                        break;
                    }

                } // end of the switch statement 
                
                /////////////////////////////////////////////////////////////////
                /// Throwing an exception if the kernel parameter is negative ///
                /////////////////////////////////////////////////////////////////

                if(dSigma < 0.0){
                    
                    // to define the message to throw
                    TString sMsgNegVal = Form(
                        "THSIC::estimateSigma\n"
                        "The computed kernel parameter [%e] for attribute [%s] is negative.",
                        dSigma, sAtt
                    );

                    // to throw the exception
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, sMsgNegVal);
                }

                //////////////////////////////////////////////////////////////////
                /// Throwing an exception if the kernel parameter is too small ///
                //////////////////////////////////////////////////////////////////

                if (dSigma < dEPSILONSigma){
                
                    // to define the message to throw
                    TString sMsgSmall = Form(
                        "THSIC::estimateSigma\n"
                        "The computed kernel parameter [%e] for attribute [%s] is smaller than dEPSILONSigma [%e].",
                        dSigma, sAtt, dEPSILONSigma
                    );

                    // to throw the exception
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__, sMsgSmall);
                }

            } // end of the try block

            catch(URANIE::Exceptions::UWarningExceptions& ue){
                
                ue.printMessage();
                throw ue;

            } // end of the catch block (warnings)

            catch(URANIE::Exceptions::UErrorExceptions& ue){

                ue.printMessage();
                throw ue;

            } // end of the catch block (errors)

            return dSigma;

        } // End of method estimateSigma

        //////////////////////////
        /// getMatrixKGaussian ///
        //////////////////////////

        TMatrixDSym THSIC::getMatrixKGaussian(const char* sAtt, Double_t dTheta){

            // to create a pointer to a TKernel object
            URANIE::Modeler::TKernel *kerGaussian = new URANIE::Modeler::TKernel(_tds, sAtt, sAtt, "nointercept,nostoreyhat");
            // 1st argument: pointer to the TDataServer object containing the data
            // 2nd argument: string with the list of all input variables (separated by ":") --> simply sAtt
            // 3rd argument: string with the list of all output attributes (separated by ":") --> simply sAtt
            // 4th argument: string containing additional options (relevant for Kriging, but irrelevant for HSIC indices)
            
            // to specify that a Gaussian kernel has to be used
            kerGaussian->setKernel(URANIE::Modeler::TKernel::kRBF, dTheta);
            // 1st argument: a member of the enumeration EKernelType
            // 2nd argument: a double specifying the value of the kernel parameter
            
            // Remark on the kernel parameter:
            // Let $\sigma$ be the bandwidth parameter, often set to the empirical standard deviation.
            // With such parameterization, the kernel parameter is equal to $\theta = 0.5 * \sigma^{-2}$.

            // no details provided in the documentation
            kerGaussian->setLambda(1e-9);

            // to compute the Gram matrix (from the constructed kernel and available data) --> kernelization
            kerGaussian->computeK();

            // to get the Gram matrix
            TMatrixDSym matK = kerGaussian->getMatrixK();

            // to delete the TKernel object pointed to by kerGaussian, and to assign the pointer to NULL
            delete kerGaussian; kerGaussian = NULL;
            
            return matK;

        } // End of method getMatrixKGaussian

        //////////////////
        /// biasedHSIC ///
        //////////////////

        Double_t THSIC::biasedHSIC(const TMatrixDSym& matKx, const TMatrixDSym& matKy){

            // Given the two Gram matrices, this function computes the V-statistic estimate of the HSIC.
            
            // to double center the output Gram matrix 
            TMatrixDSym matHKyH = doubleCentering(matKy);
            
            // Remark:
            // One can also double center the input Gram matrix.
            // However, it is somehow useless as it leads exactly to the same result.

            // to initialize the double used to compute the V-statistic
            Double_t dVstat = 0.0;

            // to compute the trace of the matrix product between matKx and matHKyH
            for(Int_t i=0; i<_nS; i++){
                for(Int_t j=0; j<_nS; j++){
                    dVstat += matKx(i,j) * matHKyH(i,j);
                }
            }

            // final value of the V-statistic estimate
            dVstat /= (1.0 * _nS * _nS);

            return dVstat;

        } // End of method biasedHSIC

        ////////////////////
        /// unbiasedHSIC ///
        ////////////////////

        Double_t THSIC::unbiasedHSIC(const TMatrixDSym& matKx, const TMatrixDSym& matKy){

            // Given the two Gram matrices, this function computes the U-statistic estimate of the HSIC.

            // to create local copies of the two arguments passed by reference
            TMatrixDSym matKxZ = TMatrixDSym(matKx);
            TMatrixDSym matKyZ = TMatrixDSym(matKy);

            // to set all diagonal coefficients to zero
            for(Int_t i=0; i<_nS; i++){
                matKxZ(i,i) = 0.0;
                matKyZ(i,i) = 0.0;
            }

            // to initialize the doubles used for intermediate computations
            Double_t dSum2 = 0.0, dSum3 = 0.0, dSum4 = 0.0;

            // to initialize the vector that will contain the column sums of matKxZ
            TVectorD vColx(_nS);
            // to initialize the vector that will contain the column sums of matKyZ
            TVectorD vColy(_nS);
            
            // to initialize the variables used to store results
            // --> for the sum of all rows in vColx
            Double_t dSumColx = 0.0;
            // --> for the sum of all columns in vColy
            Double_t dSumColy = 0.0;

            // to initialize the double used to compute the U-statistic
            Double_t dUstat = 0.0;

            /////////////////////////////////////
            /// Computation of the double sum ///
            /////////////////////////////////////

            for(Int_t j=0; j<_nS; j++){
                // at the beginning of the column, the two sums are reset to zero
                dSumColx = 0.0;
                dSumColy = 0.0;
                for(Int_t i=0; i<_nS; i++){
                    dSumColx += matKxZ(i,j);
                    dSumColy += matKyZ(i,j);
                    dSum2 += matKxZ(i,j) * matKyZ(i,j);
                }
                vColx(j) = dSumColx;
                vColy(j) = dSumColy;
            }

            /////////////////////////////////////
            /// Computation of the triple sum ///
            /////////////////////////////////////

            for(Int_t i=0; i<_nS; i++){
                dSum3 += vColx(i) * vColy(i);
            }

            dSum3 -= dSum2; 
            
            ////////////////////////////////////////
            /// Computation of the quadruple sum ///
            ////////////////////////////////////////

            dSum4 = matKxZ.Sum() * matKyZ.Sum() - 2.0 * dSum2 - 4.0 * dSum3;

            //////////////////////////////////////////////
            /// Computation of the final HSIC estimate ///
            //////////////////////////////////////////////

            dSum2 /= (1.0 * _nS * (_nS-1));
            dSum3 /= (1.0 * _nS * (_nS-1) * (_nS-2));
            dSum4 /= (1.0 * _nS * (_nS-1) * (_nS-2) * (_nS-3));

            dUstat = dSum2 - 2.0 * dSum3 + dSum4;

            return dUstat;

        } // End of method unbiasedHSIC

        //////////////////////////
        ///  getProductOfNorms ///
        //////////////////////////

        Double_t THSIC::getProductOfNorms(const TMatrixDSym& matKx, const TMatrixDSym& matKy){

            // Given the two Gram matrices, this function estimates the product of the two squared Hilbert-Schmidt (HS) norms:
            // $\lVert C_{xx} \rVert|^2 \times \lVert C_{yy} \rVert|^2$ (see Theorem 4 in [Gretton et al., 2007]).
            // The estimator is equal to the mean of all the off-diagonal coeff  icients of $( (matHKxH).*(matHKyK) ).^2$ where:
            // - matHKxH and matHKyH are the double-centered versions of matKx and matKy,
            // - .* and .^ denote the element-wise operators.
            
            // to double center the input Gram matrix 
            TMatrixDSym matHKxH = doubleCentering(matKx);

            // to double center the output Gram matrix 
            TMatrixDSym matHKyH = doubleCentering(matKy);

            // to initialize the double used for the temporary sum
            Double_t dProdOfNorms = 0.0;

            // to sum over the strictly lower triangular part of the matrix $( (matHKxH).*(matHKyK) ).^2$
            for(Int_t i=0; i<_nS; i++){
                for(Int_t j=0; j<i; j++){
                    dProdOfNorms += 2.0 * matHKxH(i,j) * matHKxH(i, j) * matHKyH(i,j) * matHKyH(i, j);
                }
            }

            // to obtain the mean (now that the sum is computed)
            dProdOfNorms /= (1.0 * _nS * (_nS-1));

            // (in debug mode only)
            // to display the final result
            if(_blog){
                cout << " *** THSIC::getSumOfMatrixBk *** " << endl;
                cout << "Product of the two HS norms [" << dProdOfNorms << "]" << endl;
            }

            return dProdOfNorms;
            
        } // End of method getProductOfNorms

        //////////////////////
        /// permuteKMatrix ///
        //////////////////////

        void THSIC::permuteKMatrix(const TMatrixDSym& theMatK, TList* tlstPerm, TList* tlstPermMat){
            
            // Given a matrix and a list of permutations, this functions constructs the permuted matrices and store them in a list.

            // to get the total number of elements in *tlstPerm (TList)
            Int_t nPerm = tlstPerm->GetEntries();
            // to get the total number of elements in *tlstPermMat (TList)
            Int_t nMat = tlstPermMat->GetEntries();
            // to get the number of rows in matK (TMatrixDSym)
            Int_t nRows = theMatK.GetNrows();

            // (in debug mode only)
            // to check the nature of the 
            if(_blog){
                // to print the function name
                cout << " *** THSIC::permuteKMatrix *** " << endl;
                // to print the dimensions of matK
                cout << Form("theMatK is a TMatrixDSym with dimensions (%d, %d).", nRows, nRows) << endl;
                // to print the length of tlstPerm
                cout << Form("At this step, tlstPerm is a TList of size [%d].", nPerm) << endl;
                // to print the length of tlstPermMat (to check if tlstPermMat is empty)
                cout << Form("At this step, tlstPermMat is a TList of size [%d].", nMat) << endl;
            }

            // the copy constructor allows to build an independent copy of theMatK (TMatrixDSym) on the heap
            TMatrixDSym* newMatK = new TMatrixDSym(theMatK);

            // the initial Gram matrix is added into tlstPermMat
            // it actually corresponds to the identity permutation
            tlstPermMat->Add(newMatK);

            // (only in debug mode)
            // to print the original Gram matrix 
            if(_blog){
                cout << "Here is the provided Gram matrix theMatK:" << endl;
                theMatK.Print("f=%7.3f");
                //  7: total width of the printed number (digits + decimal point)
                // .3: number of digits after the decimal point
                //  f: fixed-point notation
            }

            // to initialize a pointer to a permutation
            URANIE::DataServer::TPatternsEventList *pelPerm = NULL;
            
            // to initialize the indices used to scan rows and columns
            Int_t jPermuted = 0, kPermuted = 0;

            //////////////////////////////////
            /// Loop over all permutations ///
            //////////////////////////////////

            // --> the loop start to 1 because the first elements in tlstIndices is the identity permutation
            for(Int_t i=1; i<nPerm; i++){

                // (in debug mode only)
                // to display the counter of permutations
                if(_blog){
                    cout << endl << Form("Permutation (%d/%d)", i, nPerm) << endl;
                }
                        
                // to initialize a matrix with the appropriate number of rows and columns
                TMatrixDSym matCrt(nRows);
                // all coefficients are set to zero
                matCrt.Zero();

                pelPerm = (URANIE::DataServer::TPatternsEventList*) (tlstPerm->At(i));
                // The (i+1)-th object contained in *tlstPerm (TList) is extracted.
                // --> Note that the "At" method returns a pointer of type TObject*.
                // Then, this first pointer is casted to a pointer of type URANIE::DataServer::TPatternsEventList*.
                // --> Note that the TPatternsEventList class inherits from ROOT's TEventList class.
                // Finally, pelPerm is assigned this casted pointer.

                // (in debug mode only)
                // to display the new permutation 
                if(_blog){
                    cout << "Extracted permutation:" << endl;
                    pelPerm->Print("all");
                }

                // to get the length of the new permutation
                // the "GetN" method returns the number of elements in *pelPerm (URANIE::DataServer:TPatternsEventList)
                Int_t nPat = pelPerm->GetN();
                // --> the GetN method comes from ROOT's TEventList class

                // (in debug mode only)
                // to display the number of elements in the new permutation
                if(_blog){
                    cout << Form("The current permutation is of length [%d].", nPat) << endl;
                }

                /// performing the permutations of the rows and columns ///
                        
                for(Int_t j=0; j<nPat; j++){
                
                    /// for diagonal coefficients ///

                    // to get the position associated to j in pelPerm:
                    // jPermuted = sigma(j) where sigma denotes the permutation
                    jPermuted = pelPerm->GetEntry(j);
                    // --> the GetEntry method comes from ROOT's TEventList class
                    
                    // to define the value of the j-th diagonal coefficient
                    matCrt(j, j) = theMatK(jPermuted, jPermuted);

                    /// for off-diagonal coefficients ///

                    for(Int_t k=0; k<j; k++){
                    
                        // to get the position associated to k in pelPerm:
                        // kPermuted = sigma(k) where sigma denotes the permutation
                        kPermuted = pelPerm->GetEntry(k);

                        // (in debug mode only)
                        // to print the indices j, k, jPermuted and kPermuted
                        if (_blog){
                            cout << Form("Former row index (%d/%d) and new row index (%d/%d).", j, nPat, jPermuted, nPat) << endl;
                            cout << Form("Former column index (%d/%d) and new column index (%d/%d).", k, nPat, kPermuted, nPat) << endl;
                        }

                        // to define the value of the coefficient at position (j,k)
                        matCrt(j, k) = theMatK(jPermuted, kPermuted);

                        // to define the value of the coefficient at position (k,j)
                        matCrt(k, j) = theMatK(kPermuted, jPermuted);
                
                    } // end of loop over k

                } // end of loop of j

                // (in debug mode only)
                // to print the permuted Gram matrix 
                if(_blog){
                    cout << "Here is the permuted Gram matrix matCrt:" << endl;
                    matCrt.Print("f=%7.3f");
                }

                // the copy constructor allows to build an independent copy of matCrt on the heap
                // TMatrixDSym *matPerm = new TMatrixDSym(doubleCentering(matCrt));
                TMatrixDSym* matPerm = new TMatrixDSym(matCrt);

                // the permuted Gram matrix is added into tlstPermMat
                tlstPermMat->Add(matPerm);

            } // end of loop over permutations

            // (in debug mode only)
            // to indicate the end of the function
            if(_blog){
                cout << " *** End Of THSIC::permuteKMatrix *** " << endl;
            }

        } // End of method permuteKMatrix

        //////////////////////
        /// generateSample ///
        //////////////////////

        void THSIC::generateSample(Option_t* option){

            // to suppress unused argument warning
            (void) option;

            try{
                // to define the message to thow
                TString sMsgUseless =
                    "THSIC::generateSample\n"
                    "A dataset already exists. There is no need to generate input samples!";
                
                // to throw the exception
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__, sMsgUseless);

            } // end of the try block

            catch(URANIE::Exceptions::UWarningExceptions& ue){

                ue.printMessage();
                throw ue;

            } // end of the catch block

        } // End of method generateSample

        //////////////////////
        /// computeIndexes ///
        //////////////////////

        void THSIC::computeIndexes(Option_t* option){

            // creation of a TString object
            // as Option_t is an alias for const char, option cannot be modified directly
            // making a copy is necessary to go further
            TString sOption = option;

            // all characters of the string are converted to lower case
            sOption.ToLower();

            // (in verbose mode only)
            // --> in option, "quiet" means minimal verbosity
            if( !( sOption.Contains("quiet") ) ){
                // to print the name of the function
                cout << " *** THSIC::computeIndexes *** " << endl << endl;
            }

            // to estimate all relevant indices
            evaluateIndexes(option);

        } // End of method computeIndexes

        ///////////////////////
        /// evaluateIndexes ///
        ///////////////////////

        void THSIC::evaluateIndexes(Option_t* option){
            
            ////////////////////////////////////
            /// Code to retrieve information ///
            ////////////////////////////////////

            // creation of a TString object
            TString sOption = TString(option);

            // all characters of the string are converted to lower case
            sOption.ToLower();

            // to get the number of input variables
            Int_t nInputs = _listOfInputAttributes->GetEntries();

            // to get the number of output variables
            Int_t nOutputs = _listOfOutputAttributes->GetEntries();

            // to initialize the number of permutations
            Int_t nPerm = 0;
            
            // (in verbose mode only)
            if( !(sOption.Contains("quiet")) ){
                // to print the name of the function
                cout << " *** THSIC::evaluateIndexes *** " << endl << endl;
                // to print the string with all options
                cout << Form("Passed options: option [%s]", sOption.Data()) << endl;
                // to print the sample size (nb of input-output samples)
                cout << Form("Sample size: _nS [%d]", _nS) << endl;
                // to print the number of input variables
                // --> the second line checks that everything is consistent
                cout << Form("Number of input variables: _nX [%d]", _nX) << endl;
                cout << Form("Number of input variables: nInputs [%d]", nInputs) << endl;
                // to print the number of permutations (set to 0 at this step)
                cout << Form("Number of permutations: nPerm [%d]", nPerm) << endl << endl;
            }

            //////////////////////////////////////////////////////////////
            /// Code to handle the (user-requested) nb of permutations ///
            //////////////////////////////////////////////////////////////
        
            // only if the option "nperm=..." is added!
            if( sOption.Contains("nperm=") ){

                // the number of permutations is first set to its default value
                nPerm = _nPermutationSample;

                // (in verbose mode only)
                if( !(sOption.Contains("quiet")) ){
                    // to print the number of permutations (set to its default value)
                    cout << Form("Number of permutations: nPerm [%d]", nPerm) << endl << endl;
                }
                
                // to define a regular expression accepting the following patterns: 
                // "nperm=", "nperm=2", "nperm=100", "nperm=1456"
                TRegexp reg("nperm=[0-9]*");
                
                // to find the pattern (described by the regular expression) into the TString sOption
                TString sPattern = sOption(reg);
                // --> if sOption="quiet:nperm=3000", then sPattern="nperm=3000"

                // to create an independent copy of sPattern
                TString sNperm = sPattern;
                
                // to remove "nperm=" and only keep the number of permutations
                sNperm.Remove(0, 6);
                // --> 0: position of the first character to remove
                // --> 6: number of characters to remove, from the first position

                if( sNperm.IsDigit() ){
                    // to convert the string (only composed of digits) into an integer
                    nPerm = sNperm.Atoi();
                    // to update the number of permutations with the value specified by the user
                    _nPermutationSample = nPerm;
                }

                // to eliminate the pattern from the TString sOption
                sOption.ReplaceAll(sPattern.Data(), "");

                // (in verbose mode only)
                if( !(sOption.Contains("quiet")) ){
                    // to print the number of permutations (set to the value requested by the user)
                    cout << Form("Number of permutations: nPerm [%d]", nPerm) << endl << endl;
                }

                // (in debug model only)
                if(_blog){
                    // to print what remains in sOption
                    cout << Form("Remaining options: sOption [%s]", sOption.Data()) << endl << endl;
                }

            }
            
            /////////////////////////////////////////////////////////////	
            /// Code to handle the (user-requested) type of estimator ///
            /////////////////////////////////////////////////////////////

            Bool_t bUnbiased = kFALSE; 
            // By default, the HSIC estimator is biased (or equivalently, not unbiased). Therefore, V-statistics are used.

            if( sOption.Contains("unbiased") ){
                // if requested by the user, the HSIC estimator must be a U-statistic (instead of a V-statistic)
                bUnbiased = kTRUE;
            }

            // (in verbose mode only)
            // to indicate the type of estimator
            if( !(sOption.Contains("quiet")) ){
                if(bUnbiased){
                    cout << "Estimation based on U-statistics (unbiased, possibly negative)." << endl << endl;
                }else{
                    cout << "Estimation based on V-statistics (non-negative, biased, asymptotically unbiased)." << endl << endl;
                }
            }

            /////////////////////////////////////////////////////////////////
            /// Code to initialize variables for the estimation procedure ///
            /////////////////////////////////////////////////////////////////

            /// precomputed values ///
        
            // to define some recurrent numbers used in estimators
            // let N denote the sample size (_nS)
            // --> for 1/N
            Double_t d1OverN = 1.0 / ( 1.0 * _nS );
            // --> for 1/(N-1)
            Double_t d1OverNMinus1 = 1.0 / ( 1.0 * (_nS-1) );
            // --> for 1/(N^2)
            Double_t d1OverN2 = d1OverN * d1OverN;
            
            // (in debug mode only)
            // to check what are the selected input and output variables
            if(_blog){
                // to explain what is done now
                cout << "Estimation of HSIC and R2-HSIC indices" << endl;
                // to check what are the selected input variables
                cout << Form("Input variables: _sSelectedInput [%s]", _sSelectedInput.Data()) << endl;
                // to check what are the selected output variables
                cout << Form("Output variables: _sSelectedOutput [%s]", _sSelectedOutput.Data()) << endl;
            }

            /// initialization of various local variables ///
            
            // --> for kernel parametrization
            // for empirical standard deviation
            Double_t dCrtSigma = 0.0;
            // for kernel parameter
            Double_t dCrtTheta = 0.0;

            // --> for sensitivity measures
            // for HSIC(X, X) in the denominator of the R2-HSIC index
            Double_t dHSICxx = 0.0;
            // for HSIC(X, Y) --> HSIC index + numerator of the R2-HSIC index
            Double_t dHSICxy = 0.0;
            // for HSIC(Y, Y) in the denominator of the R2-HSIC index
            Double_t dHSICyy = 0.0;
            // for the R2-HSIC index
            Double_t dR2HSICxy = 0.0;
            // for the test p-value
            Double_t dPValue = 0.0;

            // --> for the test of independence (computation of the expectation and variance of the HSIC estimator)
            // sum of all the coefficient of the output Gram matrix
            Double_t dSumOfKx = 0.0;
            // sum of all the coefficient of the input Gram matrix
            Double_t dSumOfKy = 0.0;
            // for the product of the squared HS norms of the cross-covariance operators Cxx and Cyy
            Double_t dProdOfNorms = 0.0;

            // --> for the asymptotic test procedure
            // for E[K(Y, Y)]
            Double_t dEspKyy = 0.0;
            // for E[K(Y, Y')]
            Double_t dEspKyyp = 0.0;
            // for E[K(X, X)]
            Double_t dEspKxx = 0.0;
            // for E[K(X ,X')]
            Double_t dEspKxxp = 0.0;

            // --> for the approximation of the first two moments
            // for E[HSIC]
            Double_t dEspTest = 0.0;
            // for V(HSIC)
            Double_t dVarTest = 0.0;

            // --> for the parameters of the Gamma distribution used to approximate the distribution of the test statistic under H0
            // for the shape parameter
            Double_t dAlphaGamma = 0.0;
            // for the scale parameter
            Double_t dBetaGamma = 0.0;

            ///////////////////////////////////
            /// Loop over output attributes ///
            ///////////////////////////////////

            // to create the iterator for the loop over output attributes
            TIter nextoutput(_listOfOutputAttributes);

            // beginning of the loop over output attributes
            // --> the iterator scans all the objects contained in _listOfOutputAttributes
            // --> each object is first casted to a URANIE::DataServer::TAttribute* and then processed

            while(URANIE::DataServer::TAttribute* attoutput = (URANIE::DataServer::TAttribute*) nextoutput()){
                
                // to get the name of the output attribute
                const char* sNameOut = attoutput->GetName();

                // to store the name of the current output attribute in _sOutputAttribute
                sprintf(_sOutputAttribute, "%s", sNameOut);
                // _sOutputAttribute is defined in the class TSensitivity as a Char_t[MAXLENGTHSTRING] array;
                // when passed to sprintf, it becomes a pointer to the first character of _sOutputAttribute
                // this matches the expected char* type for the first argument of sprintf.

                // (in verbose mode only)
                if( !(sOption.Contains("quiet")) ){
                    cout << Form("Output attribute: _sOutputAttribute [%s]", _sOutputAttribute) << endl;
                    cout << Form("Output attribute: sNameOut [%s]", sNameOut) << endl << endl;
                }

                ///////////////////////////////////////////////////////////////
                /// Rule of thumb to set the parameter of the output kernel ///
                ///////////////////////////////////////////////////////////////
                
                // to estimate the empirical standard deviation
                dCrtSigma = estimateSigma(sNameOut, "empiri");
                // to derive the kernel parameter
                dCrtTheta = 1.0 / (2.0 * dCrtSigma * dCrtSigma);

                // (in debug mode only)
                if(_blog){
                    cout << Form("Bandwith parameter: dCrtSigma [%e]", dCrtSigma) << endl;
                    cout << Form("Kernel parameter: dCrtTheta [%e]", dCrtTheta) << endl << endl;
                }

                //////////////////////////////////////////////
                /// Construction of the output Gram matrix ///
                //////////////////////////////////////////////
                
                // to construct the Gram matrix
                TMatrixDSym matKy = getMatrixKGaussian(sNameOut, dCrtTheta);
                // to compute the sum of all coefficients in the matrix
                dSumOfKy = matKy.Sum();

                // (in debug mode only)
                if(_blog){
                    cout << Form("Gram matrix of the output attribute [%s]", sNameOut) << endl;
                    matKy.Print();
                    cout << Form("Sum of all coefficients: dSumOfKy [%e]", dSumOfKy) << endl;
                }

                ////////////////////////////////
                /// Estimation of HSIC(Y, Y) ///
                ////////////////////////////////

                // a distinction is made between based on the type of estimator
                // U-statistic (for unbiased estimation) or V-statistic (for biased estimation)
                if(bUnbiased){
                    dHSICyy = unbiasedHSIC(matKy, matKy);
                    // the expected arguments are the two Gram matrices (here matKy and matKy)
                } else {
                    dHSICyy = biasedHSIC(matKy, matKy);
                }

                // (in debug mode only)
                if(_blog){
                    cout << Form("HISC(Y, Y): dHSICyy [%e]", dHSICyy) << endl;
                }

                ////////////////////////////////
                /// Data storage preparation ///
                ////////////////////////////////
                
                // initialization of pointers to TList objects
                TList* lstPerm = NULL;
                // --> *lstPerm (TList) will be used to store the permutations of {1,...,_nS}
                TList* lstPermMat = NULL;
                // --> *lstPermMat (TList) will be used to store the permuted Gram matrices

                /////////////////////////////////
                /// Asymptotic test procedure ///
                /////////////////////////////////

                if(nPerm == 0){ 

                    /////////////////////////////////////////////////////////////////
                    /// Estimation of the expectations E[K(Y, Y)] and E[K(Y, Y')] ///
                    /////////////////////////////////////////////////////////////////

                    // to reset the values of dEspKyy and dEspKyyp
                    dEspKyy = 0.0, dEspKyyp = 0.0;

                    // to compute the sum of the off-diagonal coefficients
                    for(Int_t i=0; i<_nS; i++){
                        for(Int_t j=0; j<i; j++){
                            dEspKyyp += matKy(i, j);
                        }
                    }
                    // to convert the sum into an empirical mean
                    dEspKyyp *= 2.0 * d1OverNMinus1 * d1OverN;
                    
                    // to compute the sum of the diagonal coefficients
                    for(Int_t i=0; i<_nS; i++){
                        dEspKyy += matKy(i, i);
                    }
                    // to convert the sum into an empirical mean
                    dEspKyy *= d1OverN;

                    // (in debug mode only)
                    if(_blog){
                        // to check the values of dEspKyy and dEspKyyp
                        cout << Form("Expectation E[K(Y, Y)]: dEspKyy [%e]", dEspKyy) << endl;
                        cout << Form("Expectation E[K(Y, Y')]: dEspKyyp [%e]", dEspKyyp) << endl << endl;
                    }

                ////////////////////////////////////////
                /// Permutation-based test procedure ///
                ////////////////////////////////////////

                } else {
                    
                    // a new TList object (which is empty for now) is created on the heap --> memory allocation
                    // the pre-existing pointer lstPerm is assigned the address of this TList
                    lstPerm = new TList();

                    // to tell the TList that it will be responsible for deleting the objects it contains
                    lstPerm->SetOwner();
                    // now, *lstPerm (TList) "owns" its elements
                    // --> it will be responsible for deleting them when it will be itself deleted

                    // to change the name of the TList
                    lstPerm->SetName(Form("%d permutations for %s", nPerm, sNameOut));

                    //////////////////////////////////////////// 
                    /// Creation of the identity permutation ///
                    ////////////////////////////////////////////
                    
                    // a new TPatternsEventList is created on the heap --> memory allocation
                    // creation of a pointer pelId to this TPatternsEventList
                    URANIE::DataServer::TPatternsEventList* pelId = new URANIE::DataServer::TPatternsEventList(
                        "uranie_HSIC_perm_identity", 
                        "the identity permutation", 
                        _nS
                    );
                    // --> the constructor takes three arguments (name, title, size)
                    
                    // loop to enter all integers from 0 to (_nS - 1) into the TPatternsEventList
                    for(Int_t i=0; i<_nS; i++){
                        // to enter the integer i into the TPatternsEventList
                        pelId->Enter(i);
                    }

                    // (in debug mode only)
                    // to print the created list of integers 
                    if (_blog){	
                        pelId->Print("all");
                    }
                    
                    // to add this permutation into *lstPerm
                    lstPerm->Add(pelId);

                    //////////////////////////////////////////
                    /// Creation of all other permutations ///
                    //////////////////////////////////////////

                    // beginning of the loop over random permutations
                    // --> each new permutation is described by a TPatternsEventList which is stored in the TList tlstPerm

                    for(Int_t ib=0; ib<nPerm; ib++){

                        // a new TList object (which is empty for now) is created on the heap --> memory allocation
                        // creation of a pointer pelPerm to this TPatternsEventList
                        URANIE::DataServer::TPatternsEventList *pelPerm = new URANIE::DataServer::TPatternsEventList(
                            Form("uranie_HSIC_perm_%d", ib+1),
                            Form("random permutation %d", ib+1),
                            _nS
                        );
                        // --> the constructor takes three arguments (name, title, size)
                        // --> at each iteration, the TPatternsEventList has a different name and title
                        
                        // loop to enter all integers from 0 to (_nS - 1)
                        for(Int_t i=0; i<_nS; i++){
                            // to enter the integer i into the list
                            pelPerm->Enter(i);
                        }

                        // to (randomly) mix the integers {1,...,_nS}
                        pelPerm->Mix();

                        // (in debug mode only)
                        // to print the created list of integers 
                        if (_blog){
                            cout << Form("Permutation %d:", ib+1) << endl;	
                            pelPerm->Print("all");
                            cout << endl;
                        }

                        // to add this permutation into lstPerm
                        lstPerm->Add(pelPerm);
                        // *pelPerm (TPatternsEventList) now belongs to *lstPerm (TList)
                        // --> there is no need to delete *pelPerm as *lstPerm will do it (when it will be itself deleted)

                    } // end of the loop over random permutations

                    // (in debug mode only)
                    // to print the list of all created permutations 
                    if(_blog){	
                        lstPerm->Print("all");
                    }

                    // a new TList object (which is empty for now) is created on the heap --> memory allocation
                    // the pre-existing pointer lstPermMat is assigned the address of this TList
                    lstPermMat = new TList();
                    
                    // to tell the TList that it will be responsible for deleting the objects it contains
                    lstPermMat->SetOwner();
                    
                    // to change the name of the TList
                    lstPermMat->SetName(Form("%d permuted Gram matrices for %s", nPerm, sNameOut));
                    
                    // to generate and save all permuted Gram matrices
                    permuteKMatrix(matKy, lstPerm, lstPermMat);

                } // end of the if instruction (asymptotic test vs. permutation-based test)

                //////////////////////////////////
                /// Loop over input attributes ///
                //////////////////////////////////

                // index of the input under study
                Int_t iCrt = 1;
                
                // to create the iterator for the loop over input attributes
                TIter nextinput(_listOfInputAttributes);
                
                // beginning of the loop over input attributes
                // --> the iterator scans all the objects contained in _listOfInputAttributes
                // --> each object is first casted to a URANIE::DataServer::TAttribute* and then processed

                while(URANIE::DataServer::TAttribute* attinput = (URANIE::DataServer::TAttribute*) nextinput()){
                    
                    // to get the name of the output attribute
                    const char* sNameIn = attinput->GetName();
                    
                    // (in verbose mode only)
                    if( !(sOption.Contains("quiet")) ){
                        // to check what is sNameIn
                        cout << Form("Input attribute (%d/%d): sNameIn [%s]", iCrt++, nInputs, sNameIn) << endl << endl;
                    }
                    // rk: post-incrementation of iCrt (the value is printed and then incremented)

                    //////////////////////////////////////////////////////////////
                    /// Rule of thumb to set the parameter of the input kernel ///
                    //////////////////////////////////////////////////////////////
                    
                    // to estimate the empirical standard deviation
                    dCrtSigma = estimateSigma(sNameIn, "empiri");
                    // to derive the kernel parameter
                    dCrtTheta = 1.0 / (2.0 * dCrtSigma * dCrtSigma);

                    // (in debug mode only)
                    if(_blog){
                        // to check the values of dCrtSigma and dCrtTheta
                        cout << Form("Bandwith parameter: dCrtSigma [%e]", dCrtSigma) << endl;
                        cout << Form("Kernel parameter: dCrtTheta [%e]", dCrtTheta) << endl << endl;
                    }

                    /////////////////////////////////////////////
                    /// Construction of the input Gram matrix ///
                    /////////////////////////////////////////////
                    
                    // to construct the Gram matrix
                    TMatrixDSym matKx = getMatrixKGaussian(sNameIn, dCrtTheta);
                    // to compute the sum of all coefficients in the matrix
                    dSumOfKx = matKy.Sum();

                    // (in debug mode only)
                    if(_blog){
                        cout << Form("Gram matrix of the input attribute [%s]", sNameIn) << endl;
                        // to print matKx
                        matKx.Print();
                        // to check what is dSumOfKx
                        cout << Form("Sum of all coefficients: dSumOfKx [%e]", dSumOfKx) << endl << endl;
                    }

                    ///////////////////////////////////////////////
                    /// Estimation of HSIC(X, X) and HSIC(X, Y) ///
                    ///////////////////////////////////////////////
                    
                    if(bUnbiased){
                        // to estimate the HSIC between X and X
                        dHSICxx = unbiasedHSIC(matKx, matKx);
                        // to estimate the HSIC between X and Y
                        dHSICxy = unbiasedHSIC(matKx, matKy);
                    }else{
                        // to estimate the HSIC between X and X
                        dHSICxx = biasedHSIC(matKx, matKx);
                        // to estimate the HSIC between X and Y
                        dHSICxy = biasedHSIC(matKx, matKy);
                    }

                    ///////////////////////////////////
                    /// Estimation of R2-HSIC(X, Y) ///
                    ///////////////////////////////////

                    dR2HSICxy = dHSICxy / TMath::Sqrt(dHSICxx * dHSICyy);

                    // (in debug mode only)
                    if(_blog){
                        // to check the values of dHSICxx, dHSICxy and dR2HSICxy
                        cout << Form("HISC(X, X): dHSICxx [%e]", dHSICxx) << endl;
                        cout << Form("HISC(X, Y): dHSICxy [%e]", dHSICxy) << endl;
                        cout << Form("R2-HISC(X, Y): dR2HSICxy [%e]", dR2HSICxy) << endl << endl;
                    }
                    
                    /////////////////////////////////
                    /// Asymptotic test procedure ///
                    /////////////////////////////////
                    
                    if ( nPerm == 0){

                        /////////////////////////////////////////////////////////////////
                        /// Estimation of the expectations E[K(X, X)] and E[K(X, X')] ///
                        /////////////////////////////////////////////////////////////////

                        // to reset the values of dEspKyy and dEspKyyp
                        dEspKxx = 0.0, dEspKxxp = 0.0;

                        // to compute the sum of the off-diagonal coefficients
                        for(Int_t i=0; i<_nS; i++){
                            for(Int_t j=0; j<i; j++){
                                dEspKxxp += matKx(i, j);
                            }
                        }
                        // to convert the sum into an empirical mean
                        dEspKxxp *= 2.0 * d1OverNMinus1 * d1OverN;
                        
                        // to compute the sum of the diagonal coefficients
                        for(Int_t i=0; i<_nS; i++){
                            dEspKxx += matKy(i, i);
                        }
                        // to convert the sum into an empirical mean
                        dEspKxx *= d1OverN;

                        // (in debug mode only)
                        // to check the values of dEspKxx and dEspKxxp
                        if (_blog){
                            // to check the values of dEspKxx and dEspKxxp
                            cout << Form("Expectation E[K(X, X)]: dEspKxx [%e]", dEspKxx) << endl;
                            cout << Form("Expectation E[K(X, X')]: dEspKxxp [%e]", dEspKxxp) << endl << endl;
                        }

                        ///////////////////////////////////////////
                        /// Estimation of the first two moments ///
                        ///////////////////////////////////////////
                        
                        // This approach is based on asymptotic approximations.
                        // --> see Theorems 3 and 4 in [Gretton et al., 2007]

                        // estimation of the expectation of the HSIC estimator
                        dEspTest = (dEspKxx - dEspKxxp) * (dEspKyy - dEspKyyp) * d1OverN;

                        // computation of the sum of all non-diagonal coefficients of ((matHKxK).(matHKyH))^2 
                        // this allows to approximate the product of the squared HS norms of Cxx and Cyy
                        // --> see Theorem 4 and Appendix A.2 in [Gretton et al., 2007]
                        dProdOfNorms = getProductOfNorms(matKx, matKy);
                        
                        // estimation of the variance of the HSIC estimator
                        dVarTest = 2.0 * dProdOfNorms * (_nS-4) * (_nS-5) * d1OverN * d1OverNMinus1 / (_nS-2) / (_nS-3);

                        // (in debug mode only)
                        if(_blog){
                            // to check the values of dEspTest, dProdOfNorms and dVarTest
                            cout << Form("Expectation of the HSIC estimator: dEspTest [%e]", dEspTest) << endl;
                            cout << Form("Product of the squared HS norms: dProdOfNorms [%e]", dProdOfNorms) << endl;
                            cout << Form("Variance of the HSIC estimator: dVarTest [%e]", dVarTest) << endl << endl;
                        }

                        ///////////////////////////////////////////////////////////////////////////////
                        /// Throwing an exception if the variance of the HSIC estimator is negative ///
                        ///////////////////////////////////////////////////////////////////////////////
                        
                        try{

                            if(dVarTest < 0.0){

                                // to define the message to throw
                                TString sMsgNegVarTest = Form(
                                    "THSIC::evaluateIndexes\n"
                                    "The estimated variance [%e] of the HSIC estimator is negative.",
                                    dVarTest
                                );

                                // to throw the exception
                                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__, sMsgNegVarTest);
                            }
                            
                        } // end of the try block

                        catch (URANIE::Exceptions::UErrorExceptions& ue){

                            ue.printMessage();
                            throw ue;

                        } // end of the catch block (errors)

                        /////////////////////////////////////////////////////////////////////////////
                        /// Parameters of the distribution used to approximate the law under (H0) ///
                        /////////////////////////////////////////////////////////////////////////////

                        // The law of the test statistic under (H0) is approximated by:
                        // - a Gamma law if the HSIC is estimated with a V-statistic (biased),
                        // - a Pearson type-III law if the HSIC is estimated with a U-statistic (unbiased).

                        // The shape and scale parameters of the Gamma law are obgtained through the method of moments.
                        
                        // to compute the shape parameter
                        dAlphaGamma = dEspTest * dEspTest / dVarTest;
                        // to compute the scale parameter
                        dBetaGamma = dVarTest / dEspTest;

                        // (in debug mode only)
                        if(_blog){
                            // to check the values of dAlphaGamma and dBetaGamma
                            cout << Form("Shape parameter: dAlphaGamma [%e]", dAlphaGamma) << endl;
                            cout << Form("Scale parameter: dBetaGamma [%e]", dBetaGamma) << endl << endl;
                        }

                        ////////////////////////////////////////////
                        /// Parametric estimation of the p-value ///
                        ////////////////////////////////////////////
                        
                        if(bUnbiased){
                            dPValue = ROOT::Math::gamma_cdf_c(dHSICxy + dEspTest, dAlphaGamma, dBetaGamma, 0.0);
                        }else{
                            dPValue = ROOT::Math::gamma_cdf_c(dHSICxy, dAlphaGamma, dBetaGamma, 0.0);
                        }
                        
                        // the function "gamma_cdf_c" is the survival function of the Gamma distribution
                        // survival function = complement of the cumulative distribution function (CDF)

                        // (in debug mode only)
                        if(_blog){
                            // to check the values of dPValue
                            cout << Form("P-value of the independence test: dPValue [%e]", dPValue) << endl << endl;
                        }

                    } else {
                    
                    ////////////////////////////////////////
                    /// Permutation-based test procedure ///
                    ////////////////////////////////////////

                        // to initialize the double dCtrHSIC
                        // --> used to recompute the test statistic for all permutations
                        Double_t dCrtHSIC = 0.0;
                        
                        // to initialize the integer nExceedsHSIC
                        // --> used to count how many times the observed value of the test statistic is exceeded
                        Int_t nExceedsHSIC = 0;
                        
                        // to get the length of the list lstMatrixPerm
                        Int_t nSize =  lstPermMat->GetEntries();

                        // (in debug mode only)
                        if(_blog){
                            // to check that nPerm and nSize match (more precisely nSize = nPerm + 1)
                            cout << Form("Number of permutations: nPerm [%d]", nPerm) << endl;
                            cout << Form("Length of lstPermMat: nSize [%d]", nSize) << endl << endl;
                            cout << "Remember that tlstPermMat contains the unpermuted matrix (nSize = nPerm + 1)." << endl << endl;
                        }

                        ////////////////////////////////////////////////
                        /// Non-parametric estimation of the p-value ///
                        ////////////////////////////////////////////////

                        // beginning of the loop over permutations
                        // at each iteration:
                        // --> the permuted output Gram matrix is extracted from lstMatrixPerm
                        // --> the test statistic is recomputed, and compared to the original observed value

                        for(Int_t b=1; b<nSize; b++){

                            // (in debug mode only)
                            if(_blog){
                                cout << Form("Permutation (%d/%d) for input [%s]", b, nSize - 1, sNameIn) << endl;
                            }
                            
                            // to extract the permuted Gram matrix
                            TMatrixDSym* matKyPerm = (TMatrixDSym*) (lstPermMat->At(b));
                            // the "At" method returns a pointer of type TObject*
                            // this first pointer needs to be casted to a pointer of type TMatrixDSym*.

                            // (in debug mode only)
                            if(_blog){
                                // to print the matrix *matKyPerm
                                matKyPerm->Print("f=%7.3f");
                            }

                            // to recompute the test statistic (for the permuted dataset)
                            if(bUnbiased){
                                dCrtHSIC = unbiasedHSIC(matKx, TMatrixDSym(*matKyPerm));
                            }else{
                                dCrtHSIC = biasedHSIC(matKx, TMatrixDSym(*matKyPerm));
                            }

                            // to increment the counter if the new value of the test statistic exceeds the original observed value
                            if(dCrtHSIC > dHSICxy){
                                nExceedsHSIC++;
                            }
                            
                            // (in debug mode only)
                            // to check if the counting process works as expected 
                            if(_blog){
                                cout << Form("Current value of the test statistic: dCrtHSIC [%e]", dCrtHSIC) << endl;
                                cout << Form("Observed value of the test statistic: dHSICxy [%e]", dHSICxy) << endl; 
                                cout << Form("Current value of the counter: nExceedsHSIC [%d]", nExceedsHSIC) << endl << endl; 
                            }
                            
                        } // end of the loop over permutations
                        
                        // final estimate of the p-value
                        dPValue = 1.0 * nExceedsHSIC / nPerm;

                        // (in debug mode only)
                        if(_blog){
                            // to check the values of dPValue
                            cout << Form("P-value of the independence test: dPValue [%e]", dPValue) << endl << endl;
                        }

                    } // end of the if instruction (asymptotic test vs. permutation-based test)

                    ///////////////////////////////////////////////////
                    /// Filling the TTree with the obtained results ///
                    ///////////////////////////////////////////////////

                    // for the R2-HSIC index
                    fillIndex(sNameIn, "R2HSic", dR2HSICxy, "--R2-----");
                    // for the HSIC index
                    fillIndex(sNameIn, "HSic", dHSICxy, "--HSIC---");
                    // for the p-value
                    fillIndex(sNameIn, "pValues", dPValue, "--Pvalue-");

                    // (in verbose mode only)
                    if( !(sOption.Contains("quiet")) ){
                        // to display the three relevant quantities of interest (for this input-output pair)
                        cout << Form("HSIC index: dHSICxy [%e]", dHSICxy) << endl;
                        cout << Form("R2-HSIC index: dR2HSICxy [%e]", dR2HSICxy) << endl;
                        cout << Form("Complement of the p-value: 1.0 - dPvalue [%e]", 1.0-dPValue) << endl << endl;
                    }

                } // end of the loop over input attributes
                
                ////////////////////////////////////////////
                /// End of job for this output attribute ///
                ////////////////////////////////////////////
                
                // (in verbose mode only)
                if ( !(sOption.Contains("quiet")) ){
                    cout << Form("End of job for this output [%s]", sNameOut) << endl << endl;
                }

                ////////////////
                /// Cleaning ///
                ////////////////

                if( lstPerm != NULL ){
                    delete lstPerm;
                }
                lstPerm = NULL;

                if( lstPermMat != NULL ){
                    delete lstPermMat;
                }
                lstPermMat = NULL;

            } // end of the loop over output attributes

        } // End of method evaluateIndexes

        /////////////////////////////
        ///  drawIndexes (by UTF) ///
        /////////////////////////////

        void THSIC::drawIndexes(TString sTitle, const char *select, Option_t* option){

            TString soption(option);
            soption.ToLower();

            if ( _blog )
                cout << " *** THSIC::drawIndexes sTitle[" << sTitle << "] select[" << select << "] option[" << soption << "]" << endl;

            /// Case to activate the log (locally)
            if ( soption.Contains("log") )
                setLog();

            /// case to draw the legend
            Bool_t bDrawLegend = kFALSE;
            if ( soption.Contains("legend") )
                bDrawLegend = kTRUE;

            // Case of default output to draw : the first output
            TString sDefaultOutputDraw = Form("%s", _sOutputAttribute);
            if ( _nY > 1 ) {
                URANIE::DataServer::TAttribute *atty = (URANIE::DataServer::TAttribute *) (_listOfOutputAttributes->At(0));
                sDefaultOutputDraw = atty->GetName();
            }
            if ( _blog )
                cout << " *** _nY(" << _nY << ") sDefaultOutputDraw[" << sDefaultOutputDraw << "]" << endl;

            Double_t dThreshold = 1.05;
            TString sSelect = TString(select);
            try {
                if ( _ntresult->GetEntries() != 3 * _nX *_nY  ) {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                            Form("THSIC::drawIndexes method \n The TTree of results does not contain the good number of entry (%d) than (%d)\n Do you have call the computeIndexes() method ?", (Int_t)(_ntresult->GetEntries()), 3 * _nX *_nY));
                }

                if ( ( sSelect.Length() > 0 ) && ( sSelect.Contains("Out==") ) ) {
                        TRegexp regOutput("Out==\".*\"");
                        TString SOutTmp = sSelect(regOutput);
                        sDefaultOutputDraw = SOutTmp.Remove(0, 5);
                        sDefaultOutputDraw.ReplaceAll("\"", "");
                }

                if ( !_tds->isAttribute(sDefaultOutputDraw) )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                            Form("THSIC::drawIndexes method \n The Output attribute \"%s\" does not exist in the TDS (%s)",
                                                                sDefaultOutputDraw.Data() , _tds->GetName()));


                if (soption.Contains("pvalues="))
                {
                    TRegexp reg("pvalues=[+/-]?[0-9.]*.[0-9]*");
                    TString out = soption(reg);
                    TString sout = out.Remove(0, 8);
                    dThreshold = sout.Atof();
                    if ( _blog )
                        cout << " ** dThreshold[" << dThreshold << "] in Regular Expression" << endl;

                    if ( dThreshold < 0.0 )
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                                Form("THSIC::drawIndexes method \n The given threshold for the Pvalues is negative [%f]. It must be in 0.0 and 1.0", dThreshold));

                    if ( dThreshold > 1.0 ) {
                        dThreshold = 1.05;
                        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                                    Form("THSIC::drawIndexes method \n The given threshold for the Pvalues is greater than 1.0  [%s] .\n Then %f is selected ( no selection ).", out.Data(), dThreshold));
                    }


                }
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                throw ue;
            }
            catch (URANIE::Exceptions::UWarningExceptions& ue)
            {
                ue.printMessage();
            }

            Int_t nval = (Int_t) (_ntresult->GetEntries()/_nY);
            Int_t natt = _nX; // 3 indexes stored in the computeIndexes in the order :: dR2HSICXkY , dHSIC, dPValue

            // debug
            if ( _blog ) {
                cout << "  ** THSIC::drawIndexes" << endl;
                cout << "  ** sTitle[" << sTitle << "]" << endl;
                cout << "  ** select[" << select << "]" << endl;
                cout << "  ** sDefaultOutputDraw[" << sDefaultOutputDraw << "]" << endl;
                cout << "  ** option[" << option << "]" << endl;
                cout << "  ** dThreshold[" << dThreshold << "]" << endl;
                cout << "  ** nval[" << nval << "] natt[" << natt << "]" << endl;
                cout << "  ** _sMethodName[" << _sMethodName << "]" << endl;
            }

            // Clean the TEventList
            if( gROOT->FindObject("evlHSIC") )
                gROOT->Delete("evlHSIC");

            // _ntresult->Draw(">>evlHSIC",  Form("(Out==\"%s\")&&(Algo==\"--pvalues--\")&&(Value<%f)", sDefaultOutputDraw.Data(), dThreshold), "goff");
            _ntresult->Draw(">>evlHSIC",  Form("(Out==\"%s\")&&(Order==\"pValues\")&&(Value<%f)", sDefaultOutputDraw.Data(), dThreshold), "goff");
            TEventList *telHSIC = (TEventList*) gROOT->Get("evlHSIC");
            Int_t nattSelect = telHSIC->GetN();

            // cout << Form("nattSelect [%d]", nattSelect) << endl;

            try
            {
                if (nattSelect > _nX )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                            Form("THSIC::drawIndexes ERROR : the number of results [%d] is greater than the number of inputs [%d]",
                                                                    nattSelect, _nX));
            }catch(URANIE::Exceptions::UErrorExceptions &ue)
            {
                ue.printMessage();
                throw ue;
            }

            if ( _blog ) {
                telHSIC->Print("all");
                cout << " ** evlHSIC natt selected [" << nattSelect << "]" << endl;
            }

            // a local variable for the event type
            char etype[500];

            _ntresult->GetLeaf("Inp")->SetAddress(&etype);

            TLeaf *leafSobol = _ntresult->GetLeaf("Value");

            if( gROOT->FindObject("R2HSic") )
                gROOT->Delete("R2HSic");
            TH1F *hR2HSic = new TH1F("R2HSic","Option B example",nattSelect,1,nattSelect+1);
            TMap *mapAtt = new TMap();
            mapAtt->SetOwnerKeyValue();

            for(Int_t ipat=0; ipat< nattSelect; ipat++) {
                // R2HSic
                _ntresult->GetEntry(telHSIC->GetEntry(ipat)-2);
                if ( _blog )
                    cout << " *** (" << ipat << "/" << nattSelect << ") - Out(" << sDefaultOutputDraw << ") etype[" << etype << "] HSic[" << leafSobol->GetValue() << "]" << endl;
                hR2HSic->SetBinContent(ipat+1, leafSobol->GetValue());
                hR2HSic->GetXaxis()->SetBinLabel(ipat+1,etype);

                // PValues :: indexes of the entry stored in the TMap object indexed by the attribute name
                TObjString* key = new TObjString(Form("%s", etype));
                TObjString* value = new TObjString(Form("%d", (int)(telHSIC->GetEntry(ipat))));
                mapAtt->Add(key, value);
            }

            if ( _blog ) {
                hR2HSic->Print("all");
                mapAtt->Print("all");
            }

            hR2HSic->SetTitle("R^{2}_{H_{Sic}} and P_{values}");

            // Title of the graph
            TString sTempTitle = sTitle;

            // The output name
            sTempTitle += Form(" for \"%s\" ", sDefaultOutputDraw.Data());

            if ( dThreshold <1.0 ) {
                sTempTitle += Form(" ( P_{Values} < %.2f )", dThreshold);
            }

            if ( natt > nattSelect )
                sTempTitle += Form(" [ %d/%d attributes ]", nattSelect, natt);
            else
                sTempTitle += Form(" [ %d attributes ]", natt);


            hR2HSic->SetTitle(sTempTitle);
            hR2HSic->LabelsOption(">");
            hR2HSic->SetStats(0);
            hR2HSic->SetBarWidth(0.80);
            hR2HSic->SetBarOffset(0.2);
            hR2HSic->SetFillColor(kBlue); // 38 kBlue
            hR2HSic->SetMinimum(0.0);
            //  hR2HSic->SetMaximum(1.0);
            hR2HSic->Draw("bar2");

            // Set the X/Y axis of the histogramm
            hR2HSic->GetXaxis()->SetTickLength(0.25);
            hR2HSic->GetXaxis()->SetTickSize(0.0);
            if ( !bDrawLegend ) {
                hR2HSic->GetYaxis()->SetLabelColor(hR2HSic->GetFillColor());
                hR2HSic->GetYaxis()->SetAxisColor(hR2HSic->GetFillColor());
            }

            gPad->SetGridy();
            gPad->Modified();
            gPad->Update();

            // PValues with HSIC


            if( gROOT->FindObject("hPValues") )
                gROOT->Delete("hPValues");
            TH1F *hPValues = new TH1F("hPValues","Option B example",nattSelect,1,nattSelect+1);
            for(Int_t ipat=0; ipat< nattSelect; ipat++) {
                if( _blog )
                    cout << "  *** ipat(" << ipat << "/" << nattSelect << ") att[" << hR2HSic->GetXaxis()->GetBinLabel(ipat + 1) << "]";
                TString rangeInfo = mapAtt->GetValue(hR2HSic->GetXaxis()->GetBinLabel(ipat + 1))->GetName();
                if( _blog )
                    cout << " sRangeInfo[" <<rangeInfo << "]";
                Int_t rangeN = atoi(rangeInfo.Data());
                if( _blog )
                    cout << " rangeN[" <<rangeN << "]";
                _ntresult->GetEntry(rangeN);
                hPValues->SetBinContent(ipat+1, 1.0-leafSobol->GetValue());

                //hPValues->GetXaxis()->SetBinLabel(ipat+1,etype);
                if( _blog )
                    cout << " etype[" << etype << "] PValues[" << leafSobol->GetValue() << "]" << endl;
            }

            if( _blog )
                hPValues->Print("all");

            hPValues->SetMarkerStyle(8);
            hPValues->SetMarkerSize(1.2);
            hPValues->SetLineColor(kRed); // kRed 38);
            hPValues->SetLineWidth(2);
            hPValues->SetMarkerColor(hPValues->GetLineColor());

            // Scale hPValues to the pad coordinates
            Float_t rightmax = 1.1*hPValues->GetMaximum();
            Float_t scale = gPad->GetUymax()/rightmax;
            hPValues->Scale(scale, "nosw2");
            hPValues->Draw("LP,SAME");

            // Draw an axis on the right side
            TGaxis *axis = new TGaxis(gPad->GetUxmax(),gPad->GetUymin(), gPad->GetUxmax(), gPad->GetUymax(),0,rightmax,510,"+LW", -0.80);
            axis->SetLineColor(hPValues->GetMarkerColor());
            axis->SetLabelColor(hPValues->GetMarkerColor());
            axis->SetTickLength(0.50);
            axis->SetMaxDigits(5);
            axis->Draw();

            if ( bDrawLegend  ) {
                TLegend *legend = new TLegend(0.70,0.70,0.87,0.80,NULL,"brNDC");
                legend->SetBorderSize(0);
                legend->SetLineColor(1);
                legend->SetLineStyle(1);
                legend->SetLineWidth(1);

                legend->AddEntry(hR2HSic,"R^{2}_{H_{Sic}}","f");
                legend->AddEntry(hPValues,"1.0 - P_{values}","lp");
                legend->Draw("b");
            } else {
                // Draw the R2Hsic title axis
                TLatex *R2HSictitle = new TLatex(1.1,1.05*gPad->GetUymax(),"R^{2}_{H_{Sic}}");
                R2HSictitle->Draw();
                R2HSictitle->SetTextSize(0.04);
                R2HSictitle->SetTextColor(kBlack);
                R2HSictitle->SetTextColor(hR2HSic->GetFillColor());
                R2HSictitle->SetTextAngle(0.);
                R2HSictitle->SetTextAlign(31);

                // Draw the PValues title axis
                TLatex *PvaluesHSictitle = new TLatex(nattSelect+1.0,1.05*gPad->GetUymax(),"1.0 - P_{values}");
                PvaluesHSictitle->Draw();
                PvaluesHSictitle->SetTextSize(0.04);
                PvaluesHSictitle->SetTextColor(hPValues->GetMarkerColor());
                PvaluesHSictitle->SetTextAngle(0.);
                PvaluesHSictitle->SetTextAlign(31);
            }

            gPad->Modified();
            gPad->Update();

            if ( soption.Contains("log") )
                unsetLog();

            //clean
            delete mapAtt; mapAtt = NULL;

        } // End of drawIndexes

        ////////////////////
        /// printSummary ///
        ////////////////////

        void THSIC::printSummary(){

            // This functions prints the SA results in the terminal in a synthetic way.

            ///////////////////////////////////////////
            /// Description of the original context ///
            ///////////////////////////////////////////

            // to print the name of the function 
            cout << " *** THSIC::printSummary *** " << endl << endl;

            // to print the sample size (dataset stored in the TDataServer)
            cout << Form("Sample size: _nS [%d]", _nS) << endl;
            // to print the number of input variables
            cout << Form("Number of input variables: _nX [%d]", _nX) << endl;
            // to print the number of output variables
            cout << Form("Number of output variables: _nY [%d]", _nY) << endl;
            // to print the number of permutations used by the independence test
            cout << Form("Number of permutations: _nPermutationSample [%d]", _nPermutationSample) << endl << endl;

            ///////////////////
            /// Double loop ///
            ///////////////////

            // to initialize the integers used to loop over inputs and outputs
            Int_t iCrtX = 1, iCrtY = 1;

            // to initialize the doubles used to store the values of the sensitivity measures
            Double_t dHSICxy = 0.0, dR2HSICxy = 0.0, dPValue = 0.0;

            // to create an iterator to loop over the output attributes
            TIter nextoutput(_listOfOutputAttributes);

                // beginning of the loop over the output attributes
                // rk: the object extracted from the TList is a pointer to a TObject
                // --> it must be casted to a pointer of type URANIE::DataServer::TAttribute*

                while(URANIE::DataServer::TAttribute* attoutput = (URANIE::DataServer::TAttribute*) nextoutput()){

                // to get the name of the current output attribute
                const char* sNameOut = attoutput->GetName();

                // to print the name of this output attribute
                cout << Form("Output under study: [%s] (%d/%d)", sNameOut, iCrtY++, _nY) << endl << endl;

                // to create an iterator to loop over the input attributes
                TIter nextinput(_listOfInputAttributes);

                // to reset the integer iCtrX (useful starting from iteration #2)
                iCrtX = 1;

                // beginning of the loop over the input attributes

                while (URANIE::DataServer::TAttribute* attinput = (URANIE::DataServer::TAttribute*) nextinput()){

                    // to get the name of the current input attribute
                    const char* sNameIn = attinput->GetName();

                    // to print the name of this input attribute
                    cout << Form("Source of uncertainty: [%s] (%d/%d)", sNameIn, iCrtX++, _nX) << endl << endl;

                    // to define the filter used to select the correct output variable
                    TString sFilter = Form("Out==\"%s\"", sNameOut);

                    // to extract the sensitivity measures from the TTree where they are stored
                    dHSICxy = getValue("HSic", sNameIn, sFilter.Data());
                    dR2HSICxy = getValue("R2HSic", sNameIn, sFilter.Data());
                    dPValue = getValue("pValues", sNameIn, sFilter.Data());

                    // to print the values of the sensitivity measures
                    cout << Form("HSIC index: [%e]", dHSICxy) << endl;
                    cout << Form("Normalized HSIC index (R2-HSIC): [%e]", dR2HSICxy) << endl;
                    cout << Form("Complement to the p-value: [%e]", 1.0 - dPValue) << endl << endl;

                } // end of the loop over input attributes
                        
            } // end of the loop over output attributes

            // to indicate the end of the function 
            cout << " *** End of THSIC::printSummary *** " << endl << endl;

        } // End of method printSummary

        ////////////////////////
        /// printLog (by GS) ///
        ////////////////////////

        void THSIC::printLog(Option_t* option){

            // This function simply prints all class members.

            // to print the name of the function 
            cout << " ** THSIC::printLog[]" << endl;

            // to print the rule of thumb
            TString sRule = convertRule(_nEstimatedVariance);
            cout << Form(" ** Rule of thumb to compute kernel parameters: %s", sRule.Data()) << endl;

            // to print the user-requested number of permutations
            cout << Form(" ** The number of permutations is currently set to: _nPermutationSample [%d]", _nPermutationSample) << endl;

            // to print the maximum number of permutations allowed
            cout << Form(" ** Maximum number of permutations allowed: _nThresholdPermutation [%d]", _nThresholdPermutation) << endl;

            // to print all class members inherited from the TSensitivityTest class
            TSensitivityTest::printLog(option);
            
            // to indicate the end of the function
            cout << " ** End of THSIC::printLog[]" << endl;

        } // End of method printLog

    } // End of namespace Sensitivity

} // End of namespace URANIE
