//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2025 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   THSIC.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>, Rudy Chocat <rudy.chocat@cea.fr>, Gabriel Sarazin <gabriel.sarazin@cea.fr>
 \date   2025/01/15

 \brief Interface of the URANIE::Sensitivity::THSIC class

 */
/**
 *
 * @class URANIE::Sensitivity::THSIC
 *
 * @brief Description of the class THSIC.
 *
 * \section sectionIntro Description of the class THSIC.
 *
 * \subsection subsectionIntro Introduction
 *
 * The goal of this class is to find the list of input attributes which are dependent with the output attributes
 * using the \b "HSIC" values (for "Hilbert-Schmidt Independence Criterion").
 * This approach was introduced by Gretton (2005, 2007).
 *
 * Moreover, in the same time, we compute the \f$ R^{2}_{HSIC} \f$ proposed by Da Veiga (2015) given by the formula
 * \f[
 *  R^{2}_{HSIC}  ( X, Y) \; = \; \frac{ HSIC ( X, Y) }{\sqrt{ HSIC ( X, X) \, \times \, HSIC ( Y, Y)}}
 * \f]
 *
 * \subsection subsectionDraw Draw Indexes
 *
 * We use a \b "Pareto chart" to draw the \b "p-Values" and the \b \f$ R^{2}_{HSIC} \f$.
 *
 * \subsection subsectionBiblio Bibliography
 *
 * \li Da Veiga, S. " Global Sensitivity Analysis with Dependence Measures ", https://arxiv.org/abs/1311.2483, 2013
 * \li Zhang, Q., Filippi, S., Gretton, A., Sejdinovic, D., "Large-Scale Kernel Methods for Independence Testing", 2016
 * \li Gretton, A., Fukumizu, K., Teo, C.H., Song, L., Schoelkopf, B., et Smola, A.,
 *     "A Kernel Statistical Test of Independence,", NIPS 2007
 * \li Gretton, A., Herbrich, R., Smola, A., Bousquet, O., et Schoelkopf, B.,
 *     "Kernel Methods for Measuring Independence", Journal of Machine Learning Research, 6 , pp.2075--2129, 2005
 * \li The Pareto chart, https://en.wikipedia.org/wiki/Pareto_chart
 */

#ifndef THSIC_H
#define THSIC_H

/// ROOT
#include "TVector.h"
#include "TMatrixDSymfwd.h"

/// Uranie
#include "TSensitivityTest.h"

/// Default Values
#define NDefaultPermutation 500
#define NDefaultThresholdPermutation 7
#define NDefaultThresholdGamma 7
#define NMinimumPatterns 6
#define dEPSILONSigma 1.e-12

namespace URANIE {
namespace Sensitivity {
class THSIC: public TSensitivityTest {
public:
    
    enum EEstimatedVariance
    {
        kUnknown,
        kStdEmpirical,
        kMedianeDelta
    };

    //---------------------------------------------
    /** @name Static methods
     **/
    //@{

    /*!
     * \brief Outputs the string representation of any enumerator of EEstimatedVariance.
     *
     * \param evRule (EEstimatedVariance) enumerator of EEstimatedVariance that must be represented by a string
     */
    static TString convertRule(EEstimatedVariance evRule);

    //@}



    //---------------------------------------------
    /** @name Constructors and destructor
     **/
    //@{

    /*!
     * \brief Default constructor
     */
    THSIC();

    /*!
     * \brief THSIC constructor with data contined in the TDS and the inputs and outputs attributes
     *
     * \param tds
     * \param sInputs
     * \param sOutputs
     * \param option (Option_t*) option to pass [""]
     */
    THSIC(URANIE::DataServer::TDataServer* tds, const char* sInputs, const char* sOutputs, Option_t* option="");

    /*!
     * \brief Default destructor
     */
    virtual ~THSIC();

    //@}



    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{

    /*!
     * \brief Which method used to estimate the variance of the gaussian kernel ( kUnknown | kStdEmpirical | kMedianeDelta )
     *
     * \li \b kStdEmpirical : the \f$ \sigma \f$ parameters is estimated by the emperical standatd deviation of the attributes
     * \li \b kMedianeDelta : use the median distance between points
     * \return the method used to estimated the variance of the gaussian kernel ( kStdEmpirical | kMedianeDelta )
     */
    EEstimatedVariance getEstimatedVariance(){
        return _nEstimatedVariance;
    }

    /*!
     * \brief Set the method used to estimate the variance of the gaussian kernel ( kUnknown | kStdEmpirical | kMedianeDelta )
     *
     * \param nMethod (EEstimatedVariance) One of the ( kStdEmpirical | kMedianeDelta ) choice
     * \exception UErrorExceptions for the case (kUnknown)
     */
    void setEstimatedVariance(EEstimatedVariance nMethod);

    /*!
     * \brief getNPermutationSample
     * \return The number of Permutation sample (permutation zize) to estimate the pValues
     */
    Int_t getNPermutationSample(){
        return _nPermutationSample;
    }

    /*!
     * \brief Set the Permutation sample permutation Size (default NDefaultPermutation)
     * \param ns size of the permutation permutation size
     */
    void setNPermutationSample(Int_t ns);

    /*! Set the threshold (nS/nX) to use the permutation permutation test to cumpute the p-value (default NDefaultThresholdPermutation)
      */
    void setThresholdPermutation(Int_t n);
    /*!
     * \brief Get the Threshold (nS/nX) to use the permutation permutation test to compute the p-value (default NDefaultThresholdPermutation)
     *
     * \return The current Threshold (nS/nX) for using the permutation permutation test
    */
    Int_t getThresholdPermutation(){
        return _nThresholdPermutation;
    }

    /*!
     * \brief Set the Threshold (nS/nX) to use the Gamma approximation to cumpute the p-value (default NDefaultThresholdGamma)
     *
     * \param n (Int_t) the Threshold (nS/nX) for using the Gamma approximation
     */
    void setThresholdGamma(Int_t n);

    /*!
     * \brief Get the Threshold (nS/nX) to use the Gamma approximation to cumpute the p-value (default NDefaultThresholdGamma)
     *
     * \return The current Threshold (nS/nX) for using the Gamma approximation
     */
    Int_t getThresholdGamma() {
        return _nThresholdGamma;
    }

    //@}



    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{

    /** @name Parse the option
     *
     * The options treated by the THSIC object were :
     * \li "empiri" : Choose the Standard deviation of each attributes to compute the \f$ \sigma\f$ parameter
     * \li "median" : Choose the median distance between points of each attributes to compute the \f$ \sigma\f$ parameter
     *
     * \see TSensitivity::parseOption()
     **/

    /*!
     * \brief Parse the option given by the user
     * \param option (Option_t) the string of option
     */
    virtual void parseOption(Option_t *option = "");

    //@}



    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{

    /*!
     * \brief Prints a concise summary of results to the terminal.
     * 
     * Specifically, for each input-output pair \f$ (X_i, Y_j) \f$, the following results are displayed:
     * - the HSIC index,
     * - the normalized HSIC index (or R2-HSIC index),
     * - the p-value of the independence test.
     */
    void printSummary();

    /*!
     * \brief Prints all attributes of the class for the current instance.
     *
     * \param option (Option_t*) Directly forwarded to the TSensitivityTest::printLog method.
     */
    virtual void printLog(Option_t* option="");

    //@}



    //---------------------------------------------
    /** @name All methods used for elementary tasks
     *
     **/
    //@{

    /*!
     * \brief generate the \f$ H \f$ matrix
     *
     * The \f$ H \f$ matrix is defined by the formula
     * \f[ H (i, j) \, = \,  1\!\!I_{nS, nS} \: - \: \frac{1}{nS} \: \delta_{i=j} \f]
     * \param option (Option_t *) option to pass [""]
     */
    void generateHMatrix(Option_t *option = "");

    /*!
     * \brief Centers the rows and columns of the provided matrix.
     *
     * \param matK (TMatrixD) The \f$ n \f$-by-\f$ m \f$ matrix \f$ K \f$ that needs to be centered.
     * \return (TMatrixD) The double-centered matrix produced from $\f$ K \f$.
     * 
     * It is defined as \f$ \widetilde{K} := H_n \, K \, H_m \f$ where:
     * \f$ \forall \, r \geq 1 \f$, \f$ H_r \f$ is the centering matrix of dimension \f$ r \f$.
     * 
     * Note that the sum of each row (resp. column) of \f$ \widetilde{K} \f$ is equal to zero.
     */
    TMatrixD doubleCenteringRect(const TMatrixD& matK);

    /*!
     * \brief Centers the rows and columns of the provided matrix.
     *
     * \param matK (TMatrixDSym) The \f$ n \f$-by-\f$ n \f$ matrix \f$ K \f$ that needs to be centered.
     * \return (TMatrixDSym) The double-centered matrix produced from $\f$ K \f$.
     * 
     * It is defined as \f$ \widetilde{K} := H \, K \, H \f$ where:
     * \f$ H \f$ is the centering matrix of dimension \f$ n \f$.
     * 
     * Note that the sum of each row (resp. column) of \f$ \widetilde{K} \f$ is equal to zero.
     */
    TMatrixDSym doubleCentering(const TMatrixDSym& matK);

    /*!
     * \brief Computes the bandwidth parameter of the specified attribute.
     *
     * This function calculates the bandwidth parameter based on the observations of the specified attribute, which are stored in the attached TDataServer.
     *
     * \param sAtt (const char*) The name of the attribute for which the bandwith parameter is to be computed.
     * \param option (Option_t*) The rule of thumb used to compute the bandwidth parameter from the data. 
     * The user can choose between:
     * - "empiri": empirical standard deviation,
     * - "median": median of pair distances.
     * 
     * \return (Double_t) The computed value of the bandwidth parameter.
     */
    Double_t estimateSigma(const char* sAtt, Option_t* option="");

    /*!
     * \brief Constructs the Gram matrix associated with the specified attribute.
     * 
     * This function uses the observations of the specified attribute, stored in the attached TDataServer, for kernelization.
     * The kernel function applied to the data is a Gaussian kernel, whose parameter is provided as second argument.
     *
     * \param sAtt (const char*) The name of the attribute for which the Gram matrix is to be constructed.
     * \param dTheta (Double_t) The parameter of the Gaussian kernel.
     * 
     * \return (TMatrixDSym) The constructed Gram matrix.
     */
    TMatrixDSym getMatrixKGaussian(const char* sAtt, Double_t dTheta);

    /*!
     * \brief Computes the V-statistic estimate of \f$ HSIC(X, Y) \f$ where \f$ X \f$.
     * The computation is based on the Gram matrices \f$ K_{x} \f$ and \f$ K_{y} \f$ which are constructed beforehand from the observations of \f$ X \f$ and \f$ Y \f$.
     * 
     * The formula of the V-statistic estimator can be found in:
     * [Zhang et al., 2018] "Large-scale kernel methods for independence testing", Statistics and Computing", page 117, Eq (9).
     *
     * Let \f$ \widetilde{K}_{x} \f$ be the double-centered Gram matrix built from the observations of \f$ X \f$.
     * Let \f$ \widetilde{K}_{y} \f$ be the double-centered Gram matrix built from the observations of \f$ Y \f$.
     * 
     * The formula of the V-statistic estimator is given by:
     * 
     * \f[ \widehat{HSIC}_{V}(X, Y) \, = \, \frac{1}{nS^2} \, Tr( \widetilde{K}_{x} \widetilde{K}_{y} ) \f]
     *
     * \param matKx (TMatrixDSym) The Gram matrix built from the observations of \f$ X \f$. 
     * \param matKy (TMatrixDSym) The Gram matrix built from the observations of \f$ Y \f$. 
     * 
     * \return (Double_t) The value of the V-statistic estimate of \f$ HSIC(X, Y) \f$.
     */
    Double_t biasedHSIC(const TMatrixDSym& matKx, const TMatrixDSym& matKy);

    /*!
     * \brief Computes the U-statistic estimate of \f$ HSIC(X, Y) \f$.
     * The computation is based on the Gram matrices \f$ K_{x} \f$ and \f$ K_{y} \f$ which are constructed beforehand from the observations of \f$ X \f$ and \f$ Y \f$.
     * 
     * The formula of the U-statistic estimator can be found in:
     * [Zhang et al., 2018] "Large-scale kernel methods for independence testing", Statistics and Computing", page 117, Eq (8).
     *
     * Let \f$ L_{x} \f$ be the matrix obtained from the Gram matrix \f$ K_{x} \f$ after setting the diagonal elements to zero.
     * Let \f$ L_{y} \f$ be the matrix obtained from the Gram matrix \f$ K_{y} \f$ after setting the diagonal elements to zero.
     * Let \f$ 1 \f$ be the vector of \f$ \mathbb{R}^{nS} \f$ filled with ones.
     * 
     * The formula of the U-statistic estimator is given by:
     * 
     * \f[ \widehat{HSIC}_{U}(X, Y) \, = \, \frac{1}{ nS (nS-3) } \, \left[ 
     * Tr( L_{x} L_{y} )
     * + \frac{1}{ (nS-1)(nS-2) } \, 1^T L_{x} 1 \times 1^T L_{y} 1
     * - \frac{2}{ nS-2 } \, 1^T L_{x} L_{y} 1 \right] \f]
     *
     * \param matKx (TMatrixDSym) The Gram matrix built from the observations of \f$ X \f$. 
     * \param matKy (TMatrixDSym) The Gram matrix built from the observations of \f$ Y \f$. 
     * 
     * \return (Double_t) The value of the U-statistic estimate of \f$ HSIC(X, Y) \f$.
     */
    Double_t unbiasedHSIC(const TMatrixDSym& matKx, const TMatrixDSym& matKy);

    /*!
     * \brief Estimates the product of the squared Hilbert-Schmidt (HS) norms of the cross-covariance operators \f$ C_{xx} \f$ and \f$ C_{yy} \f$.
     * 
     * This quantity appears in the asymptotic variance of the test statistic (either the U-statistic or V-statistic estimator of the HSIC) under the null hypothesis.
     * The estimator is equal to the mean of all the off-diagonal coefficients of the following matrix:
     * \f$ \left( \widetilde{K}_{x} \odot \widetilde{K}_{y} \right)^{ \odot 2} \f$ where:
     * - \f$ \widetilde{K}_{x} \f$ and \f$ \widetilde{K}_{y} \f$ are the double-centered matrices derived from \f$ K_{x} \f$ and \f$ K_{y} \f$,
     * - \f$ \odot \f$ denotes the Hadamard product (element-wise product).
     * 
     * Further details can be found in [Gretton et al., 2007]. 
     * Note that the product of the squared HS norms appears in Theorem 4 and Appendix A.2.
     * 
     * \param matKx (TMatrixDSym) The input Gram matrix \f$ K_{x} \f$.
     * \param matKy (TMatrixDSym) The output Gram matrix \f$ K_{y} \f$.
     * 
     * \return (Double_t) An estimate of the product of the two squared HS norms.
     */
    Double_t getProductOfNorms(const TMatrixDSym& matKx, const TMatrixDSym& matKy);

    /*!
     * \brief Given a Gram matrix and a list of permutations, constructs the permuted matrices and store them in a list.
     *
     * Let us recall that a permutation is a one-to-one map \f$ \sigma: \{ 1, \ldots, nS \} \rightarrow \{ 1, \ldots, nS \} \f$.
     * For a given Gram matrix \f$ K = \left[ K(i,j) \right]_{1 \leq i, j \leq nS} \f$ and a given permutation \f$ \sigma \f$, the permuted Gram matrix is defined as:
     * \f[ K_{\sigma} := \left[ K \left( \sigma(i), \sigma(j) \right) \right]_{1 \leq i, j \leq nS} \f].
     *
     * \param theMatK (TMatrixDSym) Gram matrix that is to be permuted.
     *
     * \param tlstPerm (TList*) Pointer to the list of permutations.
     * Each element in *tlstPerm is a pointer to a TPatternsEventList (class defined in the DataServer module).
     * The first element in *tlstPerm must be a pointer to a TPatternsEventList representing the identity matrix.
     *
     * \param tlstPermMat (TList*) Pointer to the list of permuted Gram matrices. 
     * An empty list must be passed as argument. 
     * For each new permutation found in *tlstPerm, theMatK is permuted accordingly, and a pointer to the resulting TMatrixDSym is stored in *tlstKPermMat.
     */
    void permuteKMatrix(const TMatrixDSym& theMatK, TList* tlstPerm, TList* tlstPermMat);

    //@}



    //---------------------------------------------
    /** @name Overloaded methods
     **/
    //@{

    /*!
     * \brief No action is performed, but a warning is returned.  
     *
     * Remember that any THSIC instance is created from an already filled TDataServer.
     *
     * \param option (Option_t*) Useless here.
     */
    virtual void generateSample(Option_t* option="");

    /*!
     * \brief Calls the method evaluateIndexes.
     * 
     * \param option (Option_t*) Specifies the option to pass to the method evaluateIndexes.
     */
    void computeIndexes(Option_t* option="");

    /*!
     * \brief Applies the HSIC methodology. 
     *
     * For each input-output pair \f$ (X_i, Y_j) \f$, this function estimates three sensitivity indices:
     * - the (standard) HSIC index, 
     * - the normalized HSIC index (also called R2-HSIC index),
     * - the p-value of the independence test. 
     * 
     * \param option (Option_t*) Some possible options (level of verbosity, estimator type, test procedure).
     * 
     * By default:
     * - The function operates in verbose mode, providing many details.
     * - HSIC indices (and R2-HSIC indices) are estimated with V-statistics, which yields (slightly) biased but non-negative estimates.
     * - The p-values are estimated with the asymptotic test procedure (based on asymptotic formulas).
     * 
     * The `option` argument allows customization. It should be a concatenation of substrings, such as:
     * "[option_1]" or "[option_1],[option_2]" or "[option_1],[option_2],[option_3]".
     * 
     * Valid substrings include:
     * - "quiet" to enable the silent mode (instead of the verbose mode), 
     * - "unbiased" to estimate HSIC indices with U-statistics (instead of V-statistics),
     * - "nperm=500" to estimate the p-values with the permutation-based test procedure (instead of the asymptotic test procedure).
     * This procedure will apply \f$ B = 500 \f$ random permutations on the output data before estimating the p-value. 
     * Note that the number of permutations can be set to any integer greater than \f$ 7 \f$.
     */
    void evaluateIndexes(Option_t* option);

    /*!
     * \brief Draw the indexes computed by the method
     *
     * \param sTitle (TString) the title of the graph
     * \param select (const char*) the select argument
     * \param option (Option_t*) option to pass [""]
     */
    void drawIndexes(TString sTitle, const char* select="", Option_t* option="");

    //@}



    ClassDef(URANIE::Sensitivity::THSIC, ID_SENSITIVITY)

        //---------------------------------------------
        /** @name Private and Public attributes
         *
         **/
        //@{

        public:
            TMatrixDSym _matH;                        ///< The H matrix defined by H := ( Kronecker(i,j) - 1/nS)
        private:
            EEstimatedVariance _nEstimatedVariance;   //! the method to compute the variance for the gaussian kernel
                Int_t _nPermutationSample;                  //! The number of permutation sample ( permutation )         (default NDefaultPermutation)
                Int_t _nThresholdPermutation;                   //! The Threshold (nS/nX) to use the permutation permutation test to cumpute the p-value (default 5)
                Int_t _nThresholdGamma;                       //! The Threshold (nS/nX) to use the Gamma approximation to cumpute the p-value (default 10)

        //@}



};

} // End Of namespace Sensitivity
} // End Of namespace URANIE

#endif
