////////////////////////////////////////////////////////////// 
// Copyright (C) 2013-2025 CEA/DES 
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
\file   TestTCramerVonMises.cxx
\author Rudy Chocat <rudy.chocat@cea.fr>
\date   2024-09-09 11:19:38

\brief Unit test for URANIE::Sensitivity::TestTCramerVonMises class

*/

// Uranie
#include "TestTCramerVonMises.h"
#include "TSensitivity.h"
#include "TSobol.h"
#include "UDisableTest.h"
#include "TOutputFileKey.h"
#include "TOutputFileRow.h"
#include "TOutputFileDataServer.h"
#include "TCode.h"
#include "TUniformDistribution.h"
#include "TLauncher.h"
#include "TLauncherFunction.h"
#include "TLauncher2.h"
#include "TCJitEval.h"
#include "TSequentialRun.h"
#include "TestUserFunctions.h"
#include "TestTSensitivity.h"
#include "code_paths.h"
#include "uranie_commands.h"

// ROOT
#include "TSystem.h"
#include "TString.h"
#include "TMath.h"
#include "TTreePlayer.h"
#include "TCanvas.h"
#include "TROOT.h"
#ifdef ROOT5
#include "Api.h"
#endif

#define EPSILON 0.01
#define LVLE_EPSILON 0.001 
#define DBL_PRECISION_COMPARE 1.e-12

#include <iostream>
#include <iomanip>



TCramerVonMisesTest::TCramerVonMisesTest() :
        _tds1(NULL), _tds(NULL), _tcvm(NULL), _tcvm1(NULL)
{
    _x1 = std::make_unique<URANIE::DataServer::TNormalDistribution>("x1");
    _x2 = std::make_unique<URANIE::DataServer::TNormalDistribution>("x2");
}

TCramerVonMisesTest::~TCramerVonMisesTest()
{
    
}

void TCramerVonMisesTest::setUp()
{
    _tds1 = new URANIE::DataServer::TDataServer();
    _tds1->addAttribute(_x1.get());
    _tds1->addAttribute(_x2.get());
    _tds = new URANIE::DataServer::TDataServer("tdsflowreate",
            "DataBase flowrate");
}


void TCramerVonMisesTest::tearDown()
{
    delete _tcvm;
    _tcvm = NULL;
    delete _tcvm1;
    _tcvm1 = NULL;
    delete _tds1;
    _tds1 = NULL;
    delete _tds;
    _tds = NULL;
}


void TCramerVonMisesTest::testCramerVonMisesTest()
{

    _tcvm = new URANIE::Sensitivity::TCramerVonMises(_tds1, &GamboaCVM,"x1:x2","y",50000);    
    _tcvm->setDrawProgressBar(kFALSE);
    _tcvm->setSeed(123456);
    _tcvm->computeIndexes();

    // Double_t dsum = 0.0, jsum=0.0;
    double s1 = (6.0/TMath::Pi()) * TMath::ATan(2.0)-2.0;
    double s2 = (6.0/TMath::Pi()) * TMath::ATan(TMath::Sqrt(19.0))-2.0;
    double s12 = 1.0 - s1 - s2;
    double st1 = s1 + s12;
    double st2 = s2 + s12;
    
    Double_t ref1[2] =  {s1, s2};
    Double_t refT[2] =  {st1, st2};
    
    for (Int_t i = 0; i < 2; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(ref1[i], _tcvm->getValue("FirstCVM",_tds1->getAttribute(i)->GetName()), 0.01);
       
    }
    for (Int_t i = 0; i < 2; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(refT[i], _tcvm->getValue("TotalCVM",_tds1->getAttribute(i)->GetName()), 0.01);
    }
}

void TCramerVonMisesTest::testIshigamiCramerVonMisesWithDrawings()
{

    _tds = new URANIE::DataServer::TDataServer("tdsFlowrate", "DoE for Flowrate");
    _tds->addAttribute(new URANIE::DataServer::TUniformDistribution("rw", 0.05, 0.15));
    _tds->addAttribute(new URANIE::DataServer::TUniformDistribution("r", 100.0, 50000.0));
    _tds->addAttribute(new URANIE::DataServer::TUniformDistribution("tu", 63070.0, 115600.0));
    _tds->addAttribute(new URANIE::DataServer::TUniformDistribution("tl", 63.1, 116.0));
    _tds->addAttribute(new URANIE::DataServer::TUniformDistribution("hu", 990.0, 1110.0));
    _tds->addAttribute(new URANIE::DataServer::TUniformDistribution("hl", 700.0, 820.0));
    _tds->addAttribute(new URANIE::DataServer::TUniformDistribution("l", 1120.0, 1680.0));
    _tds->addAttribute(new URANIE::DataServer::TUniformDistribution("kw", 9855.0, 12045.0));

    TString sJDDReference = gSystem->pwd();
#ifdef WIN32
    sJDDReference += "\\";
#else
    sJDDReference += "/";
#endif  
    sJDDReference += "flowrate_input_with_values_rows.in";
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);

    URANIE::Launcher::TOutputFileKey *_fout = new URANIE::Launcher::TOutputFileKey("_output_flowrate_withKey_.dat");

     _fout->addAttribute(new URANIE::DataServer::TAttribute("yhat"));

    URANIE::Launcher::TCode *_mycode = new URANIE::Launcher::TCode(_tds, UranieTests::code_path("flowrate") + " -s -r");
    _mycode->addOutputFile(_fout);

    _tcvm = new URANIE::Sensitivity::TCramerVonMises(_tds, _mycode, 220);
    _tcvm->setDrawProgressBar(kFALSE);
    _tcvm->setSeed(123456);
    _tcvm->computeIndexes();

    double ref1[8] = {0.651916311794112, -0.03885959773481741, -0.03885959773481741, -0.0423959548813691, -0.03718592964824126, -0.03896103896103903, -0.09306930693069307, -0.06792304672163327};
    double refT[8] = {0.8399710354815352, 0.0, 0.0, 0.008138351983723369, 0.2835696413678067, 0.2681081081081079, 0.15863076601962034, 0.04112687641373636};
    
    for (Int_t i = 0; i < 8; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(ref1[i], _tcvm->getValue("FirstCVM",_tds->getAttribute(i)->GetName()), 0.001);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(refT[i], _tcvm->getValue("TotalCVM",_tds->getAttribute(i)->GetName()), 0.001);
    }

    TCanvas *Can = new TCanvas("Can","Can",1);
    _tcvm->drawIndexes("Flowrate CVM", "", "nonewcanv,hist,first,total");
    Can->Clear();
    delete Can;

     _tcvm->drawIndexes("Flowrate CVM", "", "hist,first,total");
}





void TCramerVonMisesTest::testCramerVonMisesFunctionRelauncher()
{
    URANIE::DataServer::TAttribute *out = new URANIE::DataServer::TAttribute("Ishi");
    out->delShare();
    
    URANIE::Relauncher::TCJitEval eval(&GamboaCVM);
    eval.addInput(_tds1->getAttribute("x1"));
    eval.addInput(_tds1->getAttribute("x2"));
    eval.addOutput(out);
 	
    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if( run.onMaster() )
    {
	_tcvm1 = new URANIE::Sensitivity::TCramerVonMises(_tds1, &run,3000);
	_tcvm1->setSeed(123456);
	_tcvm1->setDrawProgressBar(false);
	_tcvm1->computeIndexes();

    Double_t ref1[2] =  {0.07461589551369263, 0.6089628356371951};
    Double_t refT[2] =  {0.41319877074891165, 0.9243676247859068}; 
    
    for (Int_t i = 0; i < 2; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(ref1[i], _tcvm1->getValue("FirstCVM",_tds1->getAttribute(i)->GetName()), 0.000001);
       
    }
    for (Int_t i = 0; i < 2; i++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(refT[i], _tcvm1->getValue("TotalCVM",_tds1->getAttribute(i)->GetName()), 0.000001);
    }


	run.stopSlave();		
    }
}
