//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   ishigami.cxx
 \author Rudy Chocat Cedric <uranietm@cea.fr>
 \date   2024
 
 \brief  "Pseudo" Code \b ishigami allowing to show functionalities of Uranie Launcher. Synthesis of ishigami and ishigamiStop executables.

 \section origin_flow_rate_water Origine
 Example given in <em>Design and Modeling for Computers Experiments</em> de <em>Kai-Tai Fang</em>, <em>Runze Li</em> et <em>Agus Sudjianto</em>

 Analytic function \f$ \re^{3}\f$ in \f$\re\f$ which formula is:
 \f[
 y = f(x) = \sin(x_1) + a \sin(x_2)^{2} + b  x_{3}^{4} sin(x_2) 
 \f]

 The parameters of these function are :
 \li \f$ x_{i} \in [-\pi, \pi], i=1,2,3\f$
 \li \f$ a = 7 f$
 \li \f$ b = 0.1\f$ 


 */
#define MAX_LENGTH 256

//!  Output file with values in row and a headernew_name=
#define OUTPUT_FILE_WITHROW "_output_ishigami_withRow_.dat"
//! second output file. Contains cumulated sums for each file of all variables.
#define OUTPUT_FILE2 "_output2_.dat"
//#define  bdebug 0
//!  Output file with the  "Key = value" format.
#define OUTPUT_FILE_WITHKEY "_output_ishigami_withKey_.dat"


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <iostream>     // for I/O
#include <fstream>      // for file I/O
#include <sstream>   
#include <cstring>

#include "TamponTexte.h"

#include "TMath.h"

int bdebug = 0;
int nloop = 1;

//---------------------------------------------
/** @name Methods to construct ishigami executable.
 **/
//@{
/*!
 Returns the list of options available with ishigami.
 -d:    debug mode.
 -s:    silent mode.
 -k:    input file ishigami_input_with_keys.in with keys.
 -f:    input file with ishigami_input_with_flags.in with flags.
 -kf:   input file ishigami_input_with_keys.in with failure mode.  
 -ff:   input file ishigami_input_with_flags.in with failure mode.
 -h,-?: prints help message.  
 */
void usage(const char *scmd)
{
    printf("Usage: %s [-d] [-s] [-k|-f|-r|-x|-kf|-ff|-xf|-mkk|-mkf|-mff|-Folkk|-Folkf|-Folff] [-h|-?] [-mt] [file]\n",
            scmd);
    printf("     -d:   debug mode\n");
    printf("     -s:   silent mode\n");
    printf("     -k:   input file with keys  [ishigami_input_with_keys.in]\n");
    printf("     -f:   input file with flags [ishigami_input_with_flags.in]\n");
    printf("     -r:   input file with values in rows [ishigami_input_with_values_rows.in]\n");
    printf("   -h,-?:   this help message\n\n");

}

/*!
 Reads file with values in rows.
 */
int readInputFileWithValuesRows(const char *sfile, double *input)
{
    if (bdebug)
        std::cout << "***** readInputFileWithValuesRows file[" << sfile << "] ... "
                << std::endl;

    int nfound = -1;

    int nparamToRead = 5;

    std::ifstream fileInput(sfile, std::ios::in);
    if (fileInput)
    {
        if (bdebug)
            std::cout << " ** file open ifstream " << std::endl;
    }
    else
    {
        if (bdebug)
            std::cout << "********************* " << std::endl;
        return nfound;
    }
    nfound++;
    for (int i = 0; i < nparamToRead; i++)
    {
        fileInput >> input[i];
        if (!fileInput.good())
            break;
        if (bdebug)
            std::cout << " ** i[" << i << "/" << nparamToRead << "] value["
                    << input[i] << "]" << std::endl;
        nfound++;
    }

    fileInput.close();
    return nfound;
}


/*!
 Reads file with flags.
 */
int readInputFileWithFlags(const char *sfile, double *input)
{

    if (bdebug)
        std::cout << "***** readInputFileWithFlags file[" << sfile << "] ... "
                << std::endl;

    int nfound = 0;

    FILE *fin = fopen(sfile, "r");
    if (fin)
    {
        if (bdebug)
            std::cout << " ** file open" << std::endl;
    }
    else
    {
        if (bdebug)
            std::cout << "********************* " << std::endl;
        return -1;
    }
    TamponTexte fnomFichier(sfile, 0);
    //   fnomFichier.initSeparateurs("\t\n");
    //   fnomFichier.initCarCommentaires("#");
    fnomFichier.initSeparateurs(" \n");
    fnomFichier.initCarCommentaires("#");
    //   // elimination of white lines
    //   fnomFichier.elimineBlancs();
    //   // elimination des lignes de commentaires
    //fnomFichier.elimineCommentaires();
    // Test si le fichier ainsi obtenu est vide
    if (fnomFichier.estVide())
    {
        std::cerr << "\n le fichier: " << fnomFichier.nom() << " est vide.\n";
        return -1;
    }
    if (bdebug)
    {
        std::cout << " ** Lines : " << fnomFichier.nombreLignes() << std::endl;
    }
    for (int iline = 1; iline <= fnomFichier.nombreLignes(); iline++)
    {
        if (bdebug)
            std::cout << " ** Line " << iline << "/" << fnomFichier.nombreLignes()
                    << " : Field [ " << fnomFichier.nombreChampsLigne(iline)
                    << "]";
        if (fnomFichier.nombreChampsLigne(iline) > 1)
        {
            const char *strclef = fnomFichier.champLigne(1, iline);
            if (bdebug)
                std::cout << " - Key[" << strclef << "]";

            // Case of parameter x1 and x2
            if (strcmp("parameter12", strclef) == 0)
            {
                nfound++;
                input[0] = atof(fnomFichier.champLigne(3, iline));
                nfound++;
                input[1] = atof(fnomFichier.champLigne(4, iline));
            }
            // Case of parameter x3
            if (strcmp("parameter3", strclef) == 0)
            {
                nfound++;
                input[2] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter tl
            if (strcmp("constantA", strclef) == 0)
            {
                nfound++;
                input[3] = atof(fnomFichier.champLigne(2, iline));
            }
            // Case of parameter hu
            if (strcmp("constantB", strclef) == 0)
            {
                nfound++;
                input[4] = atof(fnomFichier.champLigne(2, iline));
            }

        }
        if (bdebug)
            std::cout << std::endl;
    }

    return nfound;
}

/*!
 Reads file with keys.
 */
int readInputFileWithKeys(const char *sfile, double *input)
{
    if (bdebug)
        std::cout << "***** readInputFileWithKeys file[" << sfile << "] ... "
                << std::endl;

    int nparamToRead = 5;
    int nfound = 0;

    FILE *fin = fopen(sfile, "r");
    if (fin)
    {
        if (bdebug)
            std::cout << " ** file open" << std::endl;
    }
    else
    {
        if (bdebug)
            std::cout << "********************* " << std::endl;
        return -1;
    }

    TamponTexte fnomFichier(sfile, 0);
    fnomFichier.initSeparateurs(" =\n");
    fnomFichier.initCarCommentaires("#");

    // Eliminatioilinen des lignes vides
    fnomFichier.elimineBlancs();
    // Elimination des lignes de commentaires
    fnomFichier.elimineCommentaires();
    // Test si le fichier ainsi obtenu est vide
    if (fnomFichier.estVide())
    {
        std::cerr << "\n le fichier: " << fnomFichier.nom() << " est vide.\n";
        exit(1);
    }
    if (bdebug)
    {
        std::cout << " ** Ouverture du fichier " << fnomFichier.nom() << " terminee."
                << std::endl;
        std::cout << " ** fichier " << fnomFichier.nom() << " : "
                << fnomFichier.nombreLignes() << " lignes initiales" << std::endl;
    }
    for (int iline = 1; iline <= fnomFichier.nombreLignes(); iline++)
    {
        if (bdebug)
            std::cout << "i[" << iline << "]" << fnomFichier.nombreChampsLigne(iline)
                    << std::endl;
        if (fnomFichier.nombreChampsLigne(iline) > 1)
        {
            const char *strclef = fnomFichier.champLigne(1, iline);
            if (bdebug)
                std::cout << "i[" << iline << "] clef[" << strclef << "]" << std::endl;

            if (strcmp("x1", strclef) == 0)
            {
                input[0] = atof(fnomFichier.champLigne(2, iline));
                nfound++;
            }
            if (strcmp("x2", strclef) == 0)
            {
                input[1] = atof(fnomFichier.champLigne(2, iline));
                nfound++;
            }
            if (strcmp("x3", strclef) == 0)
            {
                input[2] = atof(fnomFichier.champLigne(2, iline));
                nfound++;
            }
            if (strcmp("a", strclef) == 0)
            {
                input[3] = atof(fnomFichier.champLigne(2, iline));
                nfound++;
            }
            if (strcmp("b", strclef) == 0)
            {
                input[4] = atof(fnomFichier.champLigne(2, iline));
                nfound++;
            }
        }
    }

    // Close the input file
    fclose(fin);

    if (bdebug)
        std::cout << "**** End of readInputFileWithKeys file[" << sfile
                << "] - Found [" << nfound << "]" << std::endl;

    return nfound;
}

double ishigami(double *x)
{
    double x1 = x[0], x2 = x[1], x3 = x[2];
    double a = x[3], b = x[4];
    
    double term1 = TMath::Sin(x1);
    double term2 = a * TMath::Sin(x2) * TMath::Sin(x2);
    double term3 = b * x3 * x3 * x3 * x3 * TMath::Sin(x1);
    return term1 + term2 + term3;
}
//@}

//! 
/*! 
 
 \param argc le nombre d'argument
 \param argv la liste des arguments

 \return indice de l'erreur de retour
 */
int main(int argc, char *argv[])
{
    FILE * fout;
    char sfile[MAX_LENGTH];
    char * option = NULL;
    // Read the arguments
    bool bFilesFollow = false;
    bool bFileByKey = false;
    bool bFileByFlag = false;
    bool bFileByRecreate = false;
    bool bSilent = true;
    for (int i = 0; i < argc; i++)
    {

        if (!strcmp(argv[i], "-d"))
        {
            bdebug = 1;
        }
        if (!strcmp(argv[i], "-s"))
        {
            bSilent = false;
        }

        if (!strcmp(argv[i], "-r") || !strcmp(argv[i], "-rf"))
        {
            option = argv[i];
            bFileByRecreate = true;
            if (i != argc - 1)
                sprintf(sfile, "%s", argv[i + 1]);
            else
                sprintf(sfile, "%s", "ishigami_input_with_values_rows.in");
        }

        if (!strcmp(argv[i], "-k") || !strcmp(argv[i], "-kf") || !strcmp(argv[i], "-ks"))
        {
            option = argv[i];
            bFileByKey = true;
            if (i != argc - 1)
                sprintf(sfile, "%s", argv[i + 1]);
            else
                sprintf(sfile, "%s", "ishigami_input_with_keys.in");
        }

        if (!strcmp(argv[i], "-f") || !strcmp(argv[i], "-ff") || !strcmp(argv[i], "-ff"))
        {
            option = argv[i];
            bFileByFlag = true;
            if (i != argc - 1)
                sprintf(sfile, "%s", argv[i + 1]);
            else
                sprintf(sfile, "%s", "ishigami_input_with_flags.in");
        }

        
        if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "-?"))
        {
            usage(argv[0]);
            exit(0);
        }
    }

    if ((bFileByRecreate == false) && (bFileByKey == false)
	&& (bFileByFlag == false))
    {
        std::cerr
	    << " ** ERROR : You must select the flags (-r) OR (-f, -ff) OR keys (-k, -kf) OR XML (-x, -xf) option"
	    << std::endl;
        usage(argv[0]);
        exit(-1);
    }
    if (bdebug)
        std::cout << "with debug " << std::endl;

    // Allocate parameters vectors
    int nparam = 5;
    int nfound = 0;
    double *dparam = new double[nparam];
    // Default value
    dparam[3] = 7.0;
    dparam[4] = 0.1;
   


        //Reads the input parameters with one file
        if (bFileByRecreate)
            nfound = readInputFileWithValuesRows(sfile, dparam);

        if (bFileByKey)
            nfound = readInputFileWithKeys(sfile, dparam);

        if (bFileByFlag)
            nfound = readInputFileWithFlags(sfile, dparam);

    
    
    if (nfound != nparam)
    {
        std::cerr << " ** ERROR : Pb in read file [" << sfile << "] found[" << nfound
	     << "] param[" << nparam << "]" << std::endl;
        exit(-1);
    }
    if (bSilent)
    {
        std::cout << "******************" << std::endl;
        std::cout << "*** Input File[" << sfile << "] NbParam[" << nparam << "]" << std::endl;
        std::cout << "******************" << std::endl;
        for (int i = 0; i < nparam; i++)
            std::cout << "** i[" << i << "] value[" << dparam[i] << "]" << std::endl;
        std::cout << "******************" << std::endl;
    }

    // Computes the outputs
    double y = ishigami(dparam);
    

    ////////////////////////////////
    // Writes the ouput in files
    ////////////////////////////////
    // File with Row values and a header
    fout = fopen(OUTPUT_FILE_WITHROW, "w");
    fprintf(fout, "#COLUMN_NAMES: yhat \n\n%e", y);
    fclose(fout);
    // File with "key = value" format; 
    fout = fopen(OUTPUT_FILE_WITHKEY, "w");
    fprintf(fout, "yhat\t=\t%e;\n", y);
    fclose(fout);
    ////////////////////////////////

    if (bSilent)
    {
        std::cout << "******************" << std::endl;
        std::cout << "***  Ouput Files [" << OUTPUT_FILE_WITHROW << "] and ["
	     << OUTPUT_FILE_WITHKEY << "]";
        std::cout << std::endl;
        std::cout << "******************" << std::endl;
        std::cout << "** y[0] value[" << y << "]" << std::endl;
        std::cout << "******************" << std::endl;
    }

    delete[] dparam;
    dparam = NULL;
        
    return 0;

}
