// --- C++ ---
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2023 CEA/DEN
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   TCodeAsEval.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2023-03-16
 
 \brief  using launcher TCode as a TEval
*/

#include "TCodeAsEval.h"
#include "forkcode.h"

//ROOT

//URANIE


ClassImp(URANIE::Relauncher::TCodeAsEval);

namespace URANIE
{
namespace Relauncher
{

/**** constructeur destructeur ****/
TCodeAsEval::TCodeAsEval(URANIE::Launcher::TCode *code) :
    TShareEval(), URANIE::Launcher::TLauncherEnv(code), _fjob(NULL)
{
}

TCodeAsEval::~TCodeAsEval()
{
    for (auto& value : _outputValues) delete value;
    _outputValues.clear();
    for (auto& value: _inputValues) delete value;
    _inputValues.clear();
    if (_fjob) delete _fjob;
}

/**** helpers ****/
void TCodeAsEval::setJob(URANIE::Launcher::forkcode *job)
{
    URANIE::DataServer::TAttribute *att;
    
    _fjob = job;
    _code->initRun(_fjob);
    
    // alloc input and output entry
    if (_inputValues.size() != 0 && _outputValues.size() !=0) return;
    
    // inputs
    _inputValues.push_back( new URANIE::DataServer::UEntry(
            URANIE::DataServer::UEntry::Double, "iterator" )
        );
    TIter nexti(&_input);
    while ( (att = (URANIE::DataServer::TAttribute*) nexti()) ) {
        switch (att->getDataType()) {
        case URANIE::DataServer::TAttribute::kVector :
            _inputValues.push_back( new URANIE::DataServer::UEntry(
                    URANIE::DataServer::UEntry::Vector, string(att->GetName())
                ));
            break;
        case URANIE::DataServer::TAttribute::kString :
            _inputValues.push_back( new URANIE::DataServer::UEntry(
                    URANIE::DataServer::UEntry::String, string(att->GetName())
                ));
            break;
        //case URANIE::DataServer::TAttribute::kDouble :
        default :
            _inputValues.push_back( new URANIE::DataServer::UEntry(
                    URANIE::DataServer::UEntry::Double, string(att->GetName())
                ));
            break;
        }
    }
    // outputs
    _outputValues.push_back( new URANIE::DataServer::UEntry(
            URANIE::DataServer::UEntry::Double, "iterator_tot" )
        );
    _outputValues.push_back( new URANIE::DataServer::UEntry(
            URANIE::DataServer::UEntry::Double, "n__calculs__" )
        );
    TIter nexto(&_output);
    while ( (att = (URANIE::DataServer::TAttribute*) nexto()) ) {
        switch (att->getDataType()) {
        case URANIE::DataServer::TAttribute::kVector :
            _outputValues.push_back( new URANIE::DataServer::UEntry(
                    URANIE::DataServer::UEntry::Vector, string(att->GetName())
                ));
            break;
        case URANIE::DataServer::TAttribute::kString :
            _outputValues.push_back( new URANIE::DataServer::UEntry(
                    URANIE::DataServer::UEntry::String, string(att->GetName())
                ));
            break;
        //case URANIE::DataServer::TAttribute::kDouble :
        default :
            _outputValues.push_back( new URANIE::DataServer::UEntry(
                    URANIE::DataServer::UEntry::Double, string(att->GetName())
                ));
            break;
        }
    }
    
}

void TCodeAsEval::prepareEntries(URANIE::Launcher::forkcode *fjob, vector<URANIE::DataServer::UEntry*> *in)
{
    int i;
    URANIE::DataServer::TAttribute *att;
    
    // clean
    for (auto uent : _inputValues) uent->clearEntry();
    
    // set values
    i = 0;
    _inputValues.at(i)->pushEntry(1+fjob->pattern);
    TIter nexti(&_input);
    while ( (att = (URANIE::DataServer::TAttribute *) nexti()) ) {
        _inputValues.at(i+1)->copyEntry(in->at(i));
        i++;
    }
}

/**** protocole ****/
void TCodeAsEval::freeze()
{
    int i, j;
    
    if (_freezed) return;
    TBaseEval::freeze();
    
    // dir
    if (_code) {
        verifyWorkingDirectory();
        createResultsDirectory();
        _code->initFileType();
        _code->init();
        // input list
        for (i=0; i<_code->getNInputFiles(); i++) {
            for (j=0; j<_code->getNInputAttributes(i); j++) {
                _input.Add(_code->getInputAttribute(i, j));
            }
        }
        // output list
        for (i=0; i<_code->getNOutputFiles(); i++) {
            for (j=0; j<_code->getNOutputAttributes(i); j++) {
                _output.Add(_code->getAttribute(i, j));
            }
        }
    }
    
}


TEval * TCodeAsEval::rebuild(int rank, Bool_t chef)
{
    if (chef) {
        TCodeAsEval *ret = new TCodeAsEval(*this);
        auto fjob = new URANIE::Launcher::forkcode;
        fjob->pattern = rank;
        fjob->command = "";
        ret->setJob(fjob);
        return (TEval*) ret;
    }
    else return (TEval *) this; 
        
}

Bool_t TCodeAsEval::unbuild()
{
    return kTRUE;
}


int TCodeAsEval::eval(vector<URANIE::DataServer::UEntry*> *in, vector<URANIE::DataServer::UEntry*> *out, int ncalc)
{
    Bool_t post;
    URANIE::DataServer::TAttribute *att;
    
    _code->cleanOutputFiles(_fjob);
    // prepare inputs
    _fjob->pattern = ncalc;
    prepareEntries(_fjob, in);
    // code call 
    _code->preTraitment(_fjob, &_inputValues);
    _code->run(_fjob);
    post = _code->postTraitment(_fjob, &_outputValues);
    // get outputs
    if (!post) std::cout << "arg code" << std::endl;
    else {
        int i = 0;
        TIter nexto(&_output);
        while ( (att = (URANIE::DataServer::TAttribute *) nexto()) ) {
            out->at(i)->copyEntry(_outputValues.at(i+2));
            i++;
        }
    }
    
    return (post) ? getOutDimension(out) : 0 ;
}

}} //namespace

