// --- C++ ---
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2024 CEA/DEN
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   TCodeBase.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2024-07-02
 
 \brief  gestion des répertoires fichiers de travail
*/


#include "TCodeBase.h"

#ifdef WIN32
#include <io.h>
#else
//#include <sys/wait.h>
#endif

//ROOT
#include "TSystem.h"

//URANIE
#include "UExceptions.h"

ClassImp (URANIE::Relauncher::TCodeBase);

namespace URANIE
{
namespace Relauncher
{

/*** TDirEval ***/
/* Constructeur Destructeur */
TDirEval::TDirEval() :
    _isnumered(kFALSE), _keepAllFolders(false), 
    _initialdir(gSystem->pwd()),
    _workdir(""), _tmpdir("")
{
}

TDirEval::~TDirEval(void) {}

/* more definition */
void TDirEval::setWorkingDir(TString dir)
{
    //if (_freezed) throw FreezedEval();
    _workdir = (dir.Length() == 0)? _initialdir :  dir;
}

void TDirEval::setOldTmpDir()
{
    _isnumered = kTRUE;
}

void TDirEval::keepAllFolders()
{
    _keepAllFolders = kTRUE;
}

/* Close definition */
void TDirEval::buildWorkDir()
{
    /* working directory */
    if (_workdir.Length() == 0)
        _workdir = _initialdir;

#ifdef IMPLICIT
    void *dir = NULL;
    // bug mpi conflict access
    try
    {
        // Create this directory if it doesn't exist
        if (!(dir = gSystem->OpenDirectory(_workdir)))
	    if (gSystem->mkdir(_workdir, kTRUE) == -1)
		throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__, Form("TDirEval::buildDir ERROR : the directory \"%s\" can not be created", _workdir.Data()));
    }
    catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    if (dir) gSystem->FreeDirectory(dir);
#endif

}

/* execution environment */
void TDirEval::buildTmpDir(int id, Bool_t chef)
{
    void *dir = NULL;

    if (_isnumered || _keepAllFolders) {
        try {
#ifdef WIN32
            _tmpdir = TString::Format("%s\\URANIE%d", _workdir.Data(), id);
#else		
            _tmpdir = TString::Format("%s/URANIE%d", _workdir.Data(), id);
#endif
            // Create this directory if it doesn't exist
            if (chef 
                && !(dir = gSystem->OpenDirectory(_tmpdir))
                && (gSystem->mkdir(_tmpdir.Data(), kFALSE) == -1)
                ){
                throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
						      Form("TDirEval::setTmpDir ERROR : the directory \"%s\" can not be created", _tmpdir.Data())
                    );
            }
        }
        catch (URANIE::Exceptions::UExceptions& ue) {
            ue.printMessage();
            throw ue;
        }
        if (dir) gSystem->FreeDirectory(dir);
    }
    else {
        try {
            char* tmp = new char[_workdir.Length()+12];
            // tmpdir
            strcpy(tmp, _workdir.Data());
            
#ifndef WIN32	
            strcat(tmp, "/URA_XXXXXX" );
            if (!mkdtemp(tmp)) {
                throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__, 
						      Form("TDirEval::setTmpDir ERROR : the directory \"%s\" can not be created", tmp)
                    );
            }
#else	
            strcat(tmp, "\\URA_XXXXXX" );
            if (_mktemp_s(tmp,_workdir.Length()+12) || (gSystem->mkdir(tmp, kFALSE) == -1) ) {	
               	
		throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__, 
						      Form("TDirEval::setTmpDir ERROR : the directory \"%s\" can not be created", tmp)
                    );
	
            }
#endif	 		
            _tmpdir = TString(tmp);
            delete[] tmp;
        }
        catch (URANIE::Exceptions::UExceptions& ue) {
            ue.printMessage();
            throw ue;
        }
        
    }
}

/*** TIOFileEval ***/
/* Constructeur Destructeur */
TIOFileEval::TIOFileEval() : 
    TDirEval(),
    _clone(kFALSE),
    _localdir("")
{
    _scripts = new TList();
    _result = NULL;
}

TIOFileEval::~TIOFileEval(void)
{
    if (!_tmpdir.Length())
        delete _scripts;
}

/* Complete Definition */
void TIOFileEval::addInputFile(TCodeScript *f)
{
    // if (_freezed) throw FreezedEval();
    _scripts->Add((TObject *) f);
}

void TIOFileEval::addOutputFile(TCodeResult *f)
{
    //if (_freezed) throw FreezedEval();
    if (_result)
        std::cout << "Multi output file: unimplemented feature\n";
    else
        _result = f;

}

void TIOFileEval::setReference(TString dir)
{
    //if (_freezed) throw FreezedEval();
    _localdir = dir;
}

/* Close definition */
void TIOFileEval::buildDir(void)
{
    /* build file */
    if (_localdir.Length() == 0)
        _localdir = _initialdir;

    /* working directory */
    buildWorkDir();

}

/* execution environment */
void TIOFileEval::setTmpDir(int id, Bool_t chef)
{
    /* dir */
    buildTmpDir(id, chef);
}

void TIOFileEval::setClone() { _clone = kTRUE; }


// TCodeBase
TCodeBase::TCodeBase() :
    TShareEval(),
    TIOFileEval()
{
}

TCodeBase::~TCodeBase(void)
{
}

void TCodeBase::freeze(void)
{
    TCodeScript *fs;
    TIter nexts(_scripts);

    if (_freezed)
        return;
    TBaseEval::freeze();

    buildDir();
    /* build file */
    while ((fs = (TCodeScript *) nexts()))
        fs->build(_localdir, &_input);

    try
    {
        if(_result==NULL)
            throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__, "TCodeBase::freeze ERROR : there is no output file defined in this code");
    }
    catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    _result->build(&_output);
}

/* TRun use */
/* execution environment */
int TCodeBase::putAndRun(Double_t *in, int ncalc)
{
    int ret;
    TCodeScript *fs;
    TIter next(_scripts);

    if( _keepAllFolders )
        setTmpDir( ncalc, (_subeval)? kFALSE : kTRUE );

    /* input file */
    if (!_subeval) {
        while ((fs = (TCodeScript *) next())) fs->put(_localdir, _tmpdir, in);
    }
    
    if(_keepAllFolders!=true)
        _result->clearFile(_tmpdir);

    /* execute */
    ret = callFunction();	
    return ret;
}

int TCodeBase::putAndRun(vector<URANIE::DataServer::UEntry*> *in, int ncalc)
{
    int ret;
    TCodeScript *fs;
    TIter next(_scripts);
    
    if( _keepAllFolders )
        setTmpDir( ncalc, (_subeval)? kFALSE : kTRUE );

    /* input file */
    if (!_subeval) {
        while ((fs = (TCodeScript *) next())) fs->put(_localdir, _tmpdir, in);
    }

    if(_keepAllFolders!=true)
        _result->clearFile(_tmpdir);

    /* execute */
    ret = callFunction();
      
    return ret;
}

int TCodeBase::eval(vector<URANIE::DataServer::UEntry*> *in, vector<URANIE::DataServer::UEntry*> *out, int ncalc)
{
    if (_subeval && !_mpi) return _output.GetEntries();

    if (_clone) {
        putAndRun(in, ncalc);

        /* output file */
        if (_subeval) return 1;
        else {
            int ret = _result->get(_tmpdir, out);
            return ((ret!=0) ? getOutDimension(out) : ret);
        }
    }
    else return 1;
}


}} //namespace